open Items
open Pool
open Table

module type Lattice = 
sig
   type d
   val cup : d -> d -> d
   val cap : d -> d -> d
   val eq  : d -> d -> bool
   val leq : d -> d -> bool

   val bot : unit -> d
   val top : unit -> d
   val widen  : d -> d -> d
   val narrow : d -> d -> d

   type item = d
   val to_string     : item -> string
   val to_int        : item -> int
end 

module IntLattice = 
struct
   type d = int
   type item = int
   let to_int x = x
   let to_string x = string_of_int x

   let top_value = ref (max_int)
   let bot_value = ref (min_int)
   let init_top x = (top_value := x)
   let init_bot x = (bot_value := x)
   let top() = !top_value
   let bot() = !bot_value

   let cup x y = if x>y then x else y
   let cap x y = if x>y then y else x
   let eq  x y = (x=y)
   let leq x y = x <= y

   let widen x y  = cup x y 
   let narrow x y = x
end 

module BooleanLattice =
struct
   type d = bool
   type item = bool
   let to_int = function
	 false -> 0
	| true -> 1

   let to_string = function
	false -> "false"
	| true -> "true"

   let top() = true
   let bot() = false

   let cup x y = x || y
   let cap x y = x && y
   let eq  x y = (x=y)
   let leq x y = if x then y else true

   let widen  x y = x || y
   let narrow x y = x
end

module UnitLattice =
struct
   type d = unit
   type item = unit
   let to_int () = 42

   let to_string () = "()"

   let top() = ()
   let bot() = ()

   let cup x y = ()
   let cap x y = ()
   let eq  x y = true
   let leq x y = true

   let widen  x y = ()
   let narrow x y = x
end

module Product (X: Lattice) (Y: Lattice) =
struct
   module P = PairItem (X)(Y)
   include P
   type d = item
   let top () = (X.top(),Y.top())
   let bot () = (X.bot(),Y.bot())
   let cup (x1,y1) (x2,y2) = (X.cup x1 x2, Y.cup y1 y2)
   let cap (x1,y1) (x2,y2) = (X.cap x1 x2, Y.cap y1 y2)
   let widen (x1,y1) (x2,y2) = (X.widen x1 x2, Y.widen y1 y2)
   let narrow (x1,y1) (x2,y2) = (X.narrow x1 x2, Y.narrow y1 y2)
   let eq  (x1,y1) (x2,y2) = X.eq x1 x2  && Y.eq y1 y2
   let leq (x1,y1) (x2,y2) = X.leq x1 x2 && Y.leq y1 y2
end

module Sum (X: Lattice) (Y: Lattice) =
struct
   type d = Bot | First of X.d | Second of Y.d | Top
   type item = d

   let bot() = Bot
   let top() = Top

   let put_fst x = if X.leq x (X.bot()) then Bot 
	       else First x
   let put_snd y = if Y.leq y (Y.bot()) then Bot
	       else Second y

   let get_fst = function
	  Top -> X.top()
	| Bot -> X.bot()
	| First x -> x
	| Second _ -> X.bot()
   let get_snd = function
	  Top -> Y.top()
	| Bot -> Y.bot()
	| First x -> Y.bot()
	| Second y -> y

   let to_string = function
	  Bot -> "Bot"
	| Top -> "Top"
	| First x -> X.to_string x
	| Second y -> Y.to_string y

   let to_int = function
	  Bot -> 19
	| Top -> 31
	| First x -> X.to_int x
	| Second y -> Y.to_int y

   let top () = Top
   let bot () = Bot

   let cup x y = match (x,y)
	with (Bot,_) -> y
	|    (_,Bot) -> x
	| (First x, First y) -> First (X.cup x y)
	| (Second x, Second y) -> Second (Y.cup x y)
	| _ -> Top 
   let cap x y = match (x,y)
        with (Top,_) -> y
        |    (_,Top) -> x
        | (First x, First y) -> let d = X.cap x y
			        in if X.leq d (X.bot()) then Bot
				   else First d
        | (Second x, Second y) -> let d = Y.cap x y
                                  in if Y.leq d (Y.bot()) then Bot
                                     else Second d
        | _ -> Bot

   let widen x y = match (x,y)
	with (Bot,_) -> y
        |    (_,Bot) -> x
        | (First x, First y) -> First (X.widen x y)
	| (Second x, Second y) -> Second (Y.widen x y)
        | _ -> Top
   let narrow x y = match (x,y)
        with (Top,_) -> y
        |    (_,Top) -> x
        | (First x, First y) -> let d = X.narrow x y
                                in if X.leq d (X.bot()) then Bot
                                   else First d
        | (Second x, Second y) -> let d = Y.narrow x y
                                  in if Y.leq d (Y.bot()) then Bot
                                     else Second d
        | _ -> Bot

   let leq x y = match (x,y)
        with (Bot,_) -> true
	|    (_,Top) -> true
	| (First x, First y) -> X.leq x y
	| (Second x, Second y) -> Y.leq x y
	| _ -> false

   let eq  x y = match (x,y)
        with (Bot,Bot) -> true
        |    (Top,Top) -> true
        | (First x, First y) -> X.eq x y
        | (Second x, Second y) -> Y.eq x y
	| _ -> false

end

type 'a flat = Bot | Top | Val of 'a

module Flat (X: Item) =
struct
   type d = X.item flat
   type item = d

   let to_string = function 
	Top -> "Top"
      | Bot -> "Bot"
      | Val a -> X.to_string a
   let to_int = function
	Top -> 3
      | Bot -> 7
      | Val a -> 5 * X.to_int a + 13

   let cup x y = match (x,y)
	with (Bot,_) -> y
	|    (_,Bot) -> x
	|(Val a,Val b) -> if a = b then Val a else Top
	| _          -> Top

   let cap x y = match (x,y)
        with (Top,_) -> y
        |    (_,Top) -> x
        |(Val a,Val b) -> if a=b then Val a else Bot
        | _          -> Bot

   let widen      = cup
   let narrow x y = x

   let eq x y = (x=y)
  
   let leq x y = match (x,y)
	with (Bot,_) -> true
	|    (_,Top) -> true
	|(Val a,Val b) -> if a=b then true else false
	| _          -> false

   let top() = Top
   let bot() = Bot

   let to_flat a = Val a
   let val_of = function 
	Val a -> a
	| _   -> raise (Failure "illegal argument of val_of")

end
   
module Power (X: Item) = 
struct
	module S = Pool(X)
   
   type d = S.set option
   type item = d

   let to_int = function 
	  None -> 31
	| Some x -> S.to_int x

   let to_string = function 
	  None -> "All"
	| Some x -> S.to_string x

   let bot() = Some (S.new_set())
   let top() = None

   let cup x y = match (x,y)
	with (Some x, Some y) -> Some (S.cup x y)
	| _	-> None

   let cap x y = match (x,y)
	with (Some x, Some y) -> Some (S.cap x y)
	| (None,_)	-> y
	| (_,None)	-> x

   let leq x y = match (x,y)
	with (_,None) -> true
	| (Some x, Some y) -> let l = S.to_list x
			 in let rec doit = function
				[] -> true
				| x::xs -> if S.has_item y x then doit xs
					   else false
			 in doit l
	| _ -> false

   let eq x y = leq x y && leq y x

   let widen x y = cup x y
   let narrow x y = x
	
   let from_list l = Some (S.from_list l)
   let from_set s = Some s

   let to_list = function
	  Some s -> S.to_list s
	| None   -> raise (Failure "illegal argument of to_list")

   let to_list_option = function
	  Some s -> Some (S.to_list s)
	| None   -> None
end

module NegPower (X: Item) = 
(* set enumerates all missing elements *)
struct
	module S = Pool(X)
   
   type d = S.set option
   type item = d

   let to_int = function 
	  None -> 31
	| Some x -> S.to_int x

   let to_string = function 
	  None -> "Empty"
	| Some x -> S.to_string x

   let top() = Some (S.new_set())
   let bot() = None

   let cup x y = match (x,y)
	with (Some x, Some y) -> Some (S.cap x y)
	| (None,_)	-> y
	| (_,None)	-> x

   let cap x y = match (x,y)
	with (Some x, Some y) -> Some (S.cup x y)
	| _	-> None

   let leq x y = match (x,y)
	with (None,_) -> true
	| (Some y, Some x) -> let l = S.to_list x
			 in let rec doit = function
				[] -> true
				| x::xs -> if S.has_item y x then doit xs
					   else false
			 in doit l
	| _ -> false

   let eq x y = leq x y && leq y x

   let widen x y = cup x y
   let narrow x y = x
	
   let from_list l = Some (S.from_list l)
   let from_set s = Some s

   let to_list_option = function
	  Some s -> Some (S.to_list s)
	| None   -> None

   let to_list = function
	  Some s -> S.to_list s
	| None   -> raise (Failure "illegal argument of to_list")
end

module Map (X : Item) (D : Lattice) =
(* Domain of partial maps. Missing entries are interpreted as top. *)
struct
	module M = Table (X)
	type d = D.d M.table option
	type item = d

   let to_string = function
	  None   -> "Bot"
	| Some x -> M.to_string D.to_string x

   let to_int = function
          None   -> 19
	| Some x -> M.to_int D.to_int x + 3

   let bot() = None
   let top() = Some (M.new_table ())

   let cup x y = match (x,y)
	with (None,_) -> y
	|    (_,None) -> x
	| (Some x,Some y) -> let y = M.copy_table y
	      in let list = M.to_list x
	      in let doit (x,d) = match M.sub y x
			with None -> ()
			| Some b -> let d = D.cup b d
				in if D.leq (D.top()) d then 
					M.rem_item y x
				   else M.update y x d
	      in let _ = List.iter doit list
	      in Some y

   let cap x y = match (x,y)
	with (Some x, Some y) -> let y = M.copy_table y
		in let list = M.to_list x
		in let doit (x,d) = match M.sub y x
			with None -> M.update y x d
			| Some b -> M.update y x (D.cap b d)
		in Some y
	| _ -> None

   let leq x y = match (x,y)
	with (None,_) -> true
	| (Some x, Some y) -> let lx = M.to_list x
			   in let rec doit = function
				  [] -> continue ()
				| (x,d)::rest -> (match M.sub y x
					with None -> doit rest
					| Some b -> if D.leq d b then doit rest
						    else false
					)
			      and continue () = 
			      let ly = M.to_list y
			   in let rec doit1 = function
				  [] -> true
				| (y,b)::rest -> (match M.sub x y
					with None -> false
					| _ -> doit1 rest
					)
			   in doit1 ly
			   in doit lx
	| _ -> false

   let eq x y = match (x,y)
        with (None,None) -> true
        | (Some x, Some y) -> let lx = M.to_list x
                           in let rec doit = function
                                  [] -> continue ()
                                | (x,d)::rest -> (match M.sub y x
                                        with None -> if D.leq (D.top()) d then doit rest
						     else false
                                        |  Some b -> if D.eq d b then doit rest
                                                     else false
                                        )
                              and continue () =
                              let ly = M.to_list y
                           in let rec doit1 = function
                                  [] -> true
                                | (y,b)::rest -> (match M.sub x y
                                        with None -> if D.leq (D.top()) b then doit1 rest
						     else false
                                        | _ -> doit1 rest
                                        )
                           in doit1 ly
                           in doit lx
        | _ -> false

   let widen x y = cup x y
   let narrow x y = x

   let from_list l = Some (M.from_list l)
   let from_table t = Some t

   let to_list = function
	  None -> raise (Failure "illegal arument of to_list")
	| Some m -> M.to_list m

end

module Map1 (X : Item) (D : Lattice) =
(* Domain of partial maps. Missing entries are interpreted as bot. *)
struct
	module M = Table (X)
	type d = D.d M.table option
	type item = d

   let to_string = function
	  None   -> "Top"
	| Some x -> M.to_string D.to_string x

   let to_int = function
          None   -> 19
	| Some x -> M.to_int D.to_int x + 3

   let top() = None
   let bot() = Some (M.new_table ())

   let cup x y = match (x,y)
	with (Some x,Some y) -> let y = M.copy_table y
	      in let list = M.to_list x
	      in let doit (x,d) = match M.sub y x
			with None -> if D.leq d (D.bot()) then ()
				     else M.update y x d
			| Some b -> M.update y x (D.cup b d)
	      in let _ = List.iter doit list
	      in Some y
	| _ -> None

   let cap x y = match (x,y)
	with (None,_) -> y
	|    (_,None) -> x
	| (Some x, Some y) -> let y = M.copy_table y
		in let list = M.to_list x
		in let doit (x,d) = match M.sub y x
			with None -> ()
			| Some b -> let d = D.cap b d
				in if D.leq d (D.bot()) then M.rem_item y x
				   else M.update y x d
		in Some y

   let leq x y = match (x,y)
	with (_,None) -> true
	| (Some x, Some y) -> let lx = M.to_list x
			   in let rec doit = function
				  [] -> true
				| (x,d)::rest -> (match M.sub y x
					with None -> if D.leq d (D.bot()) then doit rest
						     else false
					| Some b -> if D.leq d b then doit rest
						    else false
					)
			   in doit lx
	| _ -> false

   let eq x y = match (x,y)
        with (None,None) -> true
        | (Some x, Some y) -> let lx = M.to_list x
                           in let rec doit = function
                                  [] -> continue ()
                                | (x,d)::rest -> (match M.sub y x
                                        with None -> if D.leq d (D.bot()) then doit rest
						     else false
                                        |  Some b -> if D.eq d b then doit rest
                                                     else false
                                        )
                              and continue () =
                              let ly = M.to_list y
                           in let rec doit1 = function
                                  [] -> true
                                | (y,b)::rest -> (match M.sub x y
                                        with None -> if D.leq b (D.bot()) then doit1 rest
						     else false
                                        | _ -> doit1 rest
                                        )
                           in doit1 ly
                           in doit lx
        | _ -> false

   let widen x y = cup x y
   let narrow x y = x

   let from_list l = Some (M.from_list l)
   let from_table t = Some t

   let to_list = function
	  None -> raise (Failure "illegal arument of to_list")
	| Some m -> M.to_list m

end


