open Datatypes
open IO

type state = { 	mutable clauses : (int, int, int) clause list ;
		mutable usedir : string ;
		mutable filedir : string ;
		mutable outfile : out_channel ;
		mutable observables : int list 
	     }

let state : state = { 	clauses = [] ;
			usedir = "." ;
			filedir = "." ;
			outfile = stdout ;
			observables = []
			}

let help () = 	List.iter print_string [
				"\tAvailable Commands:\n\n" ;
				"\t\t#observe p/k :\t\tadds p/k to observed predicates.\n" ;
				"\t\t#all :\t\t\tresets list of observables to all.\n" ;
				"\t\t#solve :\t\tSolves bottom-up.\n" ;
				"\t\t#topdown :\t\tSolves topdown the observed preds.\n" ;
				"\t\t#file <out> :\t\tdirects derived facts to <filedir>/<out>.\n" ;
				"\t\t#filedir <dir> :\tsets file directory to <dir>.\n" ;
				"\t\t#use <in> :\t\tinterprets commands from <usedir>/<in>.\n" ;
				"\t\t#usedir <dir> :\t\tsets use directory to <dir>.\n" ;
				"\t\t#dir <dir> :\t\tsets use- and filedir to <dir>.\n" ;
				"\t\t#reset :\t\tresets the current clause set.\n" ;
				"\t\t#dump :\t\t\tdumps all currently proven facts.\n";
				"\t\t#exit :\t\t\texits the program.\n";
				"\t\t#help :\t\t\tprints this message.\n"
			]

let observe (p,k) = let r = P.index_of (p,k)
		    in state.observables <- r :: state.observables 

let set_dir string = (state.usedir <- string; state.filedir <- string)
let set_filedir string = state.filedir <- string
let set_usedir string = state.usedir <- string
let set_outfile string = try state.outfile <- open_out (state.filedir ^"/"^string)
			 with Sys_error _ -> print_string ("Output file not found!\n")

let get_input in_channel	= 
        try ParserProlog.prog LexerProlog.token (
            Lexing.from_channel in_channel)
         with Parsing.Parse_error ->
		(Parsing.parse_error "Syntax error"; [])
        |  Invalid_argument str ->
                (print_string ("Parsing error: " ^ str ^ "\n"); flush stdout; [])
        |  Failure str -> (print_string str; flush stdout; [])

let set_clauses c = let clauses = Horn.init_clauses (
				  translate_clauses c) 
		    
		    in  Horn.reset_called ();
			state.clauses <- clauses @ state.clauses 

let output_pred (r,tuples) = let do_tup t =
                                output_string state.outfile "\t";
                                output_string state.outfile (string_of_fact (r,t));
                                output_string state.outfile ".\n"
                        in List.iter do_tup tuples;
                           output_string state.outfile "\n"

let solve () = let result = match state.observables
	       with [] -> Horn.solve0 state.clauses
	       | list  -> Horn.filtered_solve0 list state.clauses
	       in List.iter output_pred result;
		  flush state.outfile 

let topdown () = match state.observables
		with [] -> solve ()
		| _ -> let result = Horn.topdown state.observables 
	               in (List.iter output_pred result;
                           flush state.outfile
				)

let print_env_list env_list = let print_env env = 	print_string "\t" ;
					print_string (string_of_env env) ;
					print_string "\n" 
		  in match env_list 
		     with [] -> print_string "\tno\n"; 	
				flush stdout
		     | _     -> List.iter print_env env_list; 
				flush stdout 

let answer body =    let body = translate_body body
		  in let env_list = Horn.answer body	
				(* the actual local solving takes place here! *)
		  in print_env_list env_list

let dump () = let result = match state.observables
			  			with [] -> Horn.dump()
		  				| list  -> Horn.filtered_dump list
	      in List.iter output_pred result;
                 flush state.outfile

let reset () = 	Horn.reset_all ();
		state.clauses <- [] ;
		state.observables <- []

exception Eof_reached

let rec use string =  try let in_channel = open_in (state.usedir ^"/"^string)
		  
		   in let rec doit () = 
			      execute (fun () -> ()) 
				      (get_input in_channel) 
		   in doit(); 
		      flush stdout; flush state.outfile;
                      close_in in_channel
		  with Sys_error _ -> print_string "Use file not found!\n"

and execute eof commands = let do_command = function
		Use string	-> use string
	|	Prog clauses	-> set_clauses clauses
	|	Question body	-> (try answer body
				   with Failure str -> print_string ("Failure: " ^ str ^ "\n")
				)
	|	Dir string	-> set_dir string
	|	Usedir string	-> set_usedir string
	|	Outdir string	-> set_filedir string
	|	Outfile string	-> set_outfile string
	|	Observe p       -> observe p
	|	All		-> state.observables <- []
	|	Help		-> help ()
	|	Topdown 	-> (try topdown ()
				   with Failure str -> print_string 
					("Failure: " ^ str ^ "\n")
					)
	|	Solve		-> (try solve ()
				   with Failure str -> print_string ("Failure: " ^ str ^ "\n")
				)
	|	Dump		-> dump ()
	|	Reset		-> reset ()
	|	Eof		-> eof ()

    in List.iter do_command commands

let use0 () =  let commands = get_input stdin
               in execute (fun () -> ()) commands
