open Items
open Table
open Pool
open Forest
open ForestMap
open Datatypes

module Env = Table (IntItem)
type 'b env = 'b Env.table

let new_env () = Env.make_table 4

let subst env = let rec doit = function
	Const b -> Some b
	| Var x -> Env.sub env x
	| Unknown -> None
    in List.map doit 

let rec known_part eta = function
	   [] -> []
	| Const a::rest -> a :: known_part eta rest
	| Unknown::rest -> []
	| Var x::rest -> (match Env.sub eta x
		with Some a -> a :: known_part eta rest
		|    None   -> [])

let unify env args t = let env = Env.copy_table env
   in let rec doit = function
        ([],_) -> Some env
      | (_,[]) -> Some env
      | (Const a::aa, b::bb) -> if a=b then doit (aa,bb)
		else None
      | (Unknown::aa, b::bb) -> doit (aa,bb)
      | (Var x::aa, b::bb) -> (match Env.sub env x 
		with None -> (Env.add env x b; doit (aa,bb))
		| Some a -> if a=b then doit (aa,bb)
			    else None
		)
   in doit (args,t)

let rec pat_of_args eta = function
	[] -> []
      | Const a :: aa -> Some a :: pat_of_args eta aa
      | Unknown :: aa -> None   :: pat_of_args eta aa
      | Var x :: aa   -> Env.sub eta x :: pat_of_args eta aa

let rec pref_of_pat = function
	[] -> []
      | Some b :: bb -> b :: pref_of_pat bb
      | None :: _    -> []

let unify_pat args pat = let env = Env.make_table 10
   in let rec doit = function
        ([],_) -> Some env
      | (_,[]) -> Some env
      | (Const a::aa, Some b::bb) -> if a=b then doit (aa,bb)
                else None
      | (_::aa, None::bb) -> doit (aa,bb)
      | (Unknown::aa, _::bb) -> doit (aa,bb)
      | (Var x::aa, Some b::bb) -> (match Env.sub env x
                with None -> (Env.add env x b; doit (aa,bb))
                | Some a -> if a=b then doit (aa,bb)
                            else None
                )
   in doit (args,pat)

let next x f = match x with None -> ()
		| Some eta -> f eta

(* The solver ... *)

let rec iter f = function
		[]	-> ()
	|  	[x]	-> f x
	| 	x::xs	-> f x; iter f xs

module Solve (X:Item)  = struct

	module S = Table (IntItem)
        module F = Forest (X)
	module M = ForestMap (X)

type cont = X.item list -> unit

let sigma : (F.forest S.table) = S.make_table 20
let infl : (cont M.forest_map S.table) = S.make_table 20

let to_list () = let list = S.to_list sigma
	      in let doit (x,f) = (x,F.to_list f)
	      in List.map doit list

(* auxiliary functions ... *)

let insert r tuple = match S.sub sigma r
	with Some f -> F.insert f tuple
	|    None   -> S.add sigma r (F.from_tuple tuple)

let has_tuple r tuple = match S.sub sigma r
	with Some f -> F.has_value f tuple
	|    None   -> false

let register r tuple f = match S.sub infl r
	with Some m -> M.insert m tuple f
	|    None   -> S.add infl r (M.from_tuple tuple f)

let consumers r tuple = match S.sub infl r
	with Some m -> M.consumers m tuple
	|    None   -> []

(* the succinct semi-naive solver ... *)

let getValue = function Some x -> x | 
		None -> raise (Failure "Uninstantiated variable?")

let rec check (r,arg,body) eta = match body
	with []	-> let s = List.map getValue (
				subst eta arg)
		in if has_tuple r s then ()
		   else (insert r s;
			 iter (fun f -> f s) (
			 consumers r s)
			)

	 | True :: rest -> check (r,arg,rest) eta

	 | Eq (Var x,Var y) :: rest  -> let eta = Env.copy_table eta
                           in (match (Env.sub eta x, Env.sub eta y)
                                with    (Some a, Some b) -> if a = b then
                                                check (r,arg,rest) eta
                                        else    ()
                                |       (Some a, None)  -> (Env.add eta y a;
                                                check (r,arg,rest) eta)
                                |       (None, Some b)  -> (Env.add eta x b;
                                                check (r,arg,rest) eta)
                                |       _ -> raise (Failure "Floundering equality!")
                                )

        | Eq (Var x, Const b) :: rest -> let eta = Env.copy_table eta
                             in (match Env.sub eta x
                                 with   Some a  -> if a = b then
                                                check (r,arg,rest) eta
                                                else    ()
                                |       None    -> (Env.add eta x b;
                                                check (r,arg,rest) eta)
                                )

        | Eq (Const b, Var x) :: rest -> let eta = Env.copy_table eta
                             in (match Env.sub eta x
                                 with   Some a  -> if a = b then
                                                check (r,arg,rest) eta
                                                else    ()
                                |       None    -> (Env.add eta x b;
                                                check (r,arg,rest) eta)
                                )

        | Eq (Const a, Const b) :: rest -> if a = b then check (r,arg,rest) eta
                                           else ()

        | Eq _ :: rest -> check (r,arg,rest) eta

	| Neq (a1,a2) :: rest -> (match (a1,a2)
		with    (Var x, Var y) -> (match (Env.sub eta x, Env.sub eta y)
				with (Some a, Some b) -> if a = b then ()
						else check (r,arg,rest) eta
				| _ -> raise (Failure "Floundering dis-equality!")
				)
		|    (Var x, Const b) -> (match Env.sub eta x
                                with    Some a  -> if a = b then ()
					else check (r,arg,rest) eta
                                |       None    -> 
					raise (Failure "Floundering dis-equality!")
				)
		|    (Const b, Var x) -> (match Env.sub eta x
                                with    Some a  -> if a = b then ()
					else check (r,arg,rest) eta
                                |       None    -> 
					raise (Failure "Floundering dis-equality!")
				)
		|    (Const a, Const b) -> if a = b then () 
					   else check (r,arg,rest) eta

		|    _ 	-> raise (Failure "Unknown variable in dis-equality!")
		)
			
	| Builtin f :: rest -> let eta = Env.copy_table eta
			    in let eval x = Env.sub eta x
			    in (match f eval
				with None -> ()
				| Some bindings -> (
                            let rec doit = function
                                        [] -> check (r,arg,rest) eta
                                | (x,b) :: rest -> (match Env.sub eta x
                                        with Some b' -> if b=b' then doit rest
                                                        else ()
                                        | None -> (Env.add eta x b; doit rest)
					)
                            in doit bindings
				)
			)
	| Query (r',arg') :: rest -> 
		   let f s = next (unify eta arg' s) (check (r,arg,rest))
		in let t = known_part eta arg'       
		in let available = match S.sub sigma r'
			with Some forest -> F.sub forest t
			|    None	 -> []
		in (register r' t f;
		    iter f available)

	| Neg (r',arg') :: rest -> 
		   let t = known_part eta arg'
		   in if List.length t < List.length arg' then
			raise (Failure "floundering query ... ")
		      else (match S.sub sigma r'
                        with Some forest -> if F.has_value forest t then ()
					    else check (r,arg,rest) eta
			|    None        -> check (r,arg,rest) eta
			)
	| Is _ :: rest -> raise (Failure "\"is\" not supported for plain Datalog!")
	| Comp _ :: rest -> raise (Failure "Int comparisons not supported for plain Datalog!")
	| IsInt _ :: rest -> raise (Failure "Int test not supported for plain Datalog!")
	| IsString _ :: rest -> raise (Failure "String test not supported for plain Datalog!")
	| NegInt _ :: rest -> raise (Failure "Int test not supported for plain Datalog!")
	| NegString _ :: rest -> raise (Failure "String test not supported for plain Datalog!")

let solve prog = let doit (Clause (r,arg,body)) = 
			check (r,arg,body) (new_env())
	      	 in iter doit prog

(* the local solver ... *)

	module C = Pool (PairItem (IntItem) (ListItem (OptionItem(X))))

let clauses : ((int,int,X.item) clause M.forest_map S.table) = S.make_table 20
let called : C.set = C.make_set 1000

let recorded_solve prog = let _ = solve prog
		       in let preds = List.map (fun (r,_) -> r) (
				      S.to_list sigma)
		       in let do_pred r = C.insert called (r,[])
		       in List.iter do_pred preds
				      
let rec is_called p pat = 
	     if C.has_item called (p,[]) then true
	     else if C.has_item called (p,pat) then true
	     else false

let init_clauses cc = let do_clause c = match c with Clause (r,arg,body) ->
	let t = known_part (new_env()) arg
	in (match S.sub clauses r
		with Some m -> M.insert m t c
		|    None   -> S.add clauses r (M.from_tuple t c)
	   )
   in iter do_clause cc

let rec check_local (r,arg,body) eta = match body
        with [] -> let s = List.map getValue (
                                subst eta arg)
                in if has_tuple r s then ()
                   else (insert r s; 
                         iter (fun f -> f s) ( consumers r s)
			)
	
         | True :: rest -> check_local (r,arg,rest) eta
 
	 | Eq (Var x,Var y) :: rest  -> let eta = Env.copy_table eta
                           in (match (Env.sub eta x, Env.sub eta y)
                                with    (Some a, Some b) -> if a = b then
                                                check_local (r,arg,rest) eta
                                        else    ()
                                |       (Some a, None)  -> (Env.add eta y a;
                                                check_local (r,arg,rest) eta)
                                |       (None, Some b)  -> (Env.add eta x b;
                                                check_local (r,arg,rest) eta)
                                |       _ -> raise (Failure "Floundering equality!")
                                )

        | Eq (Var x, Const b) :: rest -> let eta = Env.copy_table eta
                             in (match Env.sub eta x
                                 with   Some a  -> if a = b then
                                                check_local (r,arg,rest) eta
                                                else    ()
                                |       None    -> (Env.add eta x b;
                                                check_local (r,arg,rest) eta)
                                )

        | Eq (Const b, Var x) :: rest -> let eta = Env.copy_table eta
                             in (match Env.sub eta x
                                 with   Some a  -> if a = b then
                                                check_local (r,arg,rest) eta
                                                else    ()
                                |       None    -> (Env.add eta x b;
                                                check_local (r,arg,rest) eta)
                                )

        | Eq (Const a, Const b) :: rest -> if a = b then check_local (r,arg,rest) eta
                                           else ()

        | Eq _ :: rest -> check_local (r,arg,rest) eta

	| Neq (a1,a2) :: rest -> (match (a1,a2)
                with (_, Unknown) -> raise (Failure "Unknown variable in dis-equality!")
                |    (Unknown, _) -> raise (Failure "Unknown variable in dis-equality!")
                |    (Var x, Var y) -> (match (Env.sub eta x, Env.sub eta y)
                                with (Some a, Some b) -> if a = b then ()
                                                else check_local (r,arg,rest) eta
                                | _ -> raise (Failure "Floundering dis-equality!")
                                )
                |    (Var x, Const b) -> (match Env.sub eta x
                                with    Some a  -> if a = b then ()
                                        else check_local (r,arg,rest) eta
                                |       None    ->
                                        raise (Failure "Floundering dis-equality!")
                                )
                |    (Const b, Var x) -> (match Env.sub eta x
                                with    Some a  -> if a = b then ()
                                        else check_local (r,arg,rest) eta
                                |       None    ->
                                        raise (Failure "Floundering dis-equality!")
                                )
                |    (Const a, Const b) -> if a = b then ()
					   else check_local (r,arg,rest) eta
                )

	| Builtin f :: rest -> let eta = Env.copy_table eta
			    in let eval x = Env.sub eta x
			    in (match f eval
			       with None -> ()
				|   Some bindings ->
			    let rec doit = function 
					[] -> check_local (r,arg,rest) eta
				| (x,b) :: rest ->  (match Env.sub eta x
                                        with Some b' -> if b=b' then doit rest
						        else ()
                                        | None -> (Env.add eta x b; doit rest)
					)
			    in doit bindings
			)

        | Query (r',arg') :: rest ->
                   let f s = next (unify eta arg' s) (check_local (r,arg,rest))
                in let pat = pat_of_args eta arg'
                in let t = pref_of_pat pat
                in (register r' t f;
                    solve_local r' pat f)
 	| Neg (r',arg') :: rest -> 
		   let pat = pat_of_args eta arg'
                in let t = pref_of_pat pat
		in if List.length t < List.length arg' then 
			raise (Failure "floundering negation ...")
		   else let _ = solve_local r' pat (fun _ -> ())
                in (match S.sub sigma r'
                        with Some forest -> if F.has_value forest t
					    then ()
					    else check_local (r,arg,rest) eta
                        |    None        -> check_local (r,arg,rest) eta
			)
	| Is _ :: rest -> raise (Failure "\"is\" not supported for plain Datalog!")
	| Comp _ :: rest -> raise (Failure "Int comparisons not supported for plain Datalog!")
	| IsInt _ :: rest -> raise (Failure "Int test not supported for plain Datalog!")
	| IsString _ :: rest -> raise (Failure "String test not supported for plain Datalog!")
   	| NegInt _ :: rest -> raise (Failure "Int test not supported for plain Datalog!")
        | NegString _ :: rest -> raise (Failure "String test not supported for plain Datalog!")


and solve_local r pat f = let t = pref_of_pat pat
		in if is_called r pat then
		   let available = match S.sub sigma r
                        with Some forest -> F.sub forest t
                        |    None        -> []
		   in iter f available
	else let _ = C.insert called (r,pat)
	  in let candidates = match S.sub clauses r
			with Some m -> M.all_consumers m t
                        |    None        -> []
          in let do_clause (Clause (r,arg,body)) = 
			next (unify_pat arg pat)
			     (check_local (r,arg,body))
	  in let _ = iter do_clause candidates
	  in let available = match S.sub sigma r
                        with Some forest -> F.sub forest t
                        |    None        -> []
          in iter f available

	module E = Pool (ListItem (PairItem(IntItem)(X)))

let topdown r = let result = ref []
	     in let f s = (result := s :: !result)
	     in let _ = solve_local r [] f
	     in List.rev (!result)

let answer body = let answers = E.new_set ()
	in let rec answer0 body eta = match body
        with [] -> E.insert answers (Env.to_list eta)
	
    	| True :: rest -> answer0 rest eta

        | Eq (Var x,Var y) :: rest  -> let eta = Env.copy_table eta
                           in (match (Env.sub eta x, Env.sub eta y)
                                with    (Some a, Some b) -> if a = b then
                                                answer0 rest eta
                                        else    ()
                                |       (Some a, None)  -> (Env.add eta y a;
                                                answer0 rest eta)
                                |       (None, Some b)  -> (Env.add eta x b;
                                                answer0 rest eta)
                                |       _ -> raise (Failure "Floundering equality!")
                                )

        | Eq (Var x, Const b) :: rest -> let eta = Env.copy_table eta
                             in (match Env.sub eta x
                                 with   Some a  -> if a = b then
                                                answer0 rest eta
                                                else    ()
                                |       None    -> (Env.add eta x b;
                                                answer0 rest eta)
                                )

        | Eq (Const b, Var x) :: rest -> let eta = Env.copy_table eta
                             in (match Env.sub eta x
                                 with   Some a  -> if a = b then
                                                answer0 rest eta
                                                else    ()
                                |       None    -> (Env.add eta x b;
                                                answer0 rest eta)
                                )

	| Eq (Const a, Const b) :: rest -> if a = b then answer0 rest eta
					   else ()

        | Eq _ :: rest -> answer0 rest eta

	| Neq (a1,a2) :: rest -> (match (a1,a2)
                with (_, Unknown) -> raise (Failure "Unknown variable in dis-equality!")
                |    (Unknown, _) -> raise (Failure "Unknown variable in dis-equality!")
                |    (Var x, Var y) -> (match (Env.sub eta x, Env.sub eta y)
                                with (Some a, Some b) -> if a = b then ()
                                                else answer0 rest eta
                                | _ -> raise (Failure "Floundering dis-equality!")
                                )
                |    (Var x, Const b) -> (match Env.sub eta x
                                with    Some a  -> if a = b then ()
                                        else answer0 rest eta
                                |       None    ->
                                        raise (Failure "Floundering dis-equality!")
                                )
                |    (Const b, Var x) -> (match Env.sub eta x
                                with    Some a  -> if a = b then ()
                                        else answer0 rest eta
                                |       None    ->
                                        raise (Failure "Floundering dis-equality!")
                                )
                |    (Const a, Const b) -> if a = b then ()
					   else answer0 rest eta
                )

	|  Builtin f :: rest -> let eta = Env.copy_table eta
                            in let eval x = Env.sub eta x
                            in (match f eval
			       with None -> ()
				| Some bindings -> 
			    let rec doit = function
					 [] -> answer0 rest eta
                                | (x,b) :: rest ->  (match Env.sub eta x
                                        with Some b' -> if b=b' then doit rest
                                                        else ()
                                        | None -> Env.add eta x b; doit rest
					)
                            in doit bindings
			)
        | Query (r,arg) :: rest ->
                   let f s = next (unify eta arg s) (answer0 rest)
		in let pat = pat_of_args eta arg
                in let t = pref_of_pat pat
                in (register r t f;
                   solve_local r pat f)
        | Neg (r,arg) :: rest ->
		   let pat = pat_of_args eta arg
                in let t = pref_of_pat pat
                in if List.length t < List.length arg then
                        raise (Failure "floundering negation ...")
                   else let _ = solve_local r pat (fun _ -> ())
                in (match S.sub sigma r
                        with Some forest -> if F.has_value forest t
                                            then ()
                                            else answer0 rest eta
                        |    None        -> answer0 rest eta
                        )
	| Is _ :: rest -> raise (Failure "\"is\" not supported for plain Datalog!")
	| Comp _ :: rest -> raise (Failure "Int comparisons not supported for plain Datalog!")
	| IsInt _ :: rest -> raise (Failure "Int test not supported for plain Datalog!")
	| IsString _ :: rest -> raise (Failure "String test not supported for plain Datalog!")
   	| NegInt _ :: rest -> raise (Failure "Int test not supported for plain Datalog!")
        | NegString _ :: rest -> raise (Failure "String test not supported for plain Datalog!")


	in (answer0 body (new_env());
	    E.to_list answers)




let reset_called () = C.reset called 20;
		      S.reset_table sigma 20;
		      S.reset_table infl 20

let reset_all () = 	reset_called ();
			S.reset_table clauses 20
		
               
end
