{
open ParserProlog;;
open Lexing;;

let inc_line lexbuf = let pos = lexbuf.lex_curr_p
                 in let n = pos.pos_lnum
                 in lexbuf.lex_curr_p <- {pos_lnum = n+1;
                                           pos_fname = pos.pos_fname;
                                           pos_bol = pos.pos_cnum;
                                           pos_cnum = pos.pos_cnum
                                        }

let new_token lexbuf = lexbuf.lex_start_p <- lexbuf.lex_curr_p

let error lexbuf msg = let pos = lexbuf.lex_start_p
                       in let line = pos.pos_lnum
                       in let cnum = pos.pos_cnum
                       in let bol = pos.pos_bol
                       in raise (Failure
                                (String.concat "" [

                                       " line ";  string_of_int (line);
                                       ", char "; string_of_int (cnum-bol);
                                       " : "; msg; " ";
                                        Lexing.lexeme lexbuf; "\n"
                                        ]));;

let string_chars s = String.sub s 1 ((String.length s)-2);;
}
rule token =  parse 
  [' ' '\t']			{token lexbuf} 		    (* ignore whitespace *)
| '\n'                          { inc_line lexbuf; token lexbuf } (* ignore whitespace *)
| "/*"				{ comment lexbuf }
|  '('                          { new_token lexbuf; LPAR }
|  ')'                          { new_token lexbuf; RPAR }
|  '['				{ new_token lexbuf; BEGINLIST }
|  ']'				{ new_token lexbuf; ENDLIST }
|  "::"				{ new_token lexbuf; CONS }
|  "\\="			{ new_token lexbuf; NEQ }
|  '='				{ new_token lexbuf; EQ }
|  '<'				{ new_token lexbuf; ILT }
|  '>'				{ new_token lexbuf; IGT }
|  "=="				{ new_token lexbuf; IEQ }
|  "<>"				{ new_token lexbuf; INE }
|  ">="				{ new_token lexbuf; IGE }
|  "<="				{ new_token lexbuf; ILE }
|  ','                          { new_token lexbuf; SEP }
|  ';'                          { new_token lexbuf; LISTSEP }
|  '/'				{ new_token lexbuf; SLASH }
|  '.'                          { new_token lexbuf; DOT }
|  "is"                         { new_token lexbuf; IS }
|  '-'				{ new_token lexbuf; SUB }
|  '+'				{ new_token lexbuf; ADD }
|  '*'				{ new_token lexbuf; MUL }
|  '%'				{ new_token lexbuf; MOD }
|  ":-"                         { new_token lexbuf; IMPLIED }
|  "?-"                         { new_token lexbuf; QMARK }
|  ['0'-'9']+			{ new_token lexbuf; INT (int_of_string (Lexing.lexeme lexbuf)) }
|  "not"			{ new_token lexbuf; NOT }
|  "true"			{ new_token lexbuf; TRUE }
|  "int"			{ new_token lexbuf; ISINT }
|  "string"			{ new_token lexbuf; ISSTRING }
|  ['a'-'z''$'] ['_''0'-'9''A'-'Z''a'-'z''$']* 
				{ new_token lexbuf; NAME (Lexing.lexeme lexbuf) }
| '_'				{ new_token lexbuf; UNKNOWN }
|  ['_''A'-'Z']['_''0'-'9''A'-'Z''a'-'z''$']*  
				{ new_token lexbuf; VAR (Lexing.lexeme lexbuf) }
|  '\'' [^'\'']* '\''              { new_token lexbuf; NAME (Lexing.lexeme lexbuf) }
|  '"' [^'"']* '"'              { new_token lexbuf; STRING (string_chars (Lexing.lexeme lexbuf)) }
|  eof                          { new_token lexbuf; EOF }
|  "#observe"			{ new_token lexbuf; OBSERVE }
|  "#o"			{ new_token lexbuf; OBSERVE }
|  "#help"			{ new_token lexbuf; HELP }
|  "#h"			{ new_token lexbuf; HELP }
|  "#all"			{ new_token lexbuf; ALL }
|  "#a"			{ new_token lexbuf; ALL }
|  "#topdown"			{ new_token lexbuf; TOPDOWN }
|  "#t"			{ new_token lexbuf; TOPDOWN }
|  "#solve"			{ new_token lexbuf; SOLVE }
|  "#s"			{ new_token lexbuf; SOLVE }
|  "#use"			{ new_token lexbuf; USE }
|  "#u"			{ new_token lexbuf; USE }
|  "#reset"			{ new_token lexbuf; RESET }
|  "#r"			{ new_token lexbuf; RESET }
|  "#file"			{ new_token lexbuf; OUTFILE }
|  "#f"			{ new_token lexbuf; OUTFILE }
|  "#dir"			{ new_token lexbuf; DIR }
|  "#d"			{ new_token lexbuf; DIR }
|  "#usedir"			{ new_token lexbuf; USEDIR }
|  "#ud"			{ new_token lexbuf; USEDIR }
|  "#filedir"			{ new_token lexbuf; OUTDIR }
|  "#fd"			{ new_token lexbuf; OUTDIR }
|  "#dump"			{ new_token lexbuf; DUMP }
|  "#exit"			{ new_token lexbuf; EXIT }
|  "#e"			{ new_token lexbuf; EXIT }
| _				{ 
				 new_token lexbuf; 
				 error lexbuf "Lexing error: illegal token"
				}

and comment = parse
   ( [^'*' '\n']+ | '*'+ [^'/' '*' '\n'] )* '*'+ '/' { token lexbuf }
 | ( [^'*' '\n']+ | '*'+ [^'/' '*' '\n'] )* '*'* '\n' {inc_line lexbuf; comment lexbuf }
 | ( [^'*' '\n']+ | '*'+ [^'/' '*' '\n'] )* '*'* eof
                        { error lexbuf "Unterminated comment" }
 | _                    { error lexbuf "Illegal character!" }




