open Items
open Table
open Pool
open Datatypes
open Datalog
open Tree

	module S = Solve (IntItem)
	module T = Tree
	module O = Pool(IntItem)

	type annotation = Ground | Instantiated | Nonground 
	let max x y = if x < y then y else x

	let count = ref (-1)
	let reset_vars () = count := (-1)
	let fresh_var () = let n = !count
			   in count := n-1;
			      n

	let rec eval exp eta = match exp
		with EInt n	->  n
		|    EVar x	-> (match eta x
			with Some n	->	(match T.recover n
				with Int n -> n
				|    _	-> raise (Failure "non-int in arithmetic expression!")
				)	
			|    None	->	raise (Failure "Floundering is?")
			)
		|    Add (e1,e2)	-> eval e1 eta + eval e2 eta
		|    Sub (e1,e2)	-> eval e1 eta - eval e2 eta
		|    Mul (e1,e2)	-> eval e1 eta * eval e2 eta
		|    Div (e1,e2)	-> eval e1 eta / eval e2 eta
		|    Mod (e1,e2)	-> eval e1 eta mod eval e2 eta
		|    Minus e		-> - (eval e eta)

	let check_is (x,exp) =  let assign eta = 
					let j = eval exp eta
					in let j = T.get_int (Int j)
					in match eta x
					with Some i -> if i=j then Some []
							      else None
					|    None -> Some [(x,j)]
				in Builtin assign
		
	let check_int v = let doit eta = match eta v
			with Some i -> (match T.recover i
				with Int _ -> Some []
				|	_  -> None
				)
			|  None -> raise (Failure "Floundering int test!")
			in Builtin doit

	let check_string v = let doit eta = match eta v
			with Some i -> (match T.recover i
				with String _ -> Some []
				|	_  -> None
				)
			|  None -> raise (Failure "Floundering string test!")
			in Builtin doit

	let check_negint v = let doit eta = match eta v
                        with Some i -> (match T.recover i
                                with Int _ -> None
                                |       _  -> Some []
                                )
                        |  None -> raise (Failure "Floundering neg int test!")
                        in Builtin doit

        let check_negstring v = let doit eta = match eta v
                        with Some i -> (match T.recover i
                                with String _ -> None
                                |       _  -> Some []
                                )
                        |  None -> raise (Failure "Floundering neg string test!")
                        in Builtin doit

	let check_comp (c,e1,e2) = 
		let comp eta = let j1 = eval e1 eta
			    in let j2 = eval e2 eta
			    in match c
				with Lt -> if j1 < j2 then Some []
					   else None
				|    Gt -> if j1 > j2 then Some []
					   else None
				|    Le -> if j1 <= j2 then Some []
					   else None
				|    Ge -> if j1 >= j2 then Some []
					   else None
				|    E -> if j1 = j2 then Some []
					   else None
				|    Ne -> if j1 <> j2 then Some []
					   else None
		in Builtin comp	

	let is_ground t  = 
		let rec do_tree = function
			Int _ -> true
		|	TUnknown -> false
		|	String _ -> true
		|	Atom _ -> true
		|	TVar x -> false
		|	Cons (_,list) -> let rec do_list = function
					[]	-> true
					| t::ts	-> if (do_tree t) then (do_list ts)
						   else false
				in do_list list
		in do_tree t

	 let translate_tree t = let rec do_tree = function
                        Int n -> Int n
                |       TUnknown -> TVar (fresh_var())
                |       String s -> String s
                |       Atom a -> Atom a
                |       TVar x -> TVar x
                |       Cons (c,list) -> let rec do_list = function
                                        []      -> []
                                        | t::ts -> (do_tree t) :: (do_list ts)
                                in Cons (c,do_list list)
                in do_tree t

         let check_query (r,arg) = let rec doit = function
                        [] -> ([],[],[])
                |  Const t :: aa -> if is_ground t then
                                         	   let i = T.get_int t
                                                in let (left,aa,right) = doit aa
                                                in (left,Const i :: aa,right)
				 else  let x = fresh_var ()
				       in let t = translate_tree t
                                       in let b0 = Builtin (fun eta -> try
							Some [(x, T.construct t eta)]
							with _ -> Some []
							)
				       in let b1 = Builtin (fun eta -> (match eta x
                                                        with Some i -> T.destruct t i
                                                        |    None   -> raise (Failure 
								"Floundering destruct?")
							))
				       in let (left, aa, right) = doit aa
                                       in (b0::left, Var x :: aa, b1::right)
		| Var x :: aa -> let (left, aa, right) = doit aa
				 in  (left,Var x :: aa, right)
		| Unknown :: aa -> let (left, aa, right) = doit aa
                                 in  (left, Unknown :: aa, right)

		  in let (left,arg,right) = doit arg
                  in left @ Query (r,arg) :: right


	 let check_eq (a1,a2) = match (a1,a2)
				with (Var x, Const t) -> if is_ground t then
						   Eq (Var x, Const (T.get_int t))
				   	else let t = translate_tree t
					  in let test eta = match eta x
						with Some i -> T.destruct t i
						|    None   -> Some [(x, T.construct t eta)]
			      		  in Builtin test
					
				|    (Const t, Var x) -> if is_ground t then
                                                   Eq (Var x, Const (T.get_int t))
                                    	else let t = translate_tree t
                                          in let test eta = match eta x
                                                with Some i -> T.destruct t i
                                                |    None   -> Some [(x, T.construct t eta)]
                                          in Builtin test

				|    (Var x, Var y) -> Eq (Var x, Var y)

				| (Const t1, Const t2) -> let t1 = translate_tree t1
						       in let t2 = translate_tree t2
					in (match (is_ground t1, is_ground t2)
					with (true, true) -> 
						Eq (Const (T.get_int t1), Const (T.get_int t2))
					| (true,_) -> let i = T.get_int t1 
						     in	let test _ = T.destruct t2 i
                                                     in Builtin test
					| (_,true) -> let i = T.get_int t2 
						     in	let test _ = T.destruct t1 i
                                                     in Builtin test
					| _ -> let test eta = try
							let i = T.construct t1 eta
					        	in T.destruct t2 i 
						    with _ -> (try 
							let i = T.construct t2 eta
                                                	in T.destruct t1 i
						    with _ -> raise (Failure "Floundering equality?"))
					       in Builtin test
					)
					
				| _ -> True

	 let check_neq (a1,a2) = match (a1,a2)
		with (Var x, Var y) -> Neq (Var x, Var y)

		|    (Var x, Const t) -> if is_ground t then
                           		Neq (Var x, Const (T.get_int t))
                                 else let t = translate_tree t
				   in let test eta = let j = T.construct t eta
                                		     in match eta x
                                        	with Some i ->  if i = j then None
                                                		else Some []
                                        	| None  -> raise (Failure "Floundering dis-equality!")
                                   in Builtin test

		|    (Const t, Var x) -> if is_ground t then
			   		Neq (Var x, Const (T.get_int t))
			   	 else let t = translate_tree t
				   in let test eta = let j = T.construct t eta
						     in match eta x
						with Some i ->  if i = j then None
							else Some []
						| None	-> raise (Failure "Floundering dis-equality!")
				   in Builtin test
			
		|    (Const t1, Const t2) -> let t1 = translate_tree t1
					  in let t2 = translate_tree t2
				in (match (is_ground t1, is_ground t2)
				with (true, true) -> Neq (Const (T.get_int t1), Const (T.get_int t2))
				| _ -> 	let test eta = 
					let j1 = T.construct t1 eta
				    in 	let j2 = T.construct t2 eta
				    in 	if j1 = j2 then None
					else Some []
				    in Builtin test
			)
				
		|  _ -> raise (Failure "Unknown variable in dis-equality!")
	 
	 let check_neg (r,arg) = let rec doit = function
                        [] -> ([],[])
                |  Const t :: aa -> if is_ground t then
                                       	   let i = T.get_int t
                                        in let (left,aa) = doit aa
                                        in (left,Const i :: aa)
                                    else let t = translate_tree t
				      in let x = fresh_var ()
                                      in let b = Builtin (fun eta -> Some [(x, T.construct t eta)])
                                      in let (left,aa) = doit aa
                                      in (b::left, Var x :: aa)
                                       
		| Unknown :: aa -> let (left,aa) = doit aa
                                   in (left, Unknown :: aa)
                | Var x :: aa ->   let (left,aa) = doit aa
                                   in (left, Var x :: aa)

                        in let (left,arg) = doit arg
                        in left @ [Neg (r,arg)]

	let check_body qlist = let rec doit qlist = match qlist
			with	[] -> []
			| q :: qq  -> (match q
				with Query (r,arg) -> (check_query (r,arg)) @ doit qq
				|  Is (x,exp)	->    (check_is (x,exp)) :: doit qq
				|  Comp (c,e1,e2) ->  (check_comp (c,e1,e2)) :: doit qq
				|  IsInt (x) ->       (check_int x) :: doit qq
				|  IsString (x) ->    (check_string x) :: doit qq
				|  NegInt (x) ->      (check_negint x) :: doit qq
				|  NegString (x) ->   (check_negstring x) :: doit qq
				|  True		-> 	doit qq
				|  Eq (a1,a2)   ->    (check_eq (a1,a2)) :: doit qq
				|  Neq (a1,a2)   ->   (check_neq (a1,a2)) :: doit qq
				|  Neg (r,arg) ->     (check_neg (r,arg)) @ doit qq
				|   _	-> raise (Failure "Builtin in Horn clause?")
				)
		in doit qlist

	let check_head (r,arg) = let rec doit = function
                        [] -> ([],[],[])
                |  Const t :: aa -> if is_ground t then let i = T.get_int t
                                                in let (left,aa,right) = doit aa
                                                in (left, Const i :: aa, right)
                                    else let t = translate_tree t
				      in let x = fresh_var ()
                                      in let b0 = Builtin (fun eta -> match eta x
							with Some i -> T.destruct t i
							| None -> Some []
							)
                                      in let b = Builtin (fun eta -> 
						Some [(x, T.construct t eta)]
						)
                                      in let (left,aa,right) = doit aa
                                             in (b0::left, Var x :: aa, b::right)
		| Unknown :: aa -> let (left,aa,right) = doit aa
                                   in (left, Var (fresh_var ()) :: aa, right)
		|  Var x :: aa -> let (left,aa,right) = doit aa
				  in (left, Var x :: aa, right)

            in doit arg

	let translate_clause (Clause (r,arg,body)) = let _ = reset_vars ()
					in let body = check_body body
					in let (left,arg,right) = check_head (r,arg) 
					in Clause (r,arg, left @ body @ right)

	let translate clauses = List.map translate_clause clauses

	let solve0 clauses = let _ = S.recorded_solve clauses
			 in let result = S.to_list ()
			 in let do_tuple tuple = List.map T.recover tuple
			 in let do_pred (r,tuples) = (r, List.map do_tuple tuples)
			 in List.map do_pred result

        let filtered_solve0  observables clauses =
			let p (pred,tuples) = List.mem pred observables
			in let _ = S.recorded_solve clauses
                        in let result = List.filter p (S.to_list ())
                        in let do_tuple tuple = List.map T.recover tuple
                        in let do_pred (r,tuples) = (r, List.map do_tuple tuples)
			in List.map do_pred result

	let solve clauses = solve0 (translate clauses)
	let filtered_solve list clauses = filtered_solve0 list (translate clauses)

	let topdown r = let result = S.topdown r 
			in let do_tuple tuple = List.map T.recover tuple
			in List.map do_tuple result

        let topdown list = let do_pred r = (r, topdown r)
			in List.map do_pred list

	let init_clauses clauses = let clauses = translate clauses
			   	   in S.init_clauses clauses;
				      clauses

	let answer body = let _ = reset_vars ()
		       in let body = check_body body
		       in let answers = S.answer body
		       in let do_env list = List.map (
				fun (x,i) -> (x, T.recover i)) list
		       in List.map do_env answers 

	let dump () = let result = S.to_list ()
                   in let do_tuple tuple = List.map T.recover tuple
                   in let do_pred (r,tuples) = (r, List.map do_tuple tuples)
                   in List.map do_pred result

    let filtered_dump observables = 
		    let p (pred,tuples) = List.mem pred observables
		in  let result = List.filter p (S.to_list ())
		in  let do_tuple tuple = List.map T.recover tuple
        in  let do_pred (r,tuples) = (r, List.map do_tuple tuples)
		in  List.map do_pred result


	let reset_all () = S.reset_all ()
	let reset_called () = S.reset_called ()
