open Items
open Table
open Symbol
open Datatypes
open Tree

	module S = SymbolTable (StringItem)
	module P = SymbolTable (PairItem (StringItem) (IntItem))

(* Input related functions ... *)

let rec translate_exp = function
		EInt x	-> EInt x
	|	EVar x	-> EVar (S.index_of x)
	|	Add (e1,e2) -> Add (translate_exp e1, translate_exp e2)
	|	Sub (e1,e2) -> Sub (translate_exp e1, translate_exp e2)
	|	Mul (e1,e2) -> Mul (translate_exp e1, translate_exp e2)
	|	Div (e1,e2) -> Div (translate_exp e1, translate_exp e2)
	|	Mod (e1,e2) -> Mod (translate_exp e1, translate_exp e2)
	|	Minus e     -> Minus (translate_exp e)

let rec translate_tree = function
		Int x 	-> Int x
	|	String str -> String (S.index_of ("\"" ^ str ^ "\""))
	|	TUnknown -> TUnknown
	|	TVar x 	-> TVar (S.index_of x)
	|	Atom x	-> Atom (S.index_of x)
	|	Cons (a,list) -> Cons (P.index_of (a, List.length list), List.map translate_tree list)

let translate_arg = function
		Const t	-> Const (translate_tree t)
	|	Unknown -> Unknown
	|	Var x	-> Var (S.index_of x)

let translate_args list = List.map translate_arg list

let translate_query = function
		Query (r,arg)	-> Query (P.index_of (r, List.length arg), translate_args arg)
	|	True		-> True
	|	Eq (a1,a2)	-> Eq (translate_arg a1, translate_arg a2)
	|	Neq (a1,a2)	-> Neq (translate_arg a1, translate_arg a2)
	|	Is (x,e)	-> Is (S.index_of x, translate_exp e)
	|	IsInt x		-> IsInt (S.index_of x)
	|	IsString x	-> IsString (S.index_of x)
	|	NegInt x		-> NegInt (S.index_of x)
	|	NegString x	-> NegString (S.index_of x)
	|	Comp (c,e1,e2)	-> Comp (c, translate_exp e1, translate_exp e2)
	|	Neg (r,arg)	-> Neg (P.index_of (r, List.length arg), translate_args arg)
	|	Builtin	_	-> raise (Failure "Builtins so far not supported!")

let translate_body body = List.map translate_query body

let translate_clause (Clause (r,arg,body))	= 
		Clause (P.index_of (r, List.length arg), 
			translate_args arg, 
			translate_body body)

let translate_clauses clauses = List.map translate_clause clauses

(*
let file_input_prog file = let file = open_in file
			in
	try ParserProlog.prog LexerProlog.token (
	    Lexing.from_channel file)
         with Parsing.Parse_error ->
      		print_string (parsing_error "syntax error?\n");
		[]
*)

(* Output related functions ... *)

let string_of_pred r = match P.symbol_of r
		with (name,_) -> name
		
let string_of_cons r = match P.symbol_of r
		with (name,_) -> name
		
let string_of_atom a = S.symbol_of a
let string_of_string str = S.symbol_of str

let string_of_var x = if x < 0 then "_" ^ string_of_int x
		      else S.symbol_of x

let string_of_tree t = let rec do_tree t acc = match t
	with	Int int -> string_of_int int :: acc
	|	TUnknown -> "_" :: acc
	|	TVar x	-> string_of_var x :: acc
	|	Atom a	-> string_of_atom a :: acc
	|	String s -> string_of_string s :: acc
	|	Cons (b,list) -> let rec do_list list acc = match list
			with []	-> ")" :: acc
			| h::t	-> ", " :: do_tree h (do_list t acc)

		in (match list 
			with []	->	string_of_cons b :: " ()" :: acc
			| h::t	->	string_of_cons b :: " (" :: do_tree h (do_list t acc)
			)
	in String.concat "" (do_tree t [])

let string_of_fact (r,tup) = string_of_tree (Cons (r,tup))

let string_of_env eta = let do_one (x,t) = if x < 0 then []
					   else [
					   string_of_var x;
                                           " = ";
                                           string_of_tree t
					]
 		     in let rec do_rest = function
			[]	->	[]
		|   (x,t) :: r	-> 	"\t" ::
					do_one (x,t) @
					do_rest r
		in match eta
			with	[]	->	"yes"
			| (x,t)::r	-> String.concat "" (do_one (x,t) @ do_rest r)

					



