open RightHandSide
open Printf

module H = Hashtbl


let next_states action = 
   let set = H.create 101 in
   let rec next =  function
      Empty_rhs -> () |
      Parameter i -> () |
      Symbol (_,_,rhs_list) -> List.iter next rhs_list | 
      Call (q,_,_,rhs_list) -> 
         if (H.mem set q) then
            List.iter next rhs_list
         else
            H.add set q true; List.iter next rhs_list in
   let next_st = function
      Mtt.Undefined -> () |
      Mtt.Det rhs -> next rhs |
      Mtt.Non_det rhs_list -> List.iter next rhs_list 
   in
      next_st action;
      let res = ref [] in
         H.iter (fun x y -> res := x :: (!res)) set; (!res)



let to_graph mtt =
   let g = ref (Graph.empty ()) in
      Mtt.iter
         (fun (state,sym,action) ->
            let nx = next_states action in
            g := Graph.add_adjacency_list (!g) state (next_states action)
         ) mtt;     
      (!g)



let used_states mtt =
   let dep_graph = to_graph mtt in
   List.fold_left (fun result init ->
      (Graph.reachable dep_graph init) @ result
   ) [] (Mtt.get_init mtt)



let used_parameters mtt =
   let dep_graph = to_graph mtt in
   let comp_graph = Graph.component_graph dep_graph in
   let top_sort = List.rev (Graph.topological_sort comp_graph) in
   let color = H.create (List.length top_sort) in
   let prop = H.create 101 in
   let par = H.create 101 in
   let parOfState = H.create 101 in
   let rec init = function
      [] -> () |
      u::us -> (H.add color u (-1); init us) in
   let rec pars_in_rhs = function
      Empty_rhs -> () |
      Parameter i -> H.add par i true |
      Symbol (s,k,rhs_list) -> List.iter pars_in_rhs rhs_list |
      Call (q,x,k,rhs_list) -> 
         let js = try H.find parOfState q with Not_found -> [] in
            List.iter pars_in_rhs (Util.get_jth_of_list rhs_list js) in
   let rec pars_in_action = function
      Mtt.Undefined -> () |
      Mtt.Det ac -> pars_in_rhs ac |
      Mtt.Non_det ac_list -> List.iter pars_in_rhs ac_list in
   let pars_of_state mtt q =
      List.iter pars_in_action (Mtt.get_all_actions mtt q) in
   let rec propagate = function
      [] -> () |
      u::q ->
         if (H.find color u) >= 0 then
            propagate q
         else
            let adj =
               try Graph.adjacent_vertices comp_graph u
               with Not_found -> [] in
            let p' =
               H.clear par;
               H.iter (fun x y -> H.add par x true) prop;
               List.iter (pars_of_state mtt) u;
               let res = ref [] in
                  H.iter (fun x y -> res := x :: (!res)) par; (!res)
            in
               H.replace color u 1;
               List.iter (fun x -> H.add parOfState x p') u;
               List.iter (fun x -> H.add prop x true) p';
               propagate (adj @ q) in
   let rec do_propagate = function
      [] -> () |
      u::us -> 
         if (H.find color u) < 0 then
            (propagate [u]; do_propagate us)
         else 
            do_propagate us in
   let result = ref []
   in
      init top_sort;
      H.clear prop;
      do_propagate top_sort;
      H.iter (fun x y ->
          Printf.printf "%i: %s\n" x 
          (List.fold_left (fun r z -> (string_of_int z) ^","^ r) "" y )
      ) parOfState;
      parOfState


module Item = Items.IntItem
module D = Lattice.Power (Item)
module Solve_c = SolverCon.SolverCon (Item) (D)

let useful_parameters mtt = 
   let start = Mtt.get_initial_state mtt

   in let rec constr_of_rhs rhs sigma =
      match rhs with
         | Empty_rhs -> D.bot ()
         | Parameter i -> D.from_list (i :: [])
         | Symbol (s,n,rhs_lst) -> List.fold_left (fun r rhs ->
                                      D.cup r (constr_of_rhs rhs sigma)
                                   ) (D.bot ()) rhs_lst
         | Call (q,x,k,rhs_lst) -> let pars' = sigma q
                                   in let pars = match pars' with
                                                    | None -> []
                                                    | Some s -> D.S.to_list s
                                   in D.bot ()

   in let constr q = 
      let actions = Mtt.get_all_actions mtt q
      in let rhs_lst = List.fold_left (fun r action ->
                          match action with
                             | Mtt.Undefined -> r
                             | Mtt.Det rhs -> rhs :: r
                             | Mtt.Non_det lst -> lst @ r
                       ) [] actions
      in List.map constr_of_rhs rhs_lst

   in let result = List.map (fun (x,d) ->
                      match d with
                         | None -> (x,[])
                         | Some s -> (x,D.S.to_list s)
                   ) (Solve_c.T.to_list (Solve_c.solve constr start))
   (*in let result = []*)

   in let rec lst_to_str lst =
      match lst with [] -> "[]" |
                     x::xs -> (string_of_int x) ^","^ lst_to_str xs

   in printf "%s\n" (
      List.fold_left (fun r (x,l) ->
         r ^ 
         (Mtt_to_string.state_to_string mtt x)(*Item.to_string x*) ^": "^ (lst_to_str l) ^"\n"
      ) "" result 
      );
      result



let is_non_deterministic mtt =
   let test_ac  = function Mtt.Non_det _ -> true 
                         | _             -> false in
   let test_sym = function 0             -> true
                         | i when i<0    -> true
                         | _             -> false in
   let b = ref false in
      Mtt.iter (fun (_,sym,action) -> 
         b := !b || test_ac action || test_sym sym
      ) mtt;
      !b



let is_linear mtt =
   let rec doit a = function
      | Empty_rhs -> false
      | Parameter i -> false
      | Symbol (s,k,rhs_lst) -> doit' a rhs_lst
      | Call (q,x,k,rhs_lst) -> if a.(x) then true
                                else 
                                   let _ = a.(x) <- true 
                                   in doit' a rhs_lst
   and doit' a = function
      | [] -> false
      | x :: xs -> if doit a x then true
                   else doit' a xs
   in let check rank = function
      | Mtt.Undefined -> true
      | Mtt.Det rhs -> doit (Array.make (rank+1) false) rhs
      | Mtt.Non_det rhs_lst -> List.fold_left (fun r rhs ->
                                  doit (Array.make (rank+1) false) rhs || r
                               ) false rhs_lst
   in let b = ref false 
   in Mtt.iter (fun (_,symbol,action) ->
         let rank = Mtt.arity mtt symbol
         in b := !b || (check rank action)
      ) mtt;
      (not !b)



let is_weakly_single_use mtt =
   let wsu = H.create 101 in
   let rec doit' symbol = function
      Empty_rhs -> false |
      Parameter i -> false |
      Symbol (s,k,rhs_list) ->
         List.fold_right (fun x y -> doit' symbol x || y) rhs_list false |
      Call (q,s,k,rhs_list) ->
         if (H.mem wsu (symbol,q,s)) then
            true
         else
            (H.add wsu (symbol,q,s) true;
            List.fold_right (fun x y -> doit' symbol x || y) rhs_list false) in
   let doit symbol = function
      Mtt.Undefined -> false |
      Mtt.Det rhs -> doit' symbol rhs |
      Mtt.Non_det rhs_list ->
         List.fold_right (fun x y -> doit' symbol x || y) rhs_list false in
   let b = ref false in
      Mtt.iter (fun (_,symbol,action) -> b := !b || (doit symbol action) ) mtt;
      (not !b)



let is_copying mtt =
   let nc = H.create 101 in
   let rec doit' = function
      Empty_rhs -> false |
      Parameter i ->
         if (H.mem nc i) then
            true
         else
            (H.add nc i true; false) |
      Symbol (s,k,rhs_list) ->
         List.fold_right (fun x y -> doit' x || y) rhs_list false |
      Call (q,s,k,rhs_list) ->
         List.fold_right (fun x y -> doit' x || y) rhs_list false in
   let doit = function
      Mtt.Undefined -> false |
      Mtt.Det rhs -> doit' rhs |
      Mtt.Non_det rhs_list -> 
         List.fold_right (fun x y -> doit' x || y) rhs_list false in
   let b = ref false in
      Mtt.iter (fun (_,_,action) -> (H.clear nc;b:=!b || doit action)) mtt;
      !b
let is_non_copying mtt = not (is_copying mtt)


