(** This module implements an approximative type checking algorithm
    by means of a translation of the mtt into a datalog program.

    The module uses the Succinct Solver implementation by H.Seidl. *)

open RightHandSide
open Commands

module D = Datatypes
module B = Bottom_up_automaton

(* for type compability with the input types of function set_clauses
   of module Commands *)
module String_item = Items.StringItem
open String_item

exception Variable_error
exception Non_deterministic_output_type

let count = ref (1)
let reset_vars () = count := (1)
let fresh_var () = 
   let n = !count in 
   count := n+1; D.Var ("X" ^ string_of_int n)
let fresh_tvar () = 
   let n = !count in 
   count := n+1; D.TVar ("X" ^ string_of_int n)

let fresh_var_list n =
   let rec doit lst = function
      | 0 -> lst
      | i -> doit (fresh_var () :: lst) (i-1)
   in
      List.rev (doit [] n)
let fresh_tvar_list n =
   let rec doit lst = function
      | 0 -> lst
      | i -> doit (fresh_tvar () :: lst) (i-1)
   in
      List.rev (doit [] n)


let rec clauses_to_string = function
   | [] -> ""
   | cl :: clauses -> (clause_to_string cl) ^"\n"^ (clauses_to_string clauses)

and clause_to_string = function
   | D.Clause (name,args,body) ->
        name ^ 
        (let str = args_to_string args in
         if not (str = "") then "("^ str ^")" else "") ^ 
        (let str = body_to_string body in
         if String.length str > 1 then " :-\n"^ str else str)
(*    | _ -> "..." *)

and args_to_string = function
   | [] -> ""
   | a :: [] -> arg_to_string a
   | a :: args -> (arg_to_string a) ^", "^ (args_to_string args) 

and arg_to_string = function
   | D.Const v -> (const_to_string v)
   | D.Var x -> x
   | _ -> "..."

and const_to_string = function
   | D.Int i -> "Int("^ (string_of_int i) ^")"
   | D.Atom a -> a
   | _ -> "..."

and body_to_string = function
   | [] -> "."
   | q::[] -> "\t"^ (query_to_string q) ^" ."
   | q::qs -> "\t"^ (query_to_string q) ^"\n"^ (body_to_string qs)

and query_to_string = function
   | D.Eq (v1,v2) -> (arg_to_string v1) ^"="^ (arg_to_string v2)
   | D.Neq (v1,v2) -> (arg_to_string v1) ^"!="^ (arg_to_string v2)
   | D.Query (name,args) -> "Query( "^ name ^", "^ (args_to_string args) ^" )"
   | _ -> "..."


let delta = ref "delta"

let to_facts mtt (*f*) dfa =
   let f = Mtt_to_string.delta_to_string mtt in
   let g = Mtt_to_string.sigma_to_string mtt in
   let facts = ref [] in
   B.iter (fun (s,args,res) ->
      if List.length res > 1 then
         raise Non_deterministic_output_type
      else
         let s' = try f s with Not_found -> "UNDEF" in
         let arg_list = 
            D.Const (D.Int (List.hd res)) :: 
            D.Const (D.Atom (s')) :: 
            List.map (fun a -> D.Const (D.Int a)) args in
         let fact = D.Clause (!delta,arg_list,[]) in
         facts := fact :: !facts 
   ) dfa; !facts


let rec make_body f g var par_vars = function
   Empty_rhs -> [] |
      (* The datalog representation of variables can be more efficient:
         the current version [D.Eq (List.nth par_vars (i-1), var)] takes
         a new variable and checks whether it is equal to the one in
         the clause head. More efficient might be to directly use the
         variable of the clause head.  *)
   Parameter i -> [D.Eq (List.nth par_vars (i-1), var)]  |
   Symbol (s,k,args) -> 
      let new_varlst = fresh_var_list k in
      let sym_s = D.Const (D.Atom (f s)) in
      let first_query = D.Query (!delta, var :: sym_s :: new_varlst) in
      let j = ref 0 in
      let queries =
         List.fold_left (fun lst rhs ->
            let current_var = List.nth new_varlst !j in
            j := !j + 1;
            lst @ (make_body f g current_var par_vars rhs)
         ) [] args
      in
      first_query :: queries |
   Call (q,s,k,pars) -> 
      let new_varlst = fresh_var_list k in
      let sym_q = g q in
      let first_query = D.Query (sym_q, var :: new_varlst) in
      let j = ref 0 in
      let queries =
         List.fold_left (fun lst rhs ->
            let current_var = List.nth new_varlst !j in
            j := !j + 1;
            lst @ (make_body f g current_var par_vars rhs)
         ) [] pars
      in
      first_query :: queries

let to_datalog mtt =
   let clauses = ref [] in
   Mtt.iter (fun (state,symbol,action) ->
      let num = Mtt.num_of_parameters mtt state
      and symToString = Mtt_to_string.delta_to_string mtt
      and stToString = Mtt_to_string.state_to_string mtt
      and state_sym = Mtt_to_string.state_to_string mtt state in
      (* is the symbol excluded, arbitrary, or explicit *)
(*       if symbol>0 then  *)
(*          begin *)
            match action with
               | Mtt.Undefined -> ()
               | Mtt.Det rhs -> 
                   let var = fresh_var () 
                   and var_lst = fresh_var_list num in
                   let body = make_body symToString stToString var var_lst rhs in
                   let clause = D.Clause (state_sym,var :: var_lst,body) in
                   clauses := clause :: !clauses
               | Mtt.Non_det rhslst -> 
                   List.iter (fun rhs ->
                      let var = fresh_var ()
                      and var_lst = fresh_var_list num in
                      let body = make_body symToString stToString var var_lst rhs in
                      let clause = D.Clause (state_sym,var :: var_lst,body) in
                      clauses := clause :: !clauses
                   ) rhslst
(*          end  *)
(*       else if symbol=0 then *)
(*          begin *)
(*             Printf.printf "\t** arbitrary symbol %3d\n" symbol *)
(*          end *)
(*       else *)
(*          begin *)
(*             Printf.printf "\t** excluded symbol  %3d\n" symbol *)
            (*
            match action with
               | Mtt.Undefined -> ()
               | Mtt.Det rhs -> 
                   let var = fresh_var () 
                   and var_lst = fresh_var_list num in
                   let body = make_body symToString stToString var var_lst rhs in
                   let clause = D.Clause (state_sym,var :: var_lst,body) in
                   clauses := clause :: !clauses
               | Mtt.Non_det rhslst -> 
                   List.iter (fun rhs ->
                      let var = fresh_var ()
                      and var_lst = fresh_var_list num in
                      let body = make_body symToString stToString var var_lst rhs in
                      let clause = D.Clause (state_sym,var :: var_lst,body) in
                      clauses := clause :: !clauses
                   ) rhslst
            *)
(*          end *)
   ) mtt; !clauses


let make_query mtt dfa =
   let finals = List.map (fun x -> D.Const(D.Int x)) (B.final_states dfa) in
   List.fold_left (fun result init ->
      let init_sym = Mtt_to_string.state_to_string mtt init in
      List.fold_left (fun r final ->
         let pars = Mtt.num_of_parameters mtt init in
         [D.Query (init_sym, final :: fresh_var_list pars)] :: r
      ) result finals
   ) [] (Mtt.get_init mtt)


let out facts clauses queries = 
let i = List.length facts in
let j = List.length clauses in
Printf.printf "
number of facts (bottom-up automaton):\t%i\ti.e.,

%s

number of clauses (mtt)              :\t%i\ti.e.,

%s

top-down solving for the following queries:

%s

leads to the answer:\n\n" i (clauses_to_string facts) j 
(clauses_to_string clauses) 
(List.fold_left (fun r q -> r ^"\n"^ body_to_string q) "" queries) 

   (* this function corresponds to the original function print_env_list
    * of module Commands in the SuccinctSolver library; we only changed
    * the interpretation of an empty env_list, because this means in our
    * case, that our mtt is correct ;-) *)
let prt_env_list env_list =
   let prt_env env = Printf.printf "\t%s\n" (IO.string_of_env env) in
   match env_list with
      | [] -> (Printf.printf "\tcorrect.\n"; flush stdout)
      | _ -> (Printf.printf "\tincorrect.\n"; List.iter prt_env env_list; flush stdout)

let solve_mtt debug mtt dfa = 
   let facts = to_facts mtt (*(Mtt_to_string.delta_to_string mtt)*) dfa in
   let clauses = to_datalog mtt in
   let queries = make_query mtt dfa in
   let _ =  if debug then out facts clauses queries else () in
   let _ = set_clauses (facts @ clauses) in
   List.iter (fun query ->
      let body = IO.translate_body query in
      let env_list = Horn.answer body in
      prt_env_list env_list
      (* answer query *)
   ) queries;
   (* Printf.printf "\nDumping...\n\n"; dump () *)

