module type ItemEq =
   sig
      type t
      val eq : t -> t -> bool
      val to_string : t -> string
   end

module type Table =
   sig
      module I : ItemEq
   
      type item
      type 'a table
   
      val create : (('a -> 'a -> bool) * ('a -> string) * (string -> 'a)) 
                   -> 'a table
      val mem : 'a table -> 'a -> bool
      val add : 'a table -> 'a -> item -> unit
      val rm : 'a table -> 'a -> unit
      val lookup : 'a table -> 'a -> item
      val rev_lookup : 'a table -> item -> 'a list
      val lookup_all : 'a table -> 'a -> item list
      val iter : ('a -> item -> unit) -> 'a table -> unit
      val eq : 'a table -> ('a -> 'a -> bool)
      val to_string : 'a table -> ('a -> string)
      val from_string : 'a table -> (string -> 'a)
   end

module Make (Item : ItemEq) =
   struct
      module I = Item
      module H = Hashtbl
      type item = I.t

      exception Empty_item_list

      type 'a table =
         {
            mutable atoi : ('a,item list) H.t;
            mutable itoa : (item,('a * int) list) H.t;
            eq : ('a -> 'a -> bool);
            to_str : ('a -> string);
            from_str : (string -> 'a)
         }

      let create (eq,to_str,from_str) =
         {
            atoi = H.create 101;
            itoa = H.create 101;
            eq = eq;
            to_str = to_str;
            from_str = from_str
         }

      let rec list_rm eq x result = function
         [] -> result |
         (y,k)::ys -> 
            if eq x y
            then
               if k-1 = 0
               then list_rm eq x result ys
               else list_rm eq x ((y,k-1) :: result) ys
            else
               list_rm eq x ((y,k) :: result) ys

      let rec list_add eq x result = function
         [] -> (x,1) :: result |
         (y,k)::ys ->
            if eq x y
            then list_add eq x ((y,k+1) :: result) ys
            else list_add eq x ((y,k) :: result) ys

      let rec list_mk result = function
         [] -> result |
         (y,k)::ys -> list_mk (y :: result) ys

      let mem tbl = H.mem tbl.atoi

      let add tbl x i =
         let is = try H.find tbl.atoi x with Not_found -> []
         and xs = try H.find tbl.itoa i with Not_found -> []
         in
            (H.replace tbl.atoi x (i :: is);
             H.replace tbl.itoa i (list_add tbl.eq x [] xs))

      let rm tbl x = 
         let is = H.find tbl.atoi x
         in
            match is with
               [] -> raise Empty_item_list |
               y::ys -> 
                  (if List.length ys = 0
                   then H.remove tbl.atoi x
                   else H.replace tbl.atoi x ys;
                   let xs = H.find tbl.itoa y in
                   let xs' = list_rm tbl.eq x [] xs
                   in
                      if List.length xs' = 0 
                      then H.remove tbl.itoa y
                      else H.replace tbl.itoa y xs')
                      
      let lookup tbl x = List.hd (H.find tbl.atoi x)

      let lookup_all tbl x = H.find tbl.atoi x

      let rev_lookup tbl i = list_mk [] (H.find tbl.itoa i)

      let iter f tbl = H.iter (fun x l -> f x (List.hd l)) tbl.atoi

      let eq tbl = tbl.eq
      
      let to_string tbl = tbl.to_str

      let from_string tbl = tbl.from_str
   end
