(** implements bisimulation of mtt states *)

open RightHandSide
open Mtt_to_string
module H = Hashtbl


exception Undefined_action
exception Non_deterministic_mtt


type condition =
   Par_cond of int |
   Var_cond of int * ((int * int list) * int) list

let add_cond h var cond =
   if H.mem h var then
      let c = H.find h var in
      H.replace h var (cond :: c)
   else
      H.add h var [cond]

let action = function
   Empty_rhs -> None |
   Parameter i -> None |
   Symbol (s,k,rhs_list) -> Some s |
   Call (q,x,k,rhs_list) -> None


let rec inspect_rule mtt mappings (state,sym,action) =
   let (c,r) = inspect_action mtt action in
   (0,sym,r)

and inspect_action mtt action =
   match action with
      Mtt.Undefined -> raise Undefined_action |
      Mtt.Det rhs -> inspect_rhs mtt rhs |
      Mtt.Non_det rhs_list -> raise Non_deterministic_mtt

and inspect_rhs mtt rhs =
   let action = action rhs in
   let current = ref 1 in
   let cond = H.create 11 in
   let pars = H.create 11 in
   let rec doit = function
      Empty_rhs -> (-1) |
      Parameter i -> 
         let j =
            try
               H.find pars i
            with
               Not_found ->
                 let j = !current in
                 let _ = current := j + 1 in 
                 let _ = H.add pars i j in j
         in j |
      Symbol (s,k,rhs_list) -> 
         let p = List.map (doit) rhs_list in (-1) |
      Call (q,x,k,rhs_list) ->
         let i = !current in
         let _ = current := i + 1 in
         let p = List.map (doit) rhs_list in
         let _ = add_cond cond x ((q,p),i) in i in
   let _ = doit rhs
   in
   (action,cond)

and inspect_rule' mtt mappings (state,sym,action) =
   (* no parameters *)
   let (c,r) = inspect_action' mtt mappings action in
   (0,sym,r)

and inspect_action' mtt mappings action =
   match action with
      Mtt.Undefined -> raise Undefined_action |
      Mtt.Det rhs -> inspect_rhs' mtt mappings rhs |
      Mtt.Non_det rhs_list -> raise Non_deterministic_mtt

and inspect_rhs' mtt mappings rhs =
   let action = action rhs in
   let current = ref 1 in
   let vars = H.create 11 in
   let rec doit = function
      Empty_rhs -> () |
      Parameter i -> () |
      Symbol (s,k,rhs_list) -> 
         List.iter (doit) rhs_list |
      Call (q,x,k,_) ->
         let i = !current in
         let _ = current := i + 1 in
         add_cond vars x ((q,[]),i) in
   let _ = doit rhs
   in
   (action,vars)

      

let inspect mtt =
   let mappings = Dictionary.create () in
   Mtt.iter (fun (state,sym,action) ->
      let _ = Printf.printf "\ncurrent rule:\t%s\n" 
              (rule_to_string mtt (state,sym,action)) in
      let (r0,s,vars) = 
         if (Mtt.num_of_parameters mtt state) > 0 then
            inspect_rule mtt mappings (state,sym,action)
         else
            inspect_rule' mtt mappings (state,sym,action) 
      in 
      Printf.printf "==>\t(%s,%i,%s)\n\n"
         ("{"^ string_of_int r0 ^"}")
         (s)
         (H.fold (fun x l r ->
            r ^" x_"^ (string_of_int x) ^": {"^
            (List.fold_left (fun r' ((q,l'),t) ->
               r' ^" ^ (q"^ (string_of_int q) ^","^
               (List.fold_left (fun s e ->
                  s^" r"^(string_of_int e)
               ) "" l')^")->r"^ (string_of_int t)
            ) (try let ((q,l'),t) =List.hd l in
               "(q"^(string_of_int q) ^","^
               (List.fold_left (fun s e ->
                  s^" r"^(string_of_int e)
               ) "" l')^")->r"^(string_of_int t) 
               with Failure f->"") 
              (try List.tl l with Failure f->[])) ^"}"
         ) vars "")
   ) mtt


let bisim mtt =
   let _ = print mtt in 
   let _ = inspect mtt in 
   []

