module List_funs =
struct
   exception Failure of string

   let rec single = function
      [] -> [] |
      e::es -> [e] :: single es

   let rec pre x = function
      [] -> [] |
      lst::lsts -> (x::lst) :: pre x lsts

   let rec pair lst = function
      [] -> [] |
      e::es -> (pre e lst) @ pair lst es

   let rec prepend result lst = function
      [] -> result |
      e::es -> prepend ((e::lst) :: result) lst es

   let rec combine result lst = function
      [] -> result |
      e::es -> combine (prepend result [e] lst) lst es

   let k_tuples k lst =
      let rec doit = function
         0 -> [] |
         1 -> single lst |
         2 -> pair (single lst) lst |
         n -> pair (doit (n-1)) lst
      in
         doit k

   let tuple lst = 
      let rec doit = function
         [] -> [] |
         lst::[] -> single lst |
         lst1::lst2::[] -> combine [] lst1 lst2 |
         lst::lsts -> pair (doit lsts) lst
      in
         List.map List.rev (doit (List.rev lst))

   let bin_tuple lst1 lst2 =
      List.fold_left (fun result e1 ->
         (List.fold_left (fun r e2 ->
            (e1,e2) :: r
         ) [] lst2) @ result
      ) [] lst1

   let rec contains_empty = function
      [] -> false |
      lst::lsts -> (List.length lst = 0) || contains_empty lsts

   let minus lst1 lst2 =
      List.fold_left (fun result e ->
         if List.mem e lst2 then result
         else e :: result
      ) [] lst1

   let insert lst element i =
      let rec doit j c e l = function
         [] -> 
            if c=j then List.rev (e::l) else raise (Failure "insert") |
         x::xs -> 
            if c=j then 
               (List.rev (x::e::l)) @ xs 
            else 
               doit j (c+1) e (x::l) xs
      in
         doit i 0 element [] lst

   let union lst1 lst2 =
      let rec doit result = function
         [] -> result |
         x::xs ->
            if List.mem x lst1 then
               doit result xs
            else
               doit (x :: result) xs
      in
         doit lst1 lst2
end

module Equiv_table =
struct
   type elt = int
   type table = (elt * elt, bool) Hashtbl.t

   let create lst =
      let l = List.length lst in
      let tab = Hashtbl.create (l*l) in
      List.iter (fun a ->
         List.iter (fun b ->
            Hashtbl.add tab (a,b) false
         ) lst
      ) lst; tab

   let mark tab (a,b) = 
      Hashtbl.replace tab (a,b) true;
      Hashtbl.replace tab (b,a) true

   let unmark tab (a,b) = 
      Hashtbl.replace tab (a,b) false;
      Hashtbl.replace tab (b,a) false

   let is_marked tab (a,b) = Hashtbl.find tab (a,b)
   let is_unmarked tab (a,b) = not (is_marked tab (a,b))

   let iter f tab = Hashtbl.iter f tab

   let fold f tab init = Hashtbl.fold f tab init

   let iter_unmarked f tab = 
      Hashtbl.iter (fun (a,b) marked ->
         if (not marked) then f (a,b) 
      ) tab

   let fold_unmarked f tab init =
      Hashtbl.fold (fun (a,b) marked result ->
         if (not marked) then f (a,b) result else result
      ) tab init

   let to_string tab =
      Hashtbl.fold (fun (a,b) marked result ->
         if a<b then
         ( (string_of_int a)^","^(string_of_int b)^" :   "^
          (if marked then "x\n" else "\n") ) ^ result
         else result
      ) tab ""

   let partition tab lst =
      let partitions = Hashtbl.create 101 in
      let inclusions = Hashtbl.create 101 in
      let current = ref 0 in
      let _ = 
         Hashtbl.iter (fun (a,b) marked ->
            if (not marked) && (a<b) then
               if Hashtbl.mem inclusions a then
                  let p = try Hashtbl.find inclusions a with Not_found ->
                          (Printf.printf "not found\n"; raise Not_found) in
                  let cls = Hashtbl.find partitions p in
                  let _ = 
                     if not (List.mem b cls) then
                        Hashtbl.replace partitions p (b::cls) in
                  Hashtbl.add inclusions b p
               else if Hashtbl.mem inclusions b then
                  let p = Hashtbl.find inclusions b in
                  let cls = Hashtbl.find partitions p in
                  let _ = 
                     if not (List.mem a cls) then
                        Hashtbl.replace partitions p (a::cls) in
                  Hashtbl.add inclusions a p
               else
                  let i = !current in
                  let _ = Hashtbl.add partitions i [a;b] in
                  let _ = Hashtbl.add inclusions a i in
                  let _ = Hashtbl.add inclusions b i in
                  current := i + 1
         ) tab in
      let _ =
         List.iter (fun a ->
            if Hashtbl.mem inclusions a then
               ()
            else
               let i = !current in
               let _ = Hashtbl.add partitions i [a] in
               let _ = Hashtbl.add inclusions a i in
               current := i + 1
         ) lst
      in
         (partitions,inclusions)          
end


exception Incompatible_rank


type state = int
   (* allow the definition of (negated) sets of symbols/don't cares *)
type symbol = int
type forest = int
type states = int
type set = Det of state | Non_det of states

type transition = (symbol, (forest, set) Hashtbl.t) Hashtbl.t 

type forests = (forest, state list) Dictionary.dictionary
type sets = (states, state list) Dictionary.dictionary

type automaton = {
   transition : transition;
   forest : forests;
   set : sets;
   final : state list
}
type t = automaton

(* don't care symbol *)
let dc = -17 

let empty_tree_state = -1

let make d f s fin = {
   transition = d;
   forest = f;
   set = s;
   final = fin
}

let trans automaton = automaton.transition
let sets automaton = automaton.set
let forests automaton = automaton.forest

let args_of_i automaton index = 
   Dictionary.lookup automaton.forest index
let i_of_args automaton forest = 
   Dictionary.lookup_rev automaton.forest forest

let args_of_i' dict index = 
   Dictionary.lookup dict index
let i_of_args' dict forest = 
   Dictionary.lookup_rev dict forest

let sets_of_i automaton index =
   Dictionary.lookup automaton.set index
let i_of_sets automaton set =
   Dictionary.lookup_rev automaton.set set

let sets_of_i' dict index =
   Dictionary.lookup dict index
let i_of_sets' dict set =
   Dictionary.lookup_rev dict set

let list_of_states automaton set =
   match set with
      Det q -> [q] |
      Non_det s -> sets_of_i' (sets automaton) s

let args_contain arguments state =
   let rec doit pos = function
      [] -> None |
      x::xs -> if x = state then Some pos else doit (pos+1) xs
   in
      doit 0 arguments
let set_contains set state = args_contain set state




let is_final_state automaton q = List.mem q automaton.final

let is_deterministic automaton =
   Hashtbl.fold (fun _ trans result ->
      Hashtbl.fold (fun _ states r ->
         match states with
            Det q -> r |
            Non_det s -> false && r
      ) trans result
   ) (trans automaton) true

let number_of_states automaton =
   let h = Hashtbl.create 53 in
   Hashtbl.fold (fun _ trans sum ->
      Hashtbl.fold (fun args stats' sum' ->
         let stats = list_of_states automaton stats' in
         (List.fold_left (fun s q ->
            if Hashtbl.mem h q then s
            else let _ = Hashtbl.add h q true in (s+1)
         ) 0 stats) +
         (List.fold_left (fun s q ->
            if Hashtbl.mem h q then s
            else let _ = Hashtbl.add h q true in (s+1)
         ) 0 (args_of_i automaton args)) + sum'
      ) trans sum
   ) (trans automaton) 0

let symbols automaton =
   let h = Hashtbl.create 101 in
   Hashtbl.fold (fun symbol _ result ->
      if Hashtbl.mem h symbol then result
      else (Hashtbl.add h symbol true; symbol :: result)
   ) (trans automaton) []

let arity_of_symbol automaton symbol =
   let trans = Hashtbl.find (trans automaton) symbol in
   Hashtbl.fold (fun args _ length ->
      List.length (args_of_i automaton args)
   ) trans 0

let states automaton =
   let h = Hashtbl.create 101 in
   Hashtbl.fold (fun _ trans result ->
      Hashtbl.fold (fun args stats' result' ->
         let stats = list_of_states automaton stats' in
         let first =
            List.fold_left (fun l q ->
               if Hashtbl.mem h q then l
               else let _ = Hashtbl.add h q true in (q :: l)
            ) result' stats
         in
         List.fold_left (fun l q ->
            if Hashtbl.mem h q then l
            else let _ = Hashtbl.add h q true in (q :: l)
         ) first (args_of_i automaton args)
      ) trans result
   ) (trans automaton) []

let final_states automaton = automaton.final

let leaf_states automaton =
   Hashtbl.fold (fun _ trans result ->
      Hashtbl.fold (fun args stats' result' ->
         let stats = list_of_states automaton stats' in
         if List.length (args_of_i automaton args) = 0 then
            List.fold_left (fun r q ->
               if List.mem q r then r else q :: r
            ) result' stats
         else
            result'
      ) trans result
   ) (trans automaton) []

let leaf_symbols automaton =
   Hashtbl.fold (fun symbol trans result ->
      Hashtbl.fold (fun args stats result' ->
         if List.length (args_of_i automaton args) = 0 then
            if List.mem symbol result' then result' 
            else symbol :: result'
         else
            result'
      ) trans result
   ) (trans automaton) []

let contains h lst =
   List.fold_left (fun b q ->
      let b' = try Hashtbl.find h q with Not_found -> false in
      b' && b
   ) true lst

let rules automaton =
   Hashtbl.fold (fun symbol trans rules ->
      Hashtbl.fold (fun args stats rules' ->
         (symbol,args,stats) :: rules'
      ) trans rules
   ) (trans automaton) []

let iter f automaton =
   Hashtbl.iter (fun symbol trans ->
      Hashtbl.iter (fun args' stats' ->
         let stats = list_of_states automaton stats'
         and args = args_of_i automaton args' 
         in
         f (symbol,args,stats)
      ) trans
   ) (trans automaton)

let reachable_states automaton =
   let reachable = Hashtbl.create 53 in
   let insert lst = 
      List.iter (fun q -> 
         let b' = try Hashtbl.find reachable q with Not_found -> false in
         if not b' then Hashtbl.add reachable q true
      ) lst in
   let rec doit rules states = function
      [] -> (rules,states) |
      (s,args,sts)::rs ->
         if contains reachable (args_of_i automaton args) then
            doit rules ((list_of_states automaton sts) @ states) rs
         else
            doit ((s,args,sts)::rules) states rs in
   let rec iterate rules =
      if List.length rules = 0 then
         reachable
      else
         let (new_rules,new_states) = doit [] [] rules in
         if List.length new_states = 0 then
            reachable
         else
            let _ = insert new_states in
            iterate new_rules
   in
   insert (leaf_states automaton);
   iterate (rules automaton)

let emptiness automaton = false

let reduce automaton =
   let reachable = reachable_states automaton in
   let new_transition = Hashtbl.create 101 in
   let _ = 
      Hashtbl.iter (fun symbol trans ->
         Hashtbl.iter (fun args states ->
            let arguments = args_of_i automaton args in
            if contains reachable arguments then
               if Hashtbl.mem new_transition symbol then
                  Hashtbl.add (Hashtbl.find new_transition symbol) args states
               else
                  let t = Hashtbl.create 53 in
                  let _ = Hashtbl.add t args states in
                  Hashtbl.add new_transition symbol t
         ) trans 
      ) (trans automaton) in
   let new_final = 
      List.fold_left (fun result q ->
         if Hashtbl.mem reachable q then
            q :: result
         else
            result
      ) [] (final_states automaton)
   in
   make new_transition (forests automaton) (sets automaton) new_final

let complement automaton =
   let states = states automaton in
   let new_final =
      List.fold_left (fun result q ->
         if List.mem q (final_states automaton) then
            result
         else
            q :: result
      ) [] states
   in
   make (trans automaton) (forests automaton) (sets automaton) new_final

(*
let union (a,enc_a) (b,enc_b) eq =
   let alpha = a.transition in
   let beta = b.transition in
   let g = Dictionary.create () in
   let gamma = Hashtbl.create 101 in
   let new_forest = Dictionary.create () in
   let new_set = Dictionary.create () in
   let _ =
      Hashtbl.iter (fun symbol trans ->
         Hashtbl.iter (fun symbol' trans' ->
            if eq symbol symbol' then
               Hashtbl.iter (fun args sts ->
                  Hashtbl.iter (fun args' sts' ->
                     let arguments = Dictionary.lookup a.forest args in
                     let arguments' = Dictionary.lookup b.forest args' in
                     let new_arguments = 
                        try
                           List.map (fun q ->
                              if not (Dictionary.mem_rev g q) then
                                 let index = Dictionary.size g + 1 in
                                 Dictionary.add g (index,q); index
                              else
                                 Dictionary.lookup_rev g q
                           ) (List.combine arguments arguments')
                        with
                           Invalid_argument f -> raise Incompatible_rank in
                     let new_args =
                        try
                           Dictionary.lookup_rev new_forest new_arguments
                        with
                           Not_found ->
                              let index = Dictionary.size new_forest + 1 in
                              Dictionary.add new_forest (index,new_arguments);
                              index in
                     let states = sets_of_i a sts in
                     let states' = sets_of_i b sts' in
                     let new_states =
                        List.map (fun q ->
                           if not (Dictionary.mem_rev g q) then
                              let index = Dictionary.size g + 1 in
                              Dictionary.add g (index,q); index
                           else
                              Dictionary.lookup_rev g q
                        ) (List_funs.bin_tuple states states') 
                     in

                     if Hashtbl.mem gamma symbol then
                        let trans = Hashtbl.find gamma symbol in
                           try 
                              let stats' = Hashtbl.find trans new_args in
                              let states = new_states @ stats' in
                              Hashtbl.replace trans new_args states
                           with
                              Not_found ->
                                 Hashtbl.add trans new_args new_states
                     else
                        let trans = Hashtbl.create 53 in
                        let _ = Hashtbl.add trans new_args new_states in
                        Hashtbl.add gamma symbol trans 

                  ) trans'
               ) trans
            else
               ()
         ) beta
      ) alpha in
   let new_final = 
      Dictionary.fold (fun q (x,y) fin ->
         if (is_final_state a x) or (is_final_state b y) then q :: fin
         else fin
      ) g [] in
   let enc q = let (x,y) = Dictionary.lookup g q in (enc_a x,enc_b y) in
   (make gamma new_forest new_set new_final,enc)
*)


let do_transition automaton symbol args debug =
   let result =
      try
         let arguments = i_of_args automaton args in
         let t = Hashtbl.find (trans automaton) symbol in
         list_of_states automaton (Hashtbl.find t arguments)
      with
         Not_found -> []
   in
   let _ = if debug then
      Printf.printf "\ttrying transition delta(%i,[%s])=?\t%s\n"
      symbol
      (List.fold_left (fun s q -> s ^ "," ^ (string_of_int q)) 
         (try string_of_int (List.hd args) with Failure f -> "")
         (try List.tl args with Failure f -> []))
      (List.fold_left (fun s q -> s ^ "," ^ (string_of_int q)) 
         (try string_of_int (List.hd result) with Failure f -> "")
         (try List.tl result with Failure f -> []))
   in
      result

let rec state automaton enc debug depth tree = 
   if Tree_input.is_node tree then
      let symbol = enc (Tree_input.label tree) in
      let _ = if debug then
         Printf.printf "[%i] current label:\t%i\n" depth symbol in
      let args = 
         List.map (state automaton enc debug (depth+1)) (Tree_input.sons tree) in
      let _ = if debug then 
         Printf.printf "[%i] argument states:\t[%s]\n"
         depth
         (List.fold_left (fun r l ->
            r ^ "[" ^ (
            List.fold_left (fun s q -> s ^ "," ^ (string_of_int q)) 
               (try string_of_int (List.hd l) with Failure f -> "")
               (try List.tl l with Failure f -> [])
            ) ^ "]"
         ) "" args)
      in
      if List.length args > 0 then
         List.fold_left (fun result args' ->
            let _ = if debug then
               Printf.printf "[%i] testing delta(%i,[%s]) ...\n"
               depth symbol
               (List.fold_left (fun r i -> r ^ "," ^ (string_of_int i)) 
                  (try string_of_int (List.hd args') with Failure f -> "")
                  (try List.tl args' with Failure f -> [])
               )
            in
            (do_transition automaton symbol args' debug) @ result
         ) [] (List_funs.tuple args)
      else
         let _ = if debug then
            Printf.printf "[%i] testing with empty list\n" depth
         in
         do_transition automaton symbol [] debug
   else
      let _ = Printf.printf ".\n" in
      []

let member tree automaton enc debug = 
   List.fold_left (fun result q ->
      let _ = if debug then
         Printf.printf "is '%i' a final state?\t..." q in
      let fin = is_final_state automaton q in
      let _ = if debug then
         Printf.printf "%s\n" (if fin then "yes\n" else "no\n")
      in
      fin || result 
   ) false (state automaton enc debug 0 tree)



let add_forest dict args =
   try
      i_of_args' dict args
   with
      Not_found ->
         let new_index = (Dictionary.size dict) + 1 in
         let _ = Dictionary.add dict (new_index,args) in
         new_index

let add_set dict set = add_forest dict set

let add_states_to_set sets set states =
   match set with
      Det q -> 
         let new_set = List_funs.union [q] states in
         if (List.length new_set) = 1 then
            Det (List.hd new_set)
         else
            Non_det (add_set sets new_set) |
      Non_det s ->
         let states' = args_of_i' sets s in
         let new_set = List_funs.union states' states 
         in
            Non_det (add_set sets new_set)

let add_rule' transitions forests sets symbol arguments states =
   let forest = add_forest forests arguments in
   let t = 
      if Hashtbl.mem transitions symbol then
         Hashtbl.find transitions symbol
      else
         Hashtbl.create 53 in
   let s =
      if Hashtbl.mem t forest then
         let set = Hashtbl.find t forest in
         add_states_to_set sets set states
      else
         if (List.length states) = 1 then
            Det (List.hd states)
         else
            Non_det (add_set sets states) in
   let _ = Hashtbl.replace t forest s
   in
      Hashtbl.replace transitions symbol t

      
let determinize automaton enc =
   if is_deterministic automaton then 
      let state_enc q = [enc q] in
      (automaton, state_enc)
   else
      let symbols = symbols automaton in
      let states = Dictionary.create () in
      let current = ref 0 in
      let new_trans = Hashtbl.create 101 in
      let new_forest = Dictionary.create () in
      let new_set = Dictionary.create () in
      let empty = let _ = Dictionary.add new_forest (0,[]) in 0 in
      let _ = 
         let i = !current in
         let _ = Dictionary.add states (i,leaf_states automaton) in
         let _ =
            List.iter (fun symbol ->
               let trans = Hashtbl.create 53 in
               let _ = Hashtbl.add trans empty (Det i) in
               Hashtbl.add new_trans symbol trans
            ) (leaf_symbols automaton)
         in
         current := i + 1 in
      let add () = 
         List.fold_left (fun result symbol ->
            let k = arity_of_symbol automaton symbol in
            let argv = List_funs.k_tuples 2 (Dictionary.to_list states) in
            List.fold_left (fun result' args' ->
               let a = List.map (Dictionary.lookup states) args' in
               let s =
                  List.fold_left (fun s' arg ->
                     s' @ (do_transition automaton symbol arg false)
                  ) [] (List_funs.tuple a) in
               if List.length s = 0 then
                  false && result'
               else
                  let s' = 
                     try
                        Dictionary.lookup_rev states s
                     with
                        Not_found ->
                           let index = !current in
                           let _ = Dictionary.add states (index, s) in
                           let _ = current := index + 1 in
                           index in
                  let _ =
                     add_rule' new_trans new_forest new_set symbol args' [s']
                  in
                  true && result'
            ) result argv 
         ) false symbols in
      let _ = 
         let rec iterate b =
            if b then iterate (add ())
         in
            iterate true in
      let new_final = 
         Dictionary.fold (fun q set result ->
            let contains_final () = 
               List.fold_left (fun r p ->
                  (is_final_state automaton p) || r
               ) false set
            in
               if contains_final () then
                  q :: result
               else
                  result
         ) states [] in
      let state_enc q = 
         let s = Dictionary.lookup states q in
         List.fold_left (fun r q -> (enc q) :: r) [] s
      in
      ((make new_trans new_forest new_set new_final), state_enc)


let minimize a enc = 
   let a' = reduce a in
   let (automaton,enc') = determinize a' enc in
   let states = (empty_tree_state (*-1*)) :: (states automaton) in
   let symbols = symbols automaton in
   let equiv_tab = Equiv_table.create states in
   let test q q' eq =
      List.fold_left (fun result symbol ->
         let rules = Hashtbl.find (trans automaton) symbol in
         let k = arity_of_symbol automaton symbol in
         let tuples = 
            if k > 0 then 
               List_funs.k_tuples (k-1) states
            else
               [] in
         result ||
         (List.fold_left (fun r tuple ->
            let rec doit b i =
               let arg1 = List_funs.insert tuple q i in
               let s1' = do_transition automaton symbol arg1 false in
               let s1 = if (List.length s1' < 1) then [empty_tree_state (*-1*)] else s1' in
               let arg2 = List_funs.insert tuple q' i in
               let s2' = do_transition automaton symbol arg2 false in
               let s2 = if (List.length s2' < 1) then [empty_tree_state (*-1*)] else s2' in
               let marked = 
                  List.fold_left (fun r' (p,p') ->
                     let m = Equiv_table.is_marked eq (p,p') in m || r'
                  ) false (List_funs.bin_tuple s1 s2)
               in
               match i with 
                  0 -> b || marked |
                  j -> b || marked || doit b (j-1) 
            in
            doit r (k-1) 
         ) false tuples)
      ) false symbols in
   let init () =
      let test q = is_final_state automaton q || q = (empty_tree_state (*-1*)) in
      Equiv_table.iter (fun (q,q') marked ->
         if q<q' && not marked then
         if (test q) && not (test q') then
            Equiv_table.mark equiv_tab (q,q')
         else if not (test q) && (test q') then
            Equiv_table.mark equiv_tab (q,q')
         else
            ()
      ) equiv_tab in
   let refine () =
      let changed = ref false in
      let _ =
         Equiv_table.iter (fun (q,q') marked ->
            if q<q' && not marked then
            if test q q' equiv_tab then
               (Equiv_table.mark equiv_tab (q,q'); changed := true)
            else
               ()
         ) equiv_tab in !changed in
   let _ = init () in
   let _ = 
      let rec iterate b c =
         if b then iterate (refine ()) (c+1)
      in
         iterate true 1 in
   let (partitions,inclusions) = Equiv_table.partition equiv_tab states in
   (*
   let _ = Printf.printf "partitions:\t%s\n"
      (Hashtbl.fold (fun i l s ->
         "P"^(string_of_int i)^"= {"^(List.fold_left (fun s' e ->
            s' ^ "," ^ (string_of_int e)
         ) (try string_of_int (List.hd l) with Failure f -> "")
         (try List.tl l with Failure f -> []))^"}  " ^ s
      ) partitions "") in
   *)
   let new_trans = Hashtbl.create 101 in
   let new_forest = Dictionary.create () in
   let new_set = Dictionary.create () in
   let current = ref 0 in
   let _ = 
      Hashtbl.iter (fun symbol trans ->
         Hashtbl.iter (fun args state ->
            let arguments = args_of_i automaton args in
            let arguments' = List.map (Hashtbl.find inclusions) arguments in
            let state' = list_of_states automaton state in
            let states' = List.map (Hashtbl.find inclusions) state' in
            add_rule' new_trans new_forest new_set symbol arguments' states'
         ) trans 
      ) automaton.transition in
   let new_final = 
      Hashtbl.fold (fun p qs result ->
         let b = 
            List.fold_left (fun b' q -> 
               b' || is_final_state automaton q
            ) false qs 
         in
         if b then p :: result else result
      ) partitions [] in
   let state_enc q = 
      let s = Hashtbl.find partitions q in
      List.fold_left (fun r q ->
         if q >= 0 then (enc q) :: r else r
      ) [] s
   in
   (make new_trans new_forest new_set new_final, state_enc)



let from_list rules fin =
   let rec do_rules h f s = function
      [] -> () |
      (sym,args,stats)::rs -> 
         let _ = add_rule' h f s sym args stats
         in
         do_rules h f s rs in
   let mk_rules l =
      let h = Hashtbl.create (List.length l) in
      let f = Dictionary.create () in
      let s = Dictionary.create () in
      let _ = do_rules h f s l in
      (h,f,s) in
   let (new_rules,new_forests,new_sets) = mk_rules rules in
   make new_rules new_forests new_sets fin


let rules_to_string automaton sym_enc state_enc =
   let f q = 
      (if is_final_state automaton q then "." else "") ^ (state_enc q) in
   Hashtbl.fold (fun symbol trans result ->
      Hashtbl.fold (fun arguments stats' result' ->
         let stats = list_of_states automaton stats' in
         let args = args_of_i automaton arguments in
         result' ^ "   (" ^
         (sym_enc symbol) ^ ",<" ^
         (List.fold_left (fun r i -> r ^ "," ^ (state_enc i)) 
            (try state_enc (List.hd args) with Failure f -> "")
            (try List.tl args with Failure f -> []))
         ^ ">) -> {" ^
         (List.fold_left (fun r i -> r ^ "," ^ (f i)) 
            (try f (List.hd stats) with Failure f -> "")
            (try List.tl stats with Failure f -> []))
         ^ "}\n"
      ) trans result
   ) (trans automaton) ""

let rules_to_short_string automaton =
   rules_to_string automaton string_of_int string_of_int

let to_string automaton sym_enc state_enc =
   "\nFinal states:\n   {" ^
   (List.fold_left (fun result q -> result ^","^ (state_enc q)) 
      (try state_enc (List.hd automaton.final) with Failure f -> "")
      (try List.tl automaton.final with Failure f -> []))
   ^ "}\nTransition rules:\n" ^
   (rules_to_string automaton sym_enc state_enc) ^ "\n\n"

let to_short_string automaton =
   to_string automaton string_of_int string_of_int

