(* implementation of an emptiness check for mtts by means of solving
   a constraint system *)


open RightHandSide
open Printf

module D = Lattice.BooleanLattice
module H = Hashtbl

module Item =
   struct
      type item = Mtt.state list

      let state_to_string = ref (fun x -> "")
      let set_to_string mtt = 
         state_to_string := Mtt_to_string.state_to_string mtt

      let to_string item = 
         let rec doit = function
            | [] -> ["]"]
            | x :: xs -> ("; "^ !state_to_string x ^" ("^ string_of_int x ^")") :: doit xs
         in
         match item with
            | [] -> "[]"
            | x :: xs -> String.concat "" (("["^ !state_to_string x ^" ("^
            string_of_int x ^")") :: doit xs)

      let rec to_int = function
         | [] -> 19
         | x :: xs -> x + 31 * to_int xs

      let compare a b = 
         let a' = (to_int a) in
         let b' = (to_int b) in
         Pervasives.compare a' b'

      let eq a b = (compare a b) = 0
   end

module Solve_c = SolverCon.SolverCon (Item) (D)
module Solve_e = SolverEq.SolverEq (Item) (D)



   (* adds el to list lst (if el is not element of lst) and returns
      a sorted list; 
      the compare function comp is given as list -- if this list is
      empty, we try the builtin compare function *)
let add lst el comp =
   let compare = try List.hd comp
                 with Failure e -> Pervasives.compare
   in let l' = if not (List.mem el lst) then el :: lst
               else lst
   in List.fast_sort compare l'



let rec get_calling_funs x = function
   | Call (q,y,k,rhs_lst) -> let t = get_funs' x rhs_lst
                             in if x=y then add t q []
                                else t
   | Symbol (s,n,rhs_lst) -> get_funs' x rhs_lst
   | _ -> [] 

and get_funs' x rhs_lst = List.concat (List.map (get_calling_funs x) rhs_lst)



let calls mtt qs s =
   let rank = Mtt.arity mtt s

   in if rank >= 0 then

      let var_map rhs =
         let a = Array.init (rank+1) (fun x -> [])
         in let rec doit i =
            match i with
               | 0 -> ()
               | _ -> (a.(i) <- get_calling_funs i rhs; doit (i-1))
         in let _ = doit rank
         in a
   
      in let actions = List.map (fun q -> 
                          let action = Mtt.get_action mtt q s
                          in match action with
                             | Mtt.Det rhs -> (rhs, var_map rhs) :: []
                             | Mtt.Non_det rhss -> List.map (fun rhs ->
                                                      (rhs, var_map rhs)
                                                   ) rhss
                             | _ -> []
                       ) qs
   
      in let rec combine_x x result = function
         | [] -> result
         | (rhs,map) :: r -> let qs = map.(x)
                             in let rec doit result' = function
                                | [] -> result'
                                | p::ps -> doit (add result' p []) ps
                             in combine_x x (doit result qs) r
   
      in let rec combine r = function
         | [] -> r
         | t :: xs -> let rec doit r' = function
                         | 0 -> r'
                         | i -> let r'' = add r' (combine_x i [] t) [Item.compare]
                                in doit r'' (i-1)
                      in (doit [] rank) :: r
   
      in combine [] (Util.tuple actions)

   else []



let goals mtt b =
   let h = H.create 101
   in let any = ref []
   in let _ = List.iter (fun q -> 
                 let dom = Mtt.get_domain mtt q 
                 in match dom with
                    | Mtt.OneOf l -> List.iter (fun s ->
                                        if H.mem h s then
                                           let t = H.find h s 
                                           in H.replace h s (add t q [])
                                        else
                                           H.add h s (q :: [])
                                     ) l
                    | Mtt.Any -> any := q :: !any
                    | _ -> () (* todo: NoneOf *)
              ) b 
   in H.fold (fun s qs result ->
      (calls mtt qs) s @ result
   ) h []

   

let empty mtt = 
   let _ = Item.set_to_string mtt

   in let start = List.fold_left (fun r e ->
                   (e :: []) :: r
                ) [] (Mtt.get_initial_state mtt)
   
   in let constr_of_list lst sigma =
      let rec doit = function
         | [] -> D.top ()
         | x::xs -> if (D.top () = sigma x)  or (List.length x = 0) then doit xs
                    else D.bot ()
      in doit lst

   in let rec constr b = List.map constr_of_list (goals mtt b)

   in let result = Solve_c.T.to_list (Solve_c.solve constr start)

   in printf "%s\n" (
      List.fold_left (fun r (x,b) ->
         r ^ 
         Item.to_string x ^": "^ (if b then "true" else "false") ^"\n"
      ) "" result 
      );
      printf "\n%s\n" (
      if (List.fold_left (fun r (x,b) -> if List.mem x start then r || b else r)
      false result) then "no" else "yes"
      )














