type 'a t = ('a, 'a list) Hashtbl.t     

let empty () = Hashtbl.create 50

let copy = Hashtbl.copy

let _add_edge g x y =
   let adj_list = 
      try
         Hashtbl.find g x 
      with
         Not_found -> [] 
   in
      if (not (List.mem y adj_list)) then
         Hashtbl.replace g x (y :: adj_list)

let add_edge g x y =
   let g' = copy g in
   let adj_list = 
      try Hashtbl.find g x with Not_found -> [] 
   in
      if (not (List.mem y adj_list)) then
         Hashtbl.replace g' x (y :: adj_list); g'

let _remove_vertex g x =
   if (Hashtbl.mem g x) then
      Hashtbl.remove g x

let remove_vertex g x =
   let g' = copy g in
      _remove_vertex g' x; g'

let _add_vertex g x =
   if (not (Hashtbl.mem g x)) then
      Hashtbl.add g x []

let add_vertex g x =
   let g' = copy g in
      _add_vertex g' x; g'

let rec _add_adjacency_list g v = function
   [] -> 
      if (not (Hashtbl.mem g v)) then
         Hashtbl.add g v [] |
   u::us ->
      _add_edge g v u;
      _add_adjacency_list g v us

let add_adjacency_list g v l = 
   let g' = copy g in
      _add_adjacency_list g' v l; g'
   
let rec _from_list g = function
   [] -> () |
   (x,y)::es -> _add_edge g x y; _from_list g es

let from_list g l = 
   let g' = copy g in
      _from_list g' l; g'

let adjacent_vertices g x =
   Hashtbl.find g x

let rec dfs g visited = function
   [] -> visited |
   u::q ->
      if List.mem u visited then
         dfs g visited q
      else
         let adj =
            try
               adjacent_vertices g u
            with
               Not_found -> []
         in
            dfs g (u :: visited) (adj @ q)

let reachable g x = dfs g [] [x]
   
let out_degree g x =
   try 
      List.length (Hashtbl.find g x) 
   with
      Not_found -> 0

let in_degree g x =
   let r = ref 0 in
      Hashtbl.iter
         (fun v adj_list ->
             if (List.mem x adj_list) then
                r := (!r) + 1) g; (!r)

let edges g =
   let r = ref [] in
      Hashtbl.iter 
         (fun v adj_list -> 
             List.iter (fun x -> r := (v,x) :: (!r)) adj_list) g; (!r)

let vertices g =
   let r = ref [] in
      Hashtbl.iter 
         (fun v adj_list -> 
            if (not (List.mem v (!r))) then
               r := v :: (!r);
            List.iter 
               (fun x -> 
                  if (not (List.mem x (!r))) then
                     r := x :: (!r)) 
                adj_list) g; (!r)

let transpose g =
   let g_t = empty () in
   let e = edges g in
      List.iter (fun (v,u) -> _add_edge g_t u v) e;
      g_t

let dfs1 g =
   let v_set = vertices g in
   let color = Hashtbl.create (List.length v_set) in
   let d = Hashtbl.create (List.length v_set) in
   let f = Hashtbl.create (List.length v_set) in
   let time = ref 0 in
   let rec init = function
      [] -> () |
      u::us ->
         (Hashtbl.add color u (-1); init us) in
   let rec dfs_visit u =
      (Hashtbl.replace color u 0;
      time := !time + 1;
      Hashtbl.replace d u !time;
      List.iter
         (fun v -> if (Hashtbl.find color v) < 0 then dfs_visit v)
         (try adjacent_vertices g u with Not_found -> []);
      Hashtbl.replace color u 1;
      time := !time + 1;
      Hashtbl.replace f u !time) in
   let rec do_dfs = function
      [] -> () |
      u::us ->
         if (Hashtbl.find color u) < 0 then
            (dfs_visit u; do_dfs us)
         else
            do_dfs us in
   let rec out r = function
      [] -> r |
      u::us -> 
         let c' = Hashtbl.find color u 
         and d' = Hashtbl.find d u
         and f' = Hashtbl.find f u in
            out ((u, c', d', f') :: r) us
   in
      (init v_set;
      do_dfs v_set;
      out [] v_set)

let is_empty g =
   0 == Hashtbl.fold (fun v u n -> n+1) g 0

let _top g =
   let v_set = vertices g in
   let color = Hashtbl.create (List.length v_set) in
   let f = Hashtbl.create (List.length v_set) in
   let time = ref 0 in
   let rec init = function
      [] -> () |
      u::us ->
         (Hashtbl.add color u (-1); init us) in
   let rec dfs_visit u =
      (Hashtbl.replace color u 0;
      time := !time + 1;
      List.iter
         (fun v -> if (Hashtbl.find color v) < 0 then dfs_visit v)
         (try adjacent_vertices g u with Not_found -> []);
      Hashtbl.replace color u 1;
      time := !time + 1;
      Hashtbl.replace f !time u) in
   let rec do_dfs = function
      [] -> () |
      u::us ->
         if (Hashtbl.find color u) < 0 then
            (dfs_visit u; do_dfs us)
         else
            do_dfs us in
   let rec out l = 
      (* result list contains tuples (u,f(u)) where u is a vertex and
         f(u) is the finishing time of the depth-first search *)
      Hashtbl.fold (fun x u l' -> (u,x) :: l') f l
   in
      (init v_set;
      do_dfs v_set;
      List.sort (fun (u,x) (v,y) -> Pervasives.compare y x) (out []) )

let topological_sort g = fst (List.split (_top g))

let _scc g vs =
   let color = Hashtbl.create (List.length vs) in
   let rec init = function
      [] -> () |
      u::us ->
         (Hashtbl.add color u (-1); init us) in
   let rec dfs_visit visited = function
      [] -> visited |
      u::q ->
         if (Hashtbl.find color u) >= 0 then
            dfs_visit visited q
         else
            let adj = try adjacent_vertices g u with Not_found -> []
            in
               Hashtbl.replace color u 1;
               dfs_visit (u :: visited) (adj @ q) in
   let rec do_dfs comp = function
      [] -> comp |
      u::us ->
         if (Hashtbl.find color u) < 0 then
            do_dfs ((dfs_visit [] [u])::comp) us
         else
            do_dfs comp us
   in
      (init vs;
      do_dfs [] vs)

let scc g = _scc (transpose g) (topological_sort g)   

let component_graph g =
   let components = scc g in
   let g' = empty () in
   let comp = Hashtbl.create (List.length components) in
   let rec init c = function
      [] -> () |
      u::us -> Hashtbl.add comp u c; init c us in
      (* shrinking over the set of vertices *)
   let rec shrink1 = function
      [] -> () |
      u::us ->
         let adj_vertices = adjacent_vertices g u in
         let c = Hashtbl.find comp u in
            _add_vertex g' c;
            List.iter
               (fun x ->
                  let c' = Hashtbl.find comp x in
                  if not (List.mem u c') then
                     _add_edge g' c c')
               adj_vertices;
            shrink1 us in
      (* shrinking over the set of edges *)
   let rec shrink = function
      [] -> () |
      (u,v)::es -> 
         let c = Hashtbl.find comp u
         and c' = Hashtbl.find comp v in
            if not (List.mem u c') then
               _add_edge g' c c';
         shrink es 
   in 
      List.iter (fun x -> init x x) components;
(*       shrink (edges g); g' *)
      shrink1 (vertices g); g'
