(** Directed graphs.

    This module implements directed graphs (digraphs). A directed graph is
    a set {i V} of vertices and a set {i E} of ordered pairs ({i a},{i b})
    called edges, where {i a}, {i b} are in {i V}. The vertex {i a} is 
    called the initial vertex of the edge and {i b} the terminal vertex.
*)

type 'a t
   (** The type of graphs for vertices of type ['a]. *)

val empty : unit -> 'a t
   (** [empty ()] returns an empty graph. *)

val copy : 'a t -> 'a t
   (** Return a copy of the given graph. *)

val add_edge : 'a t -> 'a -> 'a -> 'a t
   (** [add_edge g x y] adds to graph [g] the edge from [x] to [y],
       i.e. the pair ({i x},{i y}) is added to the set {i E} of edges. *)

val add_vertex : 'a t -> 'a -> 'a t
   (** [add_vertex g v] adds vertex {i v} to graph [g]. This function
       is useful to add a vertex that is not connected with other vertices
       of the graph. *)

val add_adjacency_list : 'a t -> 'a -> 'a list -> 'a t
   (** [add_adjacency_list g v l] adds for all vertices {i u} in [l]
       an edge ({i v},{i u}) to graph [g]. *)

val from_list : 'a t -> ('a * 'a) list -> 'a t
   (** [from_list g l] adds to graph [g] all edges that are contained
       in list [l]. *)

val adjacent_vertices : 'a t -> 'a -> 'a list
   (** [adjacent_vertices g v] returns all vertices {i u} such that
       there exists an edge ({i v},{i u}) in [g]. *)

val dfs1 : 'a t -> ('a * int * int * int) list

val is_empty : 'a t -> bool
   (** Returns [true] if and only if the set of vertices of the given
       graph is empty. *)

val topological_sort : 'a t -> 'a list
   (** [top_sort g] returns a topological sort of graph [g] such that
       if [g] contains an edge ({i u},{i v}), then {i u} appears
       before {i v} in the ordering. *)

val scc : 'a t -> ('a list) list
   (** [scc g] returns the strongly connected components of [g]. A
       {e strongly connected component} of a directed graph 
       {i G}=({i V},{i E}) is a maximal set of vertices {i U}
       ({i U} is a subset of {i V}) such that for every pair of
       vertices {i v} and {i u} in {i U}, there is a path from
       {i u} to {i v} and vice versa; that is, vertices {i u} and 
       {i v} are reachable from each other. *)

val component_graph : 'a t -> ('a list) t
   (** [component_graph G] returns the component graph {b SCC}({i G})
       obtained by shrinking each strongly connected component of {i G}
       to a single vertex. *)

val reachable : 'a t -> 'a -> 'a list
   (** [reachable g v] returns all vertices {i u} that are reachable
       from vertex [v], i.e. for all {i u} of the result there 
       exists a path from {i v} to {i u}. *)

val out_degree : 'a t -> 'a -> int
   (** [out_degree g x] returns the number of edges with [x] as their
       initial vertex. *)

val in_degree :  'a t -> 'a -> int
   (** [in_degree g x] returns the number of edges with [x] as their
       terminal vertex. *)

val edges : 'a t -> ('a * 'a) list
   (** [edges g] returns the set {i E} of edges of [g] as a list of 
       pairs. *)

val vertices : 'a t -> 'a list
   (** [vertices g] returns the set {i V} of vertices of [g] as a 
       list. *)

val transpose : 'a t -> 'a t
   (** [transpose g] returns the transpose of [g].
       For a graph {i G}=({i V},{i E}) the transpose of {i G} is defined
       as {b T}({i G}) = ({i V},{b T}({i E})), where
       {b T}({i E}) contains all edges ({i u},{i v}) such that
       ({i v},{i u}) are in {i E}. *)
