(* main routine *)

module B = Bottom_up_automaton
open Error

let equals x y = String.compare x y == 0
let to_string x = x
let of_string x = x

type action = Help | Top_Down | No_Action | Test
type data = 
   {
      mutable cmd_name : string;
      mutable mtt : string;
      mutable tl : string;
      mutable input : string;
      mutable output : string;
      mutable prefix : string;
      mutable action : action;
      mutable debug : bool
   }


let data = 
   {
      cmd_name = "";
      mtt = "";
      tl = "";
      input = "";
      output = "";
      prefix = ".";
      action = No_Action;
      debug = false
   }

let with_tl = ref false
let with_mtt = ref false

let index = ref (-1)
let reset_index () = index := (-1)
let set_index i = index := i


let translate rules f =
   List.map (fun (symbol,args,res) -> 
      (* let _ = Printf.printf "translating symbol >%s< into " (symbol) in *)
      let symbol' =
         try f symbol
         with Not_found -> 
            (* Printf.printf "(not found) ";*)
            index := !index + 1; !index
      in
      (* let _ = Printf.printf ">%i< \n" (symbol') in *)
      (symbol',args,res)
   ) rules



let top_down () =
   let file = (data.prefix ^ "/" ^ (if !with_tl then data.tl else data.mtt)) in
   let mtt = if !with_tl then Read_tl.read data.debug file
             else ReadMtt.read data.debug file in
   let dummy = try Mtt.ord_of_sigma mtt "."
               with Not_found -> -42 in

   (*
   let _ = Printf.printf "sigma dictionary:\n" in
   let _ = 
      Mtt.iter_sigma (fun s i ->
         Printf.printf "%s -> %i\n" s i
      ) mtt in
   *)

   let (out_rules,out_finals,out_dict) = 
      try Read_automaton.read (data.prefix ^ "/" ^ data.output) 
      with Sys_error(e) -> output_error ("Error " ^ e ^ "\nSince you left the output type unspecified, I have to quit now.\n"); exit 0 in
   let t1 x = Mtt.ord_of_sigma mtt x in
   let t2 x = Mtt.ord_of_delta mtt x in
   let _ = set_index (SymbolTable.last_index mtt.Mtt.sigma) in
   let _ = if data.debug then Printf.printf "Last index for sigma:\t%i\n" (!index) in
   let out_rules' = translate out_rules t2 in
   let out_dfa = Bottom_up_automaton.from_list out_rules' out_finals in
   if data.debug then Printf.printf "Output automaton:\n%s\n" (Bottom_up_automaton.to_short_string out_dfa);
   Printf.printf "%s is \n" file;
   Approx.solve_mtt data.debug mtt out_dfa;
   print_newline ();
   exit 0;;


   (* undocumented test feature for new functions/models etc. *)
let test () = 
   (*
   let file = (data.prefix ^ "/" ^ data.mtt) in
   let _ = if data.debug then Printf.printf "reading mtt '%s'...\n" file in
   let mtt = ReadMtt.read data.debug file in
   let (out_rules,out_finals,out_dict) = 
      try Read_automaton.read (data.prefix ^ "/" ^ data.output) 
      with Sys_error(e) -> output_error ("Error " ^ e ^ "\nSince you left the output type unspecified, I have to quit now.\n"); exit 0 in
   let t1 x = Mtt.ord_of_sigma mtt x in
   let t2 x = Mtt.ord_of_delta mtt x in
   let _ = set_index (SymbolTable.last_index mtt.Mtt.sigma) in
   let _ = if data.debug then Printf.printf "Last index for sigma:\t%i\n" (!index) in
   let out_rules' = translate out_rules t2 in
   let out_dfa = Bottom_up_automaton.from_list out_rules' out_finals in
   if data.debug then Printf.printf "Output automaton:\n%s\n" (Bottom_up_automaton.to_short_string out_dfa);
   Test_main.test data.debug mtt out_dfa file;
   exit 0;;
   *)
   let file = (data.prefix ^ "/" ^ data.tl) in
   let tl = Read_tl.read false file in
   exit 0;;



let help () =
Printf.printf "
Type Checking XML Transformations with Macro Tree Transducers

   %s <action> [-in <file>] -out <file> -mtt <file>    or
   %s <action> [-in <file>] -out <file> -tl <file>

where <action> is one of

   -top                   - type check the transformation by top-down
                            solving (requires -out switch)

   -help                  - print this help message

and the other options are

   -prefix <dir>          - specifies a directory, where the input and 
                            output type and the mtt can be found

   -in <file>             - file name of the input type given as dfa
                            (can be abbreviated as -i)

   -out <file>            - file name of the output type given as dfa
                            (can be abbreviated as -o)

   -mtt <file>            - file name of the mtt definition
                            (can be abbreviated as -m)

   -tl <file>             - file name of the tl program
                            (can be abbreviated as -t)

" data.cmd_name data.cmd_name


let doit () =
   match data.action with
      | Top_Down -> top_down ()
      | Test -> test ()
      | Help -> help ()
      | No_Action -> ()


let rec read_args = function
   | "-in" :: name :: rest -> (data.input <- name; read_args rest)
   | "-i" :: name :: rest -> (data.input <- name; read_args rest)
   | "-out" :: name :: rest -> (data.output <- name; read_args rest)
   | "-o" :: name :: rest -> (data.output <- name; read_args rest)
   | "-prefix" :: name :: rest -> (data.prefix <- name; read_args rest)
   | "-mtt" :: name :: rest -> (data.mtt <- name; 
                                if !with_tl
                                then (Printf.printf "Specify either -mtt or -tl!\n\n"; data.action <- Help) 
                                else (with_mtt := true; read_args rest))
   | "-m" :: name :: rest -> (data.mtt <- name;
                                if !with_tl
                                then (Printf.printf "Specify either -mtt or -tl!\n\n"; data.action <- Help) 
                                else (with_mtt := true; read_args rest))
   | "-tl" :: name :: rest -> (data.tl <- name; 
                               if !with_mtt 
                               then (Printf.printf "Specify either -mtt or -tl!\n\n"; data.action <- Help) 
                               else (with_tl := true; read_args rest))
   | "-t" :: name :: rest -> (data.tl <- name;
                               if !with_mtt 
                               then (Printf.printf "\n"; data.action <- Help) 
                               else (with_tl := true; read_args rest))
   | "-debug" :: rest -> (data.debug <- true; read_args rest)
   | "-top" :: rest -> (data.action <- Top_Down; read_args rest)
   | "-test" :: rest -> (data.action <- Test; read_args rest)
   | "-help" :: rest -> (data.action <- Help; read_args rest)
   | [] -> ()
   | _ -> data.action <- Help;;


let main () =
   let arg_list = Array.to_list Sys.argv in
   let _ = data.cmd_name <- (List.hd arg_list) in
   let _ = read_args (List.tl arg_list) in
   doit ();;

main ();;
