module T = SymbolTable
module R = RightHandSide
module H = Hashtbl

type symbol = int
type state = int

type rhs = (symbol, state) R.rhs

type action =
   Undefined |
   Det of rhs | 
   Non_det of rhs list

   (* defines the domain of a state, i.e. all symbols for which a state
    * is defined;
    * None    - no symbols at all,
    * Any     - the rule is defined for all symbols,
    * NoneOf  - the rule is defined for all symbols except the listed ones,
    * OneOf   - the rule is defined only for the listed symbols. *)
type domain =
   None |
   Any |
   NoneOf of symbol list |
   OneOf of symbol list

type rule = symbol -> action
(* type rule_set = (state,(symbol list * rule)) H.t *)
type rule_set = (state,(domain * rule)) H.t
type parameter_table = (state,int) H.t

type ('a,'b,'c) mtt = 
   {
      mutable q : 'a T.symbol_table;
      mutable pars : parameter_table;
      mutable sigma : 'b T.symbol_table;
      mutable delta : 'c T.symbol_table;
      mutable init : state list;
      mutable rules : rule_set;
      mutable sigma_arity : (symbol,int) Simple_table.table
   }

let undef = function x -> Undefined
let dc = -1

let empty (f1,g1,h1) (f2,g2,h2) (f3,g3,h3) =
   {
      q = T.create (f1,g1,h1);
      pars = H.create 101;
      sigma = T.create (f2,g2,h2);
      delta = T.create (f3,g3,h3);
      init = [];
      rules = H.create 101;
      sigma_arity = Simple_table.create ()
   }

let add_state mtt state = T.add_symbol mtt.q state

let add_sigma mtt symbol arity = 
   let i = T.add mtt.sigma symbol in
   try
      Simple_table.add mtt.sigma_arity (i,arity)
   with
      Simple_table.First_entry_exists ->
         let arity' = Simple_table.lookup mtt.sigma_arity i in
         if not (arity = arity') then (
            Printf.printf "Error: Symbol '%s' should have rank %i, but is here used with rank %i\n" 
            ((T.to_string mtt.sigma) symbol) arity' arity; raise Exit
         ) 


let add_delta mtt symbol = 
   (* the new version operates only with one single alphabet sigma *)
   T.add_symbol mtt.sigma symbol

let add_rule mtt state symbol rhs = 
   let (domain,rule) = 
      try
         (H.find mtt.rules state)
      with
         Not_found -> (OneOf [],undef) in
   let action = rule symbol in

   let domain' =
      if (symbol < 0) then (
(*          Printf.printf "\t** adding rule with excl. symbol     %3d\n" symbol; *)
         match domain with
            | None -> OneOf [symbol]
            | Any -> OneOf [symbol]
            | NoneOf l -> NoneOf l
            | OneOf l -> if (List.mem symbol l) then
                            OneOf l
                         else
                            OneOf (symbol :: l) )
      else if (symbol = 0) then (
(*          Printf.printf "\t** adding rule with arbitrary symbol %3d\n" symbol; *)
         match domain with
            | None -> Any
            | Any -> Any
            | NoneOf l -> if (List.mem symbol l) then 
                             NoneOf l 
                          else 
                             NoneOf (symbol :: l)
            | OneOf l -> if (List.mem symbol l) then
                             OneOf l 
                          else 
                             OneOf (symbol :: l) )
      else (
(*          Printf.printf "\t** adding rule with symbol           %3d\n" symbol; *)
         match domain with
            | None -> OneOf [symbol]
            | Any -> OneOf [symbol]
            | NoneOf l -> NoneOf l
            | OneOf l -> if (List.mem symbol l) then 
                            OneOf l
                         else
                            OneOf (symbol :: l) ) in

   let new_ac = 
      match action with
         Undefined -> Det rhs |
         Det rhs' -> Non_det [rhs;rhs'] |
         Non_det l -> Non_det (rhs :: l) in
   let rule' = (fun x -> if x=symbol then new_ac else rule x)
   in
      H.replace mtt.rules state (domain',rule')

let get_action mtt state symbol =
   try 
      let (dom,rule) = H.find mtt.rules state in
         rule symbol 
   with 
      Not_found -> Undefined

let get_all_actions mtt state =
   try
      let (dom',rule) = H.find mtt.rules state in
      let dom = 
         match dom' with
            OneOf l -> l |
            NoneOf l -> l |
            Any -> [0] |
            None -> [] in
      let res = ref [] in
         List.iter (fun sym -> res := (rule sym) :: (!res)) dom; (!res)
   with
      Not_found -> []

let arity mtt sigma = Simple_table.lookup mtt.sigma_arity sigma

let get_domain mtt state =
   try
      let (dom,rule) = H.find mtt.rules state in
         dom
   with
      Not_found -> None (*OneOf []*)

let iter f mtt =
   H.iter (fun state (dom',rule) ->
      let dom =
         match dom' with
            OneOf l -> l |
            NoneOf l -> l |
            Any -> [0] |
            None -> []  in
      List.iter (fun sym ->
         let action = rule sym in
         f (state,sym,action)
      ) dom
   ) mtt.rules 

let iter_states f mtt = T.iter f mtt.q
let iter_sigma f mtt = T.iter f mtt.sigma
let iter_delta f mtt = 
   (* the new version operates only with one single alphabet sigma *)
   T.iter f mtt.sigma

let num_of_parameters mtt state = H.find mtt.pars state
let set_parameters mtt state k = H.add mtt.pars state k
let change_parameters mtt state k = H.replace mtt.pars state k

let ord_of_state mtt state = T.get_index mtt.q state
let ord_of_sigma mtt sigma = T.get_index mtt.sigma sigma
let ord_of_delta mtt delta = 
   (* T.get_index mtt.delta delta *)
   (* the new version operates only with one single alphabet sigma *)
   T.get_index mtt.sigma delta
let state_of_ord mtt ord = T.get_symbol mtt.q ord
let sigma_of_ord mtt ord = T.get_symbol mtt.sigma ord
let delta_of_ord mtt ord = 
   (* T.get_symbol mtt.delta ord *)
   (* the new version operates only with one single alphabet sigma *)
   T.get_symbol mtt.sigma ord 

let get_initial_state mtt = mtt.init
let get_init = get_initial_state
let set_initial_state mtt q0 = 
   let init = mtt.init in
   if not (List.mem q0 init) then mtt.init <- (q0 :: init)
let set_init = set_initial_state

let state_eq mtt = T.eq mtt.q
let sigma_eq mtt = T.eq mtt.sigma
   (* the new version operates only with one single alphabet sigma *)
let delta_eq mtt = T.eq mtt.sigma

let state_to_string mtt = T.to_string mtt.q
let sigma_to_string mtt = T.to_string mtt.sigma
   (* the new version operates only with one single alphabet sigma *)
let delta_to_string mtt = T.to_string mtt.sigma

let state_from_string mtt = T.from_string mtt.q
let sigma_from_string mtt = T.from_string mtt.sigma
   (* the new version operates only with one single alphabet sigma *)
let delta_from_string mtt = T.from_string mtt.sigma

let print_states mtt = 
   T.iter (fun x y -> Printf.printf "%s -> %i " ((state_to_string mtt) x) y)
   mtt.q; Printf.printf "\n"
let print_sigma mtt = 
   T.iter (fun x y -> Printf.printf "%s -> %i " ((sigma_to_string mtt) x) y)
   mtt.sigma; Printf.printf "\n"
let print_delta mtt = 
   (* the new version operates only with one single alphabet sigma *)
   T.iter (fun x y -> Printf.printf "%s -> %i " ((delta_to_string mtt) x) y)
   mtt.sigma; Printf.printf "\n"

