#include <stdio.h>
#include <libxml/tree.h>
#include <libxml/parser.h>
#include <caml/mlvalues.h>
#include <caml/alloc.h>
#include <caml/memory.h>
#include <caml/callback.h>

value inspect (value v) {
   if (Is_long(v))
      printf("v is an integer (%ld) : %ld", (long) v, Long_val(v));
   else if (Is_block(v))
      printf("v is a pointer");
   else
      printf("v is neither int nor pointer");
   printf(" ");
   fflush(stdout);
   return v;
}

void margin (int n) {
   while (n-- > 0) printf(".");
   return;
}

void print_block (value v, int m) {
   int size, i;
   margin(m);
   if (Is_long(v)) {
      printf("immediate val (%d)\n", Long_val(v));
      return;
   }
   printf("mem block: size=%d - ", size=Wosize_val(v));
   switch (Tag_val(v)) {
      case Closure_tag :
         printf(" closure with %d free vars\n", size-1);
         margin(m+4);
         printf("code pointer: %p\n", Code_val(v));
         for(i=1;i<size;i++)
            print_block(Field(v,i), m+4);
         break;
      case String_tag :
         printf("string: %s (%s)\n", String_val(v),(char *) v);
         break;
      case Double_tag :
         printf("float: %g\n", Double_val(v));
         break;
      case Double_array_tag :
         printf("float array: ");
         for(i=0;i<size/Double_wosize;i++)
            printf("   %g", Double_field(v,i));
         printf("\n");
         break;
      case Abstract_tag :
         printf("abstract type\n");
         break;
      case Custom_tag :
         printf("custom block\n");
         break;
      default :
         if (Tag_val(v)>=No_scan_tag) {
            printf("unknown tag"); break;
         }
         printf("structured block (tag-%d):\n", Tag_val(v));
         for(i=0;i<size;i++)
            print_block(Field(v,i), m+4);
   }
   return;
}

value inspect_block (value v) {
   print_block(v,4);
   fflush(stdout);
   return v;
}

void print_elements(xmlNode *node) {
	xmlNode *current=NULL;

	for(current=node; current; current = current->next) {
		if(current->type == XML_ELEMENT_NODE)
			printf("%s\n", current->name);
		print_elements(current->children);
	}
}

/* given an xmlNode, create_forest returns an O'Caml structure
 * representing the forest. The forest is build up by successively
 * processing the next pointer. */
value create_forest(xmlNode *node) {
	if(node==NULL)
		return Val_long(0);

	/* if the current node is an empty or whitespace only
	 * text-node it is ignorable */
	if(xmlIsBlankNode(node))
		return create_forest(node->next);

	value tree = alloc(2,0);
	Store_field(tree,0,create_tree(node));
	Store_field(tree,1,create_forest(node->next));

	return tree;
}

/* given an xmlNode, create_tree returns an O'Caml structure
 * representing the tree */
/* defining create_tree as value causes a warning:
 * type mismatch with previous implicit declaration
 * privious implicit declaration of `create_tree'
 * `create_tree' was previously implicitly declared to return `int' */ 
/* value create_tree(xmlNode *node) { */
int create_tree(xmlNode *node) {
   value tree = alloc(2,0);
	
	/* closure_encode is a pointer to an O'Caml function
	 * from XML labels to an O'Caml structure */
	static value* closure_encode = NULL;
	if(closure_encode == NULL) {
		/* first time around, look up by name */
	   closure_encode = caml_named_value("encoding");
   }

	Store_field(tree,0,callback(*closure_encode, copy_string(node->name)));
	Store_field(tree,1,create_forest(node->children));

	return tree;
}

/* f_c reads an XML file with filename v and returns an O'Caml
 * structure representing the structure of the XML document. */
value f_c(value v) {
	xmlDocPtr doc=NULL;
	xmlNode *root=NULL;
	
	char *filename = String_val(v);
	/* returns the resulting document tree */
	doc = xmlParseFile(filename);
	if (doc == NULL) {
		fprintf(stderr, "Error: Failed to parse %s\n", filename);
		return;
	}

	root = xmlDocGetRootElement(doc);
   value tree = create_tree(root);

	xmlFreeDoc(doc);
   return tree;
}

