/* File parser.mly */
/*

*/

%{

   open ParserTypes
   open RightHandSide
   open Mtt

   module H = Hashtbl

   exception Error_in_rhs

   let rec sym_to_string = function
      | Any_sym -> "."
      | Expl_sym s -> s
      | Excl_sym s -> "~"^s

   let rec list_to_string = function
      [] -> "" |
      x::[] -> x |
      x::l -> x ^ "," ^ list_to_string l

   let rec to_string = function
      Sym s -> s |
      App (s,l) -> s ^ "(" ^ (do_list l) ^ ")" |
      Any_rhs l -> ".(" ^ (do_list l) ^ ")" 
   and do_list = function
      [] -> "" |
      x::[] -> to_string x |
      x::l -> (to_string x) ^ "," ^ do_list l

   let rec rhs_to_string = function
      Empty_rhs -> "" |
      Parameter i -> "Par_"^ string_of_int i |
      Symbol (s,k,l) -> "Sym_"^ s ^ "(" ^ rhs_list l ^ ")" |
      Call (q,i,k,l) -> "Stat_"^q^"(x"^(string_of_int i)^","^rhs_list l ^ ")"
   and rhs_list = function
      [] -> "" |
      x::[] -> rhs_to_string x |
      x::l -> (rhs_to_string x) ^ "," ^ rhs_list l
   

   let cond_to_string = function
      Empty_cond -> "" |
      Opt (s,l) -> s ^ " in " ^ list_to_string l |
      Excl (s,l) -> s ^ " not in " ^ list_to_string l

%}
%token <string> ID
%token LPAREN RPAREN
%token COMMA
%token ARROW
%token WHERE
%token IN NOT
%token EOL
%token TILDE
%token ANY
%left COMMA
%left ARROW
%start main
%type <ParserTypes.parsed_main> main
%type <parsed_rhs> rhs
%type <parsed_vars> varseq
%type <parsed_pars> parseq
%type <parsed_cond> cond
%type <parsed_input> input
%type <parsed_expr> expr
%type <parsed_lhs> lhs
%type <parsed_rule> rule
%type <parsed_sym> symbol
%%
main:
        rule EOL                    { let (state,sym,vars,pars,rhs,str) = $1 in
                                      let cond = Empty_cond in
                                         (state,sym,vars,pars,rhs,cond,str) }
      | rule cond EOL               { let (state,sym,vars,pars,rhs,str) = $1 in
                                      let cond = $2 in
                                      let str' = 
                                         str ^ " [" 
                                         ^ (cond_to_string $2) ^ "]"
                                      in
                                         (state,sym,vars,pars,rhs,cond,str') }
;
rule:
        lhs ARROW rhs               { let (state,sym,vars,pars) = $1 in
                                      let rhs = $3 in
                                      (* let rhs' = mk_rhs rhs in *)
                                      let str = state ^ "(" ^ 
                                         (sym_to_string sym)
                                         ^"("^(list_to_string vars)^")"
                                         ^ ";" ^ (list_to_string pars)
                                         ^ ")" ^
                                         " => " ^ (to_string rhs) 
                                      in
                                         (state,sym,vars,pars,rhs,str) }
;
lhs:
        state LPAREN expr RPAREN    { let (sym,vars,pars) = $3 in
                                      let state = $1 in
                                         (state,sym,vars,pars) }
;
expr:
        input COMMA parseq          { let (sym,vars) = $1 in 
                                      let pars = List.rev $3 in
                                         (sym,vars,pars) }
      | input                       { let (sym,vars) = $1 in (sym,vars,[]) }
;
input:
        symbol LPAREN varseq RPAREN { ($1,List.rev $3) }
      | symbol LPAREN RPAREN        { ($1,[]) }
      | symbol                      { ($1,[]) }
;
parseq:
        parseq COMMA parameter      { $3 :: $1 }
      | parameter                   { [$1] }
;
varseq:
        varseq COMMA variable       { $3 :: $1 }
      | variable                    { [$1] }
;
rhs:
        ident LPAREN rhsseq RPAREN  { App ($1,$3) }
      | ANY LPAREN rhsseq RPAREN    { Any_rhs $3 }
      | ident LPAREN RPAREN         { Sym $1 }
      | ANY LPAREN RPAREN           { Any_rhs [] }
      | ident                       { Sym $1 }
      | ANY                         { Any_rhs [] }
;
rhsseq:
        rhsseq COMMA rhs            { $1 @ [$3] }
      | rhs                         { [$1] }
;
cond:
        WHERE ident IN symseq       { Opt ($2,$4) }
      | WHERE ident NOT IN symseq   { Excl ($2,$5) }
;
symseq:
        symb COMMA symseq           { $1 :: $3 }
      | symb                        { [$1] }
;
state:
        ID                          { $1 }
;
symbol:
        ID                          { Expl_sym $1 }
      | TILDE ID                    { Excl_sym $2 }
      | ANY                         { Any_sym }
;
symb:
        ID                          { $1 }
;
parameter:
        ID                          { $1 }
;
variable:
        ID                          { $1 }
;
ident:
        ID                          { $1 }
;
