/* File parser_automaton.mly */

%{
   
open Lexing
module D = Dictionary


let d = D.create ()

let count = ref 0
let reset_count () = count := 0

let add item = 
   if D.mem_rev d item then
      D.lookup_rev d item
   else
      let index = !count in
      D.add d (index,item); count := index + 1; index
   
let syntax_error msg =
   let pos = Parsing.symbol_start_pos () in
      (* unfortunately, pos.pos_fname is empty *)
   let file = pos.pos_fname in
   let line = pos.pos_lnum in
   let cnum = pos.pos_cnum in
   let bol = pos.pos_bol in
   let msg' = "line "^ 
              string_of_int line ^", character "^ 
              string_of_int (cnum-bol) ^"\n Syntax Error: "^ 
              msg ^"\n" in
   raise (Error.Syntax_error msg')

let parse_error msg =
   let pos = Parsing.symbol_start_pos () in
   let line = pos.pos_lnum in
   let cnum = pos.pos_cnum in
   let bol = pos.pos_bol in
   let msg' = "line "^ 
              string_of_int line ^", character "^ 
              string_of_int (cnum-bol) ^"\n"^
              msg ^"\n" in
   raise (Error.Parse_error msg')

let warning msg =
   let pos = Parsing.symbol_start_pos () in
   let line = pos.pos_lnum in
   let cnum = pos.pos_cnum in
   let bol = pos.pos_bol in
   let msg' = "Warning: line "^ 
              string_of_int line ^", character "^ 
              string_of_int (cnum-bol) ^": "^
              msg ^"\n" in
   print_string msg'; flush stdout

%}
%token <string> ID
%token SETBEGIN SETEND
%token COMMA
%token ENDRULE
%token ENDRULES
%token EOL
%token EOF
%left COMMA
%start dfa
%type <(string * int list * int list) list * int list * (int,string) Dictionary.dictionary> dfa
%%
dfa:
        rules ENDRULES finals fin                { 
                                                   let _ = reset_count () in
                                                   ((List.rev $1) @ $4,List.rev $3, d) 
                                                 }
      | rules ENDRULES fin                       { 
                                                   let _ = reset_count () in
                                                   ((List.rev $1) @ $3,[], d) 
                                                 }
      | fin                                      { 
                                                   let _ = reset_count () in
                                                   ($1,[], d) 
                                                 }
;
finals:
        finals COMMA id                          { $3 :: $1 }
      | finals COMMA error                       { syntax_error "This is not a legal name." }
      | id                                       { $1 :: [] }
;
fin:
        EOF                                      { [] }
;
rules:
        rules rule                               { 
                                                   let (sym,res,arg) = $2 in
                                                   if (List.length res) + (List.length arg) = 0 
                                                   then $1 
                                                   else (sym,res,arg) :: $1 
                                                 }
      | rule                                     { $1 :: [] }
;
rule:
        results COMMA sym COMMA args ENDRULE     { ($3,List.rev $5,$1) }
      | results COMMA sym ENDRULE                { ($3,[],$1) }
      | results COMMA error                      { syntax_error "Corrupt transition." }
      | ENDRULE                                  { 
                                                   warning "Empty rule ignored"; 
                                                   ("",[],[]) 
                                                 }
;
results:
        id                                       { $1 :: [] }
      | SETBEGIN set SETEND                      { List.rev $2 }
      | SETBEGIN set error                       { syntax_error "Unclosed set.  Maybe you have forgotten a closing '}'" }
;
set:
        set COMMA id                             { $3 :: $1 }
      | set COMMA error                          { syntax_error "This is not a legal name." }
      | id                                       { $1 :: [] }
;
sym:
        ID                                       { $1 }
;
args:
        args COMMA id                            { $3 :: $1 }
      | args COMMA error                         { syntax_error "This is not a legal name." }
      | id                                       { $1 :: [] }
;
id:
        ID                                       { let i = add $1 in i }
;

