
open RightHandSide

exception Undefined
exception Undefined_rhs
exception Non_deterministic_mtt

(* ------------ debug functions --------------------------- *)

let rec rhs_to_string mtt = function
   Empty_rhs -> "nil" |
   Parameter i -> "y_"^(string_of_int i) |
   Symbol (s,k,succ) ->
      Mtt.delta_to_string mtt (Mtt.delta_of_ord mtt s)^
      "("^(rhs_list mtt succ)^")" |
   Call (q,s,k,succ) ->
      Mtt.state_to_string mtt (Mtt.state_of_ord mtt q)^
      "(x_"^(string_of_int s)^
      (if k>0 then ","^(rhs_list mtt succ) else "")^")"
and rhs_list mtt = function
   [] -> "" |
   r::[] -> rhs_to_string mtt r |
   r::rs -> (rhs_to_string mtt r)^","^(rhs_list mtt rs)

let rec pars_to_string n = function
   0 -> "" |
   1 -> "y_" ^ (string_of_int n) |
   i -> "y_" ^ (string_of_int (n-i+1)) ^ "," ^ (pars_to_string n (i-1))

let action_to_string mtt = function
   Mtt.Undefined -> "undef" |
   Mtt.Det ac -> rhs_to_string mtt ac |
   Mtt.Non_det acs -> "["^(rhs_list mtt acs)^"]"

(* ------------ functions for list operations -------------- *)

let rec single = function
   [] -> [] |
   e::es -> [e] :: single es

let rec prepend x = function
   [] -> [] |
   l::ls -> (x::l) :: prepend x ls

let rec pair l = function
   [] -> [] |
   e::es -> (prepend e l) @ pair l es

   (* [tuple n l] creates a list of all n-tuples that can be formed
      with elements of list [l] *)
let rec tuple n l =
      (* WARNING: this function produces an overflow error for the
         call [tuple 8 l]: 
         Stack overflow during evaluation (looping recursion) *)
   let rec doit = function
      0 -> [] |
      1 -> single l |
      2 -> pair (single l) l |
      n -> pair (doit (n-1)) l
   in
      doit n

let rec prepend' result l = function
   [] -> result |
   e::es -> prepend' ((e::l) :: result) l es

let rec combine result lst = function
   [] -> result |
   e::es -> combine (prepend' result [e] lst) lst es

   (* if l is a list of lists l_1,...,l_k, [tuple' l] creates a list of
      all k-tuples (e_1,...,e_k) such that e_i is element of l_i for
      all i=1,...,k *)
let tuple' lst = 
   let rec doit = function
      [] -> [] |
      lst::[] -> single lst |
      lst1::lst2::[] -> combine [] lst1 lst2 |
      lst::lsts -> pair (doit lsts) lst
   in
      List.map List.rev (doit (List.rev lst))


let max_rank mtt =
   let r = ref 0 in
   Mtt.iter_states (fun _ q -> 
      let n = Mtt.num_of_parameters mtt q in
      if n > !r then r := n
   ) mtt; !r

   (* returns a string list and a dictionary (int -> string) *)
let mk_sigma mtt sigma =
   let e = Automat.sym_to_string (Automat.empty_tree) in
   let l = ref [e] in
   Mtt.iter_sigma (fun s i -> 
      Dictionary.add sigma (i,s);
      l := (string_of_int i) :: !l
   ) mtt; !l


exception Error_in_construction

let rec mk_mapping' h = function
   ([],[]) -> () |
   (xs,[]) -> raise Error_in_construction |
   ([],ys) -> raise Error_in_construction |
   (x::xs,y::ys) -> Hashtbl.add h x y; mk_mapping' h (xs,ys)

let mk_states mtt fsta_states tuples debug =
(*    let d = Dictionary.create () in *)
   let d = Simple_table.create () in
   let index = ref 0 in
   let get () = (let r = !index in index := !index + 1; r) in
   let l = ref [] in
   let q_p_tuples =
      let l' = ref [] in
      Mtt.iter_states (fun q_sem q ->
         if List.length tuples > 0 then
            List.iter (fun t ->
               l' := (q,t) :: !l'
            ) tuples
               (* for each possiple tuple p1,..,pm of tree states *)
         else
            l' := (q,[]) :: !l'
      ) mtt; !l' in
         (* for each state q of the mtt *)
   let k = List.length q_p_tuples in
   let results = tuple k (Dictionary.to_list fsta_states) in
   let _ = if debug then
      Printf.printf "max. rank of all q: %s\n"
         (string_of_int (max_rank mtt));
      Printf.printf "number of state tuples (p1,..,pm):\t%s\n"
         (string_of_int (List.length tuples));
      Printf.printf "number of (q,p1,..,pm) tuples:\t%s\n"
         (string_of_int (List.length q_p_tuples));
      Printf.printf "number of functions f:Q x P^(m) -> P:\t%s\n"
         (string_of_int (List.length results))
   in
   List.iter (fun r ->
      let new_index = get () in
      let h = Hashtbl.create k in
      mk_mapping' h (q_p_tuples,r);
(*       Dictionary.add d (new_index, h); *)
      Simple_table.add d (new_index, h);
      l := (string_of_int new_index) :: !l
   ) results; (!l,d)


let rec contains l1 = function
   [] -> true |
   x::xs -> (List.mem x l1) && (contains l1 xs)

let get_d new_states q args ps =
(*    Dictionary.fold (fun d d_q result -> *)
   Simple_table.fold (fun d d_q result ->
      let p = Hashtbl.find d_q (q,args) in
      if List.mem p ps then d :: result else result
   ) new_states []

let compute_d new_states =
(*    let h = Hashtbl.create (Dictionary.size new_states) in *)
   let h = Hashtbl.create (Simple_table.size new_states) in
(*    Dictionary.iter (fun d d_q -> *)
   Simple_table.iter (fun d d_q ->
      Hashtbl.iter (fun (q,args) p ->
         Hashtbl.add h (q,args,p) d
      ) d_q
   ) new_states; h

let mk_final mtt fsta new_states =
   let q_init = Mtt.get_initial_state mtt in
   let finals = Automat.final_states fsta in
(*    Dictionary.fold (fun d d_q result -> *)
   Simple_table.fold (fun d d_q result ->
      let is_final = Hashtbl.fold (fun (q,l) p b ->
         b || (List.mem q q_init && List.mem p finals)
      ) d_q false
      in
      if is_final then (string_of_int d) :: result else result
   ) new_states [] 

let get_mtt_states mtt symbol =
   let l = ref [] in
   let contains sym = function
      Mtt.None -> false |
      Mtt.Any -> true |
      Mtt.NoneOf l -> not (List.mem sym l) |
      Mtt.OneOf l -> List.mem sym l in
   Mtt.iter_states (fun sem q ->
      let dom = Mtt.get_domain mtt q in
      if (contains symbol dom) then l := q :: !l
   ) mtt; !l

let get_rhs = function
   Mtt.Undefined -> raise Undefined_rhs |
   Mtt.Det rhs -> rhs |
   Mtt.Non_det _ -> raise Non_deterministic_mtt


let next_states automaton sym args debug =
   let result = ref [] in
   let rec eq' l1 l2 =
      (* equality test for two fsta state lists *)
      List.fold_left2 (fun b x y ->
         let x' = Automat.state_to_string x 
         and y' = Automat.state_to_string y in
         b && (Hashtbl.hash x' = Hashtbl.hash y')
      ) true l1 l2 in 
   let doit qs =
      Automat.states_fold (fun q l -> 
         (int_of_string (Automat.state_to_string q)) :: l
      ) qs [] 
   in
   Automat.iter_rules (fun s qs qs' ->
      if (Automat.sym_compare s sym) == 0 then (
         let _ = if debug then
            Printf.printf "\tsymbol '%s' found\n\t-> testing [%s] = [%s]?" 
            (Automat.sym_to_string s)
            (List.fold_left (fun s x -> 
               s ^ " " ^ (Automat.state_to_string x)
            ) "" qs)
            (List.fold_left (fun s x -> 
               s ^ " " ^ (Automat.state_to_string x)
            ) "" args)
         in
         if eq' qs args  then (
            let _ = if debug then 
               Printf.printf "\t...yes\n" 
            in
            result := doit qs'
         ) else (
            if debug then
               Printf.printf "\t...no\n"
         )
      ) else (
         if debug then
            Printf.printf "\tsymbol found?\t...no\n"
      )
   ) automaton; !result


let rec simul' fsta states phi rho depth debug = function
   Empty_rhs -> [] |
   Parameter i -> [ (rho i) ] |
   Call (q,j,k,rhs_list) ->
      let d = phi j in
      let states' = 
         List.map (fun rhs ->
            simul' fsta states phi rho (depth+1) debug rhs
         ) rhs_list in
      List.fold_left (fun result args ->
         (eval' states d q args) @ result
      ) [] (tuple' states') |
   Symbol (symbol,k,rhs_list) ->
      let sym = Automat.sym_from_string (string_of_int symbol) in
      let _ = if debug then
         Printf.printf "[%i] current symbol:\t%s\n"
         depth 
         (Automat.sym_to_string sym)
      in
      let states' = 
         List.map (fun rhs ->
            simul' fsta states phi rho (depth+1) debug rhs
         ) rhs_list 
      in
      if List.length rhs_list > 0 then
         List.fold_left (fun result args ->
            let args' = 
               List.map Automat.state_from_string (List.map string_of_int args) in
            let deriv_stats = next_states fsta sym args' debug in
            deriv_stats @ result
         ) [] (tuple' states')
      else
            (* current symbol is a leaf, thus it is the empty tree epsilon *)
         next_states fsta Automat.empty_tree [] false


and eval' states d q rhs_list =
(*    let h = Dictionary.lookup states d in  *)
   let h = Simple_table.lookup states d in 
   let t = (q,rhs_list) in
   let r = Hashtbl.find h t in
   [r]


let rec simul_0' mtt sigma fsta_states fsta states phi depth debug = 
function
   Empty_rhs -> [] |
   Parameter i -> [] (* no parameters defined *) |
   Call (q,j,k,rhs_list) ->
      let _ = if debug then
         Printf.printf "\t[%i] call of:\t'%s'\n"
         depth
         ((Mtt.state_to_string mtt) (Mtt.state_of_ord mtt q))
      in
      let d = phi j in
      eval_0' mtt states d q |
   Symbol (symbol,k,rhs_list) -> 
      let sym = Automat.sym_from_string (string_of_int symbol) in
      let _ = if debug then
         Printf.printf "\t[%i] output:\t'%s'='%s'\n"
         depth 
         (Automat.sym_to_string sym)
         ((Mtt.delta_to_string mtt) (Mtt.delta_of_ord mtt symbol))
      in
      let states' = 
         List.map (fun rhs ->
            simul_0' mtt sigma fsta_states fsta states phi (depth+1) debug rhs
         ) rhs_list 
      in
      if List.length rhs_list > 0 then 
         List.fold_left (fun result args ->
            let args' = 
               List.map Automat.state_from_string (List.map string_of_int args) in
            let deriv_stats = next_states fsta sym args' debug in
            let _ = if debug then
               Printf.printf "\t[%i] beta(%s,<%s>) -> [%s]\n"
               depth
               ((Mtt.delta_to_string mtt) (Mtt.delta_of_ord mtt symbol))
               (List.fold_left (fun s q -> s^","^(Automat.state_to_string q)) 
                  (try Automat.state_to_string (List.hd args') with Failure f -> "")
                  (try List.tl args' with Failure f -> []) 
               )
               (List.fold_left (fun s q -> s^","^(string_of_int q))
                  (try string_of_int (List.hd deriv_stats) with Failure f -> "")
                  (try List.tl deriv_stats with Failure f -> [])
               )
            in
            deriv_stats @ result
         ) [] (tuple' states')
      else
            (* current symbol is a leaf, thus it is the empty tree epsilon *)
         next_states fsta Automat.empty_tree [] false

and eval_0' mtt states d q =
(*    let h = Dictionary.lookup states d in  *)
   let h = Simple_table.lookup states d in 
   let r = Hashtbl.find h (q,[]) in
   [r]

let mk_rules' mtt fsta sigma delta fsta_states fsta_tuples states states' debug =
   let e_rule = 
      (Automat.sym_to_string Automat.empty_tree,[],
       List.map string_of_int (next_states fsta Automat.empty_tree [] false)) in
   let rules = ref [e_rule] in
   let rule_counter = ref 0 in
   let error_rule_counter = ref 0 in
(*    let state_tuples = tuple 2 (Dictionary.to_list states) in *)
   let state_tuples = tuple 2 (Simple_table.to_list states) in
   let _ = Printf.printf "number of state tuples: %i\n" (List.length state_tuples) in
   let k = List.length fsta_tuples in
   if k = 0 then
         (* mtt is top-down, i.e., no parameters occurr *)
      Dictionary.iter (fun symbol symbol_sem ->
         let fsta_sym = string_of_int symbol in
         let _ = if debug then
            Printf.printf "current symbol of Sigma:\t'%s'='%s'\n"
            fsta_sym
            ((Mtt.sigma_to_string mtt) symbol_sem)
         in
         let mtt_states = get_mtt_states mtt symbol in
         List.iter (fun tuple ->
            let phi = fun i -> List.nth tuple (i-1) in
            List.iter (fun q ->
               let rhs = get_rhs (Mtt.get_action mtt q symbol) in
               let _ = if debug then
                  Printf.printf "   current rhs:\t%s\n"
                  (rhs_to_string mtt rhs)
               in
               let e = 
                  simul_0' mtt sigma fsta_states fsta states phi 0 debug rhs in
               let arguments = List.map string_of_int tuple in
(*                let d = get_d states q [] e in *)
               let d = 
                  List.fold_left (fun ds p ->
                     let d' = 
                        try
                           Some (Hashtbl.find states' (q,[],p))
                        with
                           Not_found -> None
                     in
                     match d' with
                        Some (d) -> d :: ds |
                        None -> ds
                  ) [] e in
               let result_states = List.map string_of_int e in
               let result_ds = List.map string_of_int d in
               let _ = if debug then
                  Printf.printf "   ==>  delta(%s,<%s>) -> [%s]\n"
                  fsta_sym
                  (List.fold_left (fun s a -> s^","^a)
                     (try List.hd arguments with Failure f -> "")
                     (try List.tl arguments with Failure f -> [])
                  ) 
                  (List.fold_left (fun s a -> s^","^a)
                     (try List.hd result_ds with Failure f -> "")
                     (try List.tl result_ds with Failure f -> [])
                  ) 
               in
               let _ = if !rule_counter mod 50 = 0 then
                  Printf.printf "\n" 
               in
               if List.length d = 0 then (
                  (* list of result states is empty *)
                  Printf.printf "."; 
                  error_rule_counter := !error_rule_counter + 1;
                  rule_counter := !rule_counter + 1
               ) else (
                  Printf.printf "*";
                  rule_counter := !rule_counter + 1
(*                rules := (fsta_sym,arguments,result_ds) :: !rules *)
               )
            ) mtt_states
               (* state q of mtt for which q(a(x1,..,xk),y1,..,ym) -> T *)
         ) state_tuples 
            (* d1,..,dk for a k-ary symbol of the alphabet *)
      ) sigma 
   else
         (* mtt has parameters *)
      Dictionary.iter (fun symbol symbol_sem ->
         let fsta_sym = string_of_int symbol in
         let mtt_states = get_mtt_states mtt symbol in
         List.iter (fun tuple ->
            let phi = fun i -> List.nth tuple (i-1) in
            List.iter (fun q ->
               let rhs = get_rhs (Mtt.get_action mtt q symbol) in
               List.iter (fun args ->
                  let rho = fun j -> List.nth args (j-1) in
                  let e = simul' fsta states phi rho 0 false rhs in
(*                   let d = get_d states q args e in *)
                  let d = 
                     List.fold_left (fun ds p ->
                        let d' = 
                           try
                              Some (Hashtbl.find states' (q,args,p))
                           with
                              Not_found -> None
                        in
                        match d' with
                           Some (d) -> d :: ds |
                           None -> ds
                     ) [] e in
                  let _ = if !rule_counter mod 50 = 0 then
                     Printf.printf "\n" 
                  in
                  if List.length d = 0 then (
                     (* list of result states is empty *)
                     Printf.printf "."; 
                     error_rule_counter := !error_rule_counter + 1;
                     rule_counter := !rule_counter + 1
                  ) else (
                     Printf.printf "*";
                     rule_counter := !rule_counter + 1
(*                   rules := (fsta_sym,List.map string_of_int tuple,List.map *)
(*                   string_of_int d) :: !rules *)
                  )
               ) fsta_tuples
                  (* p_1,...,p_m for a m-ary state q of the mtt *)
            ) mtt_states
               (* state q of mtt for which q(a(x1,..,xk),y1,..,ym) -> T *)
         ) state_tuples 
            (* d_1,...,d_k for a k-ary symbol of the alphabet *)
      ) sigma; 
      
      (* symbol a of alphabet *)
(*    !rules *)
   (!rule_counter,!error_rule_counter)




let pre mtt fsta delta fsta_states debug =
   let _ = if debug then 
      Printf.printf "mtt:\n";
      Mtt.iter (fun (q,s,a) -> 
         let state = Mtt.state_to_string mtt (Mtt.state_of_ord mtt q) in
         let sym = Mtt.sigma_to_string mtt (Mtt.sigma_of_ord mtt s) in
         let n = Mtt.num_of_parameters mtt q in
         let pars = pars_to_string n n in
         let act = action_to_string mtt a in
         if n = 0 then
            Printf.printf "%s(%s) -> %s\n" state sym act
         else
            Printf.printf "%s(%s,%s) -> %s\n" state sym pars act
      ) mtt in
   let _ = if debug then 
      Printf.printf "\n\nfsta:\n%s\n\n" (Automat.to_string fsta) in
   let sigma = Dictionary.create () in
   let sigma_list = mk_sigma mtt sigma in
   let rank = max_rank mtt in
      (* tuples : (int list) list *)
   let tuples = tuple rank (Dictionary.to_list fsta_states) in
   let (state_list,d) = mk_states mtt fsta_states tuples debug in
   let d' = compute_d d in
   let _ = if debug then
      Printf.printf "number of computed states:\t%i\n" 
      (List.length state_list) in
   let _ = if debug then 
      Printf.printf "states of the new automaton:\n" in
   let _ = if not debug then Simple_table.iter (fun i e ->
       Printf.printf "%i ->\n" i;
       Hashtbl.iter (fun (q,l) p ->
          Printf.printf "\t(%s%s) -> %s\n"
          ((Mtt.state_to_string mtt) (Mtt.state_of_ord mtt q))
          (if List.length l > 0 then List.fold_left (fun r a ->
             r ^ "," ^ (string_of_int a)
          ) (", "^(string_of_int (List.hd l))) (List.tl l) else "")
          (string_of_int p)
       ) e;
       Printf.printf "\n"
   ) d in
   let (rules',errors) = 
      mk_rules' mtt fsta sigma delta fsta_states tuples d d' false in
   let _ = Printf.printf "\n\nnumber of rules: %i\n" rules' in
   let _ = Printf.printf "number of error rules: %i\n\n" errors in
(*    let finals = mk_final mtt fsta d in *)
   let a = Automat.from_list sigma_list state_list [] (*finals*) [] (*rules'*) in
   (a, sigma, d)
