
open RightHandSide
module Automat = Bottom_up_automaton

exception Undefined
exception Undefined_rhs
exception Non_deterministic_mtt


let rec single = function
   [] -> [] |
   e::es -> [e] :: single es

let rec prepend x = function
   [] -> [] |
   l::ls -> (x::l) :: prepend x ls

let rec pair l = function
   [] -> [] |
   e::es -> (prepend e l) @ pair l es

   (* [tuple n l] creates a list of all n-tuples that can be formed
      with elements of list [l] *)
let rec tuple n l =
      (* WARNING: this function produces an overflow error for the
         call [tuple 8 l]: 
         Stack overflow during evaluation (looping recursion) *)
   let rec doit = function
      0 -> [] |
      1 -> single l |
      2 -> pair (single l) l |
      n -> pair (doit (n-1)) l
   in
      doit n

let rec prepend' result l = function
   [] -> result |
   e::es -> prepend' ((e::l) :: result) l es

let rec combine result lst = function
   [] -> result |
   e::es -> combine (prepend' result [e] lst) lst es

   (* if l is a list of lists l_1,...,l_k, [tuple' l] creates a list of
      all k-tuples (e_1,...,e_k) such that e_i is element of l_i for
      all i=1,...,k *)
let tuple' lst = 
   let rec doit = function
      [] -> [] |
      lst::[] -> single lst |
      lst1::lst2::[] -> combine [] lst1 lst2 |
      lst::lsts -> pair (doit lsts) lst
   in
      List.map List.rev (doit (List.rev lst))


let max_rank mtt =
   let r = ref 0 in
   Mtt.iter_states (fun _ q -> 
      let n = Mtt.num_of_parameters mtt q in
      if n > !r then r := n
   ) mtt; !r

   (* returns a string list and a dictionary (int -> string) *)
let mk_sigma mtt sigma =
(*    let e = Automat.sym_to_string (Automat.empty_tree) in *)
   let l = ref [] in
   Mtt.iter_sigma (fun s i -> 
      Dictionary.add sigma (i,s);
      l := (string_of_int i) :: !l
   ) mtt; !l


exception Error_in_construction

let rec mk_mapping' h = function
   ([],[]) -> () |
   (xs,[]) -> raise Error_in_construction |
   ([],ys) -> raise Error_in_construction |
   (x::xs,y::ys) -> Hashtbl.add h x y; mk_mapping' h (xs,ys)

let mk_states mtt fsta_states tuples debug =
   let d = Simple_table.create () in
   let index = ref 0 in
   let get () = (let r = !index in index := !index + 1; r) in
   let l = ref [] in
   let q_p_tuples =
      let l' = ref [] in
      Mtt.iter_states (fun q_sem q ->
         if List.length tuples > 0 then
            List.iter (fun t ->
               l' := (q,t) :: !l'
            ) tuples
               (* for each possiple tuple p1,..,pm of tree states *)
         else
            l' := (q,[]) :: !l'
      ) mtt; !l' in
         (* for each state q of the mtt *)
   let k = List.length q_p_tuples in
   let results = tuple k (Dictionary.to_list fsta_states) in
   let _ = if debug then
      Printf.printf "max. rank of all q: %s\n"
         (string_of_int (max_rank mtt));
      Printf.printf "number of state tuples (p1,..,pm):\t%s\n"
         (string_of_int (List.length tuples));
      Printf.printf "number of (q,p1,..,pm) tuples:\t%s\n"
         (string_of_int (List.length q_p_tuples));
      Printf.printf "number of functions f:Q x P^(m) -> P:\t%s\n"
         (string_of_int (List.length results))
   in
   List.iter (fun r ->
      let new_index = get () in
      let h = Hashtbl.create k in
      mk_mapping' h (q_p_tuples,r);
      Simple_table.add d (new_index, h);
      l := (string_of_int new_index) :: !l
   ) results; (!l,d)


let rec contains l1 = function
   [] -> true |
   x::xs -> (List.mem x l1) && (contains l1 xs)

(*
let get_d new_states q args ps =
   Simple_table.fold (fun d d_q result ->
      let p = Hashtbl.find d_q (q,args) in
      if List.mem p ps then d :: result else result
   ) new_states []
*)

let compute_d new_states =
   let h = Hashtbl.create (Simple_table.size new_states) in
   Simple_table.iter (fun d d_q ->
      Hashtbl.iter (fun (q,args) p ->
         Hashtbl.add h (q,args,p) d
      ) d_q
   ) new_states; h

let compute_d' new_states =
   let h = Hashtbl.create (Simple_table.size new_states) in
   Simple_table.iter (fun d d_q ->
      Hashtbl.iter (fun (q,args) p ->
         if (Hashtbl.mem h (q,args,p)) then
            let ds = Hashtbl.find h (q,args,p) in
            Hashtbl.replace h (q,args,p) (d :: ds)
         else
            Hashtbl.add h (q,args,p) [d]
      ) d_q
   ) new_states; h

let mk_final mtt fsta new_states =
   let q_init = Mtt.get_initial_state mtt in
   let finals = Automat.final_states fsta in
   Simple_table.fold (fun d d_q result ->
      let is_final = Hashtbl.fold (fun (q,l) p b ->
         b || (List.mem q q_init && List.mem p finals)
      ) d_q false
      in
      if is_final then d :: result else result
   ) new_states [] 

let get_mtt_states mtt symbol =
   let l = ref [] in
   let contains sym = function
      Mtt.None -> false |
      Mtt.Any -> true |
      Mtt.NoneOf l -> not (List.mem sym l) |
      Mtt.OneOf l -> List.mem sym l 
   in
   Mtt.iter_states (fun sem q ->
      let dom = Mtt.get_domain mtt q in
      if (contains symbol dom) then l := q :: !l
   ) mtt; !l

let get_rhs = function
   Mtt.Undefined -> raise Undefined_rhs |
   Mtt.Det rhs -> rhs |
   Mtt.Non_det _ -> raise Non_deterministic_mtt

let next_states automaton sym args debug =
   let result = ref [] in
   let rec eq' l1 l2 =
      (* equality test for two fsta state lists *)
      List.fold_left2 (fun b x y ->
         b && (x = y)
      ) true l1 l2 
   in 
   Automat.iter (fun (s, qs, qs') ->
      if s = sym then (
         let _ = if debug then
            Printf.printf "\tsymbol '%s' found\n\t-> testing [%s] = [%s]?" 
            (string_of_int s)
            (List.fold_left (fun s x -> 
               s ^ " " ^ (string_of_int x)
            ) "" qs)
            (List.fold_left (fun s x -> 
               s ^ " " ^ (string_of_int x)
            ) "" args)
         in
         if eq' qs args  then (
            let _ = if debug then 
               Printf.printf "\t...yes\n" 
            in
            result := qs'
         ) else (
            if debug then
               Printf.printf "\t...no\n"
         )
      ) else (
         if debug then
            Printf.printf "\tsymbol found?\t...no\n"
      )
   ) automaton; !result

let rec simul' fsta states phi rho depth debug = function
   Empty_rhs -> [] |
   Parameter i -> [ (rho i) ] |
   Call (q,j,k,rhs_list) ->
      let d = phi j in
      let states' = 
         List.map (fun rhs ->
            simul' fsta states phi rho (depth+1) debug rhs
         ) rhs_list in
      List.fold_left (fun result args ->
         (eval' states d q args) @ result
      ) [] (tuple' states') |
   Symbol (symbol,k,rhs_list) ->
      let sym = symbol in
      let _ = if debug then
         Printf.printf "[%i] current symbol:\t%s\n" depth (string_of_int sym)
      in
      let states' = 
         List.map (fun rhs ->
            simul' fsta states phi rho (depth+1) debug rhs
         ) rhs_list 
      in
      if List.length rhs_list > 0 then
         List.fold_left (fun result args ->
            let deriv_stats = next_states fsta sym args debug in
            deriv_stats @ result
         ) [] (tuple' states')
      else
         next_states fsta sym [] debug


and eval' states d q rhs_list =
   let h = Simple_table.lookup states d in 
   let t = (q,rhs_list) in
   let r = Hashtbl.find h t in
   [r]

   

let rec simul_0' mtt (*sigma*) (*fsta_states*) fsta states phi depth debug = 
function
   Empty_rhs -> [] |
   Parameter i -> [] (* no parameters defined *) |
   Call (q,j,k,rhs_list) ->
      let _ = if debug then
         Printf.printf "\t[%i] call of:\t'%s'\n"
         depth
         ((Mtt.state_to_string mtt) (Mtt.state_of_ord mtt q))
      in
      let d = phi j in
      eval_0' mtt states d q |
   Symbol (symbol,k,rhs_list) -> 
      let sym = symbol in
      let _ = if debug then
         Printf.printf "\t[%i] output:\t'%s'='%s'\n"
         depth (string_of_int sym)
         ((Mtt.delta_to_string mtt) (Mtt.delta_of_ord mtt symbol))
      in
      let states' = 
         List.map (fun rhs ->
            simul_0' mtt (*sigma*) (*fsta_states*) fsta states phi (depth+1) debug rhs
         ) rhs_list 
      in
      if List.length rhs_list > 0 then 
         List.fold_left (fun result args ->
            let deriv_stats = next_states fsta sym args debug in
            let _ = if debug then
               Printf.printf "\t[%i] beta(%s,<%s>) -> [%s]\n"
               depth
               ((Mtt.delta_to_string mtt) (Mtt.delta_of_ord mtt symbol))
               (List.fold_left (fun s q -> s^","^(string_of_int q)) 
                  (try string_of_int (List.hd args) with Failure f -> "")
                  (try List.tl args with Failure f -> []) 
               )
               (List.fold_left (fun s q -> s^","^(string_of_int q))
                  (try string_of_int (List.hd deriv_stats) with Failure f -> "")
                  (try List.tl deriv_stats with Failure f -> [])
               )
            in
            deriv_stats @ result
         ) [] (tuple' states')
      else
         next_states fsta sym [] false

and eval_0' mtt states d q =
   let h = Simple_table.lookup states d in 
   let r = Hashtbl.find h (q,[]) in
   [r]




let mk_rules mtt fsta sigma delta fsta_states fsta_tuples states states' debug =
   let rules = ref [] in
   let rule_counter = ref 0 in
   let error_rule_counter = ref 0 in
   let state_tuples = tuple 2 (Simple_table.to_list states) in
   let _ = Printf.printf "number of state tuples: %i\n" (List.length state_tuples) in
   let k = List.length fsta_tuples in
   if k = 0 then
         (* mtt is top-down, i.e., no parameters occurr *)
      Dictionary.iter (fun symbol symbol_sem ->
         let fsta_sym = symbol in
         let _ = if debug then
            Printf.printf "current symbol of Sigma:\t'%s'='%s'\n"
            (string_of_int fsta_sym)
            ((Mtt.sigma_to_string mtt) symbol_sem)
         in
         let mtt_states = get_mtt_states mtt symbol in
         List.iter (fun tuple ->
            let phi = fun i -> List.nth tuple (i-1) in
            List.iter (fun q ->
               let rhs = get_rhs (Mtt.get_action mtt q symbol) in
               let _ = if debug then
                  Printf.printf "   current rhs:\t%s\n"
                  (Mtt_to_string.rhs_to_string mtt rhs)
               in
               let e = 
                  simul_0' mtt (*sigma*) (*fsta_states*) fsta states phi 0 debug rhs in
               let arguments = List.map string_of_int tuple in
               let d = 
                  List.fold_left (fun ds p ->
                     let d' = 
                        try
                           Some (Hashtbl.find states' (q,[],p))
                        with
                           Not_found -> None
                     in
                     match d' with
                        Some (d) -> d :: ds |
                        None -> ds
                  ) [] e in
               let result_states = List.map string_of_int e in
               let result_ds = List.map string_of_int d in
               let _ = if debug then
                  Printf.printf "   ==>  delta(%s,<%s>) -> [%s]\n"
                  (string_of_int fsta_sym)
                  (List.fold_left (fun s a -> s^","^a)
                     (try List.hd arguments with Failure f -> "")
                     (try List.tl arguments with Failure f -> [])
                  ) 
                  (List.fold_left (fun s a -> s^","^a)
                     (try List.hd result_ds with Failure f -> "")
                     (try List.tl result_ds with Failure f -> [])
                  ) 
               in
               let _ = if !rule_counter mod 50 = 0 then
                  Printf.printf "\n" 
               in
               if List.length d = 0 then (
                  (* list of result states is empty *)
                  Printf.printf "."; 
                  error_rule_counter := !error_rule_counter + 1;
                  rule_counter := !rule_counter + 1
               ) else (
                  Printf.printf "*";
                  rule_counter := !rule_counter + 1;
(*                   rules := (fsta_sym,arguments,result_ds) :: !rules *)
                  rules := (fsta_sym,e,d) :: !rules
               )
            ) mtt_states
               (* state q of mtt for which q(a(x1,..,xk),y1,..,ym) -> T *)
         ) state_tuples 
            (* d1,..,dk for a k-ary symbol of the alphabet *)
      ) sigma 
   else
         (* mtt has parameters *)
      Dictionary.iter (fun symbol symbol_sem ->
         let fsta_sym = symbol in
         let mtt_states = get_mtt_states mtt symbol in
         List.iter (fun tuple ->
            let phi = fun i -> List.nth tuple (i-1) in
            List.iter (fun q ->
               let rhs = get_rhs (Mtt.get_action mtt q symbol) in
               List.iter (fun args ->
                  let rho = fun j -> List.nth args (j-1) in
                  let e = simul' fsta states phi rho 0 false rhs in
                  let d = 
                     List.fold_left (fun ds p ->
                        let d' = 
                           try
                              Some (Hashtbl.find states' (q,args,p))
                           with
                              Not_found -> None
                        in
                        match d' with
                           Some (d) -> d :: ds |
                           None -> ds
                     ) [] e in
                  let _ = if !rule_counter mod 50 = 0 then
                     Printf.printf "\n" 
                  in
                  if List.length d = 0 then (
                     (* list of result states is empty *)
                     Printf.printf "."; 
                     error_rule_counter := !error_rule_counter + 1;
                     rule_counter := !rule_counter + 1
                  ) else (
                     Printf.printf "*";
                     rule_counter := !rule_counter + 1;
(*                      rules := (fsta_sym,List.map string_of_int tuple,List.map string_of_int d) :: !rules *)
                     rules := (fsta_sym,tuple,d) :: !rules
                  )
               ) fsta_tuples
                  (* p_1,...,p_m for a m-ary state q of the mtt *)
            ) mtt_states
               (* state q of mtt for which q(a(x1,..,xk),y1,..,ym) -> T *)
         ) state_tuples 
            (* d_1,...,d_k for a k-ary symbol of the alphabet *)
      ) sigma; 
         (* symbol a of alphabet *)
(*    !rules *)
   (!rules,!rule_counter,!error_rule_counter)




exception Empty_right_hand_side
exception Domain_error
exception Rule_error

(*
type condition =
   | Par_cond of int 
   | Var_cond of int * ((int * int list) * int) list
*)

(*
let add_cond h var cond =
   if Hashtbl.mem h var then
      let c = Hashtbl.find h var in
      Hashtbl.replace h var (cond :: c)
   else
      Hashtbl.add h var [cond]
*)

let which_d states q p =
   try
      Some (Hashtbl.find states (q,[],p))
   with
      Not_found -> None

let rec eval_on_demand mtt symbols q p fsta new_states new_states' seen rs debug =
   let rec doit result = function 
      | [] -> result 
      | symbol::l  -> 
         let action = Mtt.get_action mtt q symbol in
         let rank = Mtt.arity mtt symbol in
         let _ = Hashtbl.add seen (q,p) true in
         let _ = if debug then Printf.printf "current rule:\t%s" 
            (Mtt_to_string.rule_to_string mtt (q,symbol,action)) in
         let _ = if debug then Printf.printf "rank of current symbol:\t%i\n\n" rank in
         let r = 
            if rank > 0 then
               let initial_d = which_d new_states' q p in
               match initial_d with
                  | Some d ->
                     let rhs = get_rhs (Mtt.get_action mtt q symbol) in
                     let _ = if debug then Printf.printf "\tcurrent rhs:\t%s\n"
                        (Mtt_to_string.rhs_to_string mtt rhs) in
                     let args = Array.create rank [] in
                     let next_ds = ref [] in
                     let r = 
                        inspect_rhs (q,p,rhs) mtt fsta new_states' args seen in
                     let tmp = 
                        let l = ref [] in
                        (for i = 0 to (Array.length args)-1 do
                           let _ = if debug then Printf.printf "states for x_%i:\t%i\n"
                                   (i+1) (List.length args.(i)) in
                           let d' = 
                              if (List.length args.(i)) > 0 then
                                 List.fold_left (fun r (q,p,d) ->
                                    let _ = next_ds := (q,p) :: (!next_ds) in
                                    d :: r
                                 ) [] args.(i) 
                              else
                                 (* variable x_i is not used on the rhs, thus 
                                    every input is allowed *)
                                 Simple_table.to_list new_states
                           in
                           l := d' :: (!l)
                        done; (!l)) in
                     let next =
                        List.fold_left (fun result (q,p) ->
                           let domain = Mtt.get_domain mtt q in
                           let rules' =
                              match domain with
                                 | Mtt.OneOf symbols ->
                                      let _ = if debug then Printf.printf "rec call for mtt state %i and fsta state %i" q p in
                                      if (Hashtbl.mem seen (q,p)) then
                                         let _ = if debug then Printf.printf " ...rejected (already computed)\n" in
                                         []
                                      else
                                         let _ = if debug then Printf.printf " ...\n" in
                                         eval_on_demand mtt symbols q p fsta
                                         new_states new_states' seen rs debug
                                 | _ -> raise Domain_error
                           in
                           List.fold_left (fun r (s,a,a') ->
                              if not (Simple_table.mem rs (s,a,a')) then
                                 let _ = Simple_table.add rs ((s,a,a'), true)
                                 in (s,a,a') :: r
                              else r
                           ) result rules'
                        ) [] (!next_ds) in
                     let args' = tuple' tmp in
                     List.fold_left (fun result a ->
                        List.fold_left (fun r' d' ->
                           if not (Simple_table.mem rs (symbol,a,[d'])) then
                              let _ = Simple_table.add rs ((symbol,a,[d']), true)
                              in (symbol,a,[d']) :: r'
                           else r'
                        ) result d
                     ) next args'
                  | None -> []
            else
               let rhs = get_rhs (Mtt.get_action mtt q symbol) in
                  (* since the rank equals 0, we do not care about
                     arguments, and therefore use the identity function
                     and define the argument list as [] *)
               let phi x = x in
               let args = [] in
               let e = simul_0' mtt fsta new_states phi 0 false rhs in
               let d = 
                  List.fold_left (fun ds p ->
                     match (which_d new_states' q p) with
                        | Some d ->
                            List.fold_left (fun r d' ->
                               d' :: r
                            ) ds d
                        | None -> ds
                  ) [] e in
               let new_rule =
                  if not (Simple_table.mem rs (symbol,args,d)) then
                     let _ = Simple_table.add rs ((symbol,args,d), true)
                     in [(symbol,args,d)]
                  else []
               in
                  new_rule in
         let _ = if debug then Printf.printf "\n" 
         in
            doit (r @ result) l
   in
      doit [] symbols


and inspect_rhs (q,p,rhs) mtt fsta new_states args seen debug =
   match rhs with
      | Symbol (s,k,rhs_list) -> (
           let _ = if debug then Printf.printf "\tsymbol '%s'\n" (Mtt_to_string.delta_to_string mtt s) in
           let next_ps = Array.create k [] in
           let pos = ref 0 in
           let _ =
              Automat.iter (fun (s,p',q') ->
                 let _ = if debug then Printf.printf "\tsearching for next ps\n" in
                 if (List.mem p q') then
                    let _ = if debug then Printf.printf "\tfound next ps\n" in
                    let i = ref 0 in
                    List.iter (fun p ->
                       (next_ps.(!i) <- p :: next_ps.(!i); 
                       i := (!i) + 1)
                    ) p'
                 else
                    ()
              ) fsta
           in
           List.iter (fun rhs' ->
              let ps = next_ps.(!pos) in
              List.iter (fun p' ->
                 inspect_rhs (q,p',rhs') mtt fsta new_states args seen debug
              ) ps;
              pos := (!pos) + 1
           ) rhs_list
        )
      | Call (q',x,k,rhs_list) -> (
           let _ = if debug then Printf.printf "\tcall \n" in
           match (which_d new_states q' p) with
              | Some d ->
                  (args.(x-1) <-
                     List.fold_left (fun r d' ->
                        (q',p,d') :: r
                     ) args.(x-1) d
                  )
              | None -> ()
        )
      | _ -> if debug then Printf.printf "** Error\n"



let which_d' states q args p =
   try
      Some (Hashtbl.find states (q,args,p))
   with
      Not_found -> None

let rec eval_on_demand' mtt symbols q p fsta new_states new_states' seen rs debug =
   let rec doit result = function 
      | [] -> result 
      | symbol::l  -> 
         let action = Mtt.get_action mtt q symbol in
         let rank = Mtt.arity mtt symbol in
         let par_num = Mtt.num_of_parameters mtt q in
         let _ = Hashtbl.add seen (q,p) true in
         let _ = if debug then Printf.printf "current rule:\t%s" 
            (Mtt_to_string.rule_to_string mtt (q,symbol,action)) in
         let _ = if debug then Printf.printf "rank of current symbol:\t%i\n" rank in
         let r = 
            let rhs = get_rhs (Mtt.get_action mtt q symbol) in
            let _ = 
                    if debug then Printf.printf "\tcurrent rhs:\t%s\n"
               (Mtt_to_string.rhs_to_string mtt rhs) in
            let args = Array.create rank [] in
            let pars = Array.create par_num [] in
            let next_ds = ref [] in
            let _ = if debug then Printf.printf "\tinspecting rhs...\n\n" in
            let r = 
               inspect_rhs' (q,p,rhs) rank mtt fsta new_states' args pars seen
               debug in
            let xs = 
               let l = ref [] in
                  begin
                     for i = 0 to (Array.length args)-1 do
                        let _ = if debug then Printf.printf "states for x_%i:\t%i\n"
                                (i+1) (List.length args.(i)) in
                        let d' = 
                           if (List.length args.(i)) > 0 then
                              List.fold_left (fun r (q,p,d,a) ->
                                 let _ = next_ds := (q,p) :: (!next_ds) in
                                 d :: r
                              ) [] args.(i) 
                           else
                              (* variable x_i is not used on the rhs, thus 
                                 every input is allowed *)
                              Simple_table.to_list new_states
                        in
                        l := d' :: (!l)
                     done; 
                     (!l)
                  end  in
            let ys =
               let l = ref [] in
                  begin
                     for i = 0 to (Array.length pars)-1 do
                        let p' =
                           if (List.length pars.(i)) > 0 then
                              pars.(i)
                           else
                              Automat.states fsta
                        in
                        l := p' :: (!l)
                     done;
                     (!l)
                  end in
            let next =
               List.fold_left (fun result (q,p) ->
                  let domain = Mtt.get_domain mtt q in
                  let rules' =
                     match domain with
                        | Mtt.OneOf symbols ->
                             let _ = if debug then Printf.printf "rec call for mtt state %i and fsta state %i" q p in
                             if (Hashtbl.mem seen (q,p)) then
                                let _ = if debug then Printf.printf " ...rejected (already computed)\n" in
                                []
                             else
                                let _ = if debug then Printf.printf " ...\n" in
                                eval_on_demand mtt symbols q p fsta
                                new_states new_states' seen rs debug
                        | _ -> raise Domain_error
                  in
                  List.fold_left (fun r (s,a,a') ->
                     if not (Simple_table.mem rs (s,a,a')) then
                        let _ = Simple_table.add rs ((s,a,a'), true)
                        in (s,a,a') :: r
                     else r
                  ) result rules'
               ) [] (!next_ds)
            in
            let args' = tuple' xs in
            let pars' = tuple' ys in
            if rank > 0 then
               List.fold_left (fun result a ->
                  List.fold_left (fun res ps ->
                     let ds = which_d' new_states' q ps p in
                     match ds with
                        | Some d ->
                           List.fold_left (fun r' d' ->
                              if not (Simple_table.mem rs (symbol,a,[d'])) then
                                 let _ = Simple_table.add rs ((symbol,a,[d']), true)
                                 in (symbol,a,[d']) :: r'
                              else r'
                           ) res d
                        | None -> res
                  ) result pars'
               ) next args'
            else
               let a = [] in
               List.fold_left (fun res ps ->
                  let ds = which_d' new_states' q ps p in
                  match ds with
                     | Some d ->
                        List.fold_left (fun r' d' ->
                           if not (Simple_table.mem rs (symbol,a,[d'])) then
                              let _ = Simple_table.add rs ((symbol,a,[d']), true)
                              in (symbol,a,[d']) :: r'
                           else r'
                        ) res d
                     | None -> res
               ) [] pars'
         in
         let _ = if debug then Printf.printf "\n" 
         in
            doit (r @ result) l
   in
      doit [] symbols

and inspect_rhs' (q,p,rhs) rank mtt fsta new_states args pars seen debug =
   match rhs with
      | Symbol (s,k,rhs_list) ->
           begin
              let _ = if debug then  
                 Printf.printf "\tsymbol '%s'\n" 
                 (Mtt_to_string.delta_to_string mtt s) in
              let next_ps = Array.create k [] in
              let pos = ref 0 in
              let _ =
                 Automat.iter (fun (s,p',q') ->
                    if (List.mem p q') then
                       let i = ref 0 in
                       List.iter (fun p ->
                          (next_ps.(!i) <- p :: next_ps.(!i); 
                          i := (!i) + 1)
                       ) p'
                    else
                       ()
                 ) fsta
              in
              List.iter (fun rhs' ->
                 let ps = next_ps.(!pos) in
                 List.iter (fun p' ->
                    inspect_rhs' (q,p',rhs') rank mtt fsta new_states args pars seen debug
                 ) ps;
                 pos := (!pos) + 1
              ) rhs_list
           end
      | Call (q',x,k,rhs_list) ->
           begin
              let _ = if debug then  
                 Printf.printf "\tcall of '%s'\n" 
                 (Mtt_to_string.state_to_string mtt q') in
              let next_ps = Array.create k [] in
              let pos = ref 0 in
              let _ = 
                 Hashtbl.iter (fun (q,a,s) d ->
                    if p = s then
                       let i = ref 0 in
                       List.iter (fun s' ->
                          (next_ps.(!i) <- p :: next_ps.(!i);
                          i := (!i) + 1)
                       ) a
                    else
                       ()
                 ) new_states in
              let _ = 
                 List.iter (fun rhs' ->
                    let ps = next_ps.(!pos) in
                    List.iter (fun p' ->
                       inspect_rhs' (q',p',rhs') rank mtt fsta new_states args pars seen debug
                    ) ps;
                    pos := (!pos) + 1
                 ) rhs_list in
              let arg_list = tuple' (Array.to_list next_ps)
              in   
              List.iter (fun arguments ->
                 match (which_d' new_states q' arguments p) with
                    | Some d ->
                        (args.(x-1) <-
                           List.fold_left (fun r d' ->
                              (q',p,d',arguments) :: r
                           ) args.(x-1) d
                        )
                    | None -> ()
              ) arg_list
           end
      | Parameter j ->
           begin
              let _ = if debug then  
                 Printf.printf "\tparameter 'y_%i'\n" (j) in
              pars.(j-1) <- p :: pars.(j-1)
           end
      | _ -> if debug then Printf.printf "** Error\n"


(*
   ..Call
           let next_ps = Array.create k [] in
           let pos = ref 0 in
           let _ = 
              Hashtbl.iter (fun (q,a,s) d ->
                 if p = s then
                    let i = ref 0 in
                    List.iter (fun s' ->
                       (next_ps.(!i) <- p :: next_ps.(!i);
                       i := (!i) + 1)
                    ) a
                 else
                    ()
              ) new_states in
           let _ = 
              List.iter (fun rhs' ->
                 let ps = next_ps.(!pos) in
                 List.iter (fun p' ->
                    inspect_rhs' (q',p',rhs') mtt fsta new_states args pars seen
                 ) ps;
                 pos := (!pos) + 1
              ) rhs_list in
           let arg_list = tuple' (Array.to_list next_ps)
           in   
           List.iter (fun arguments ->
              match (which_d' new_states q' arguments p) with
                 | Some d ->
                     (args.(x-1) <-
                        List.fold_left (fun r d' ->
                           (q',p,d',arguments) :: r
                        ) args.(x-1) d
                     )
                 | None -> ()
           ) arg_list
        )

   ..Symbol
           let next_ps = Array.create k [] in
           let pos = ref 0 in
           let _ =
              Automat.iter (fun (s,p',q') ->
                 if (List.mem p q') then
                    let i = ref 0 in
                    List.iter (fun p ->
                       (next_ps.(!i) <- p :: next_ps.(!i); 
                       i := (!i) + 1)
                    ) p'
                 else
                    ()
              ) fsta
           in
           List.iter (fun rhs' ->
              let ps = next_ps.(!pos) in
              List.iter (fun p' ->
                 inspect_rhs' (q,p',rhs') mtt fsta new_states args pars seen
              ) ps;
              pos := (!pos) + 1
           ) rhs_list

   ..Parameter
           pars.(j) <- p :: pars.(j)

   ..

*)



and mk_rules_on_demand mtt fsta sigma delta fsta_tuples states states' debug =
   let rules = ref [] in
   let rs = Simple_table.create () in
   let rule_counter = ref 0 in
   let error_rule_counter = ref 0 in
   let k = List.length fsta_tuples in
   let fin = Automat.final_states fsta in
   let all = Automat.states fsta in
   let _ = 
      if k = 0 then (
            (* mtt is top-down, i.e. no parameters occurr *)
         let _ = if debug then Printf.printf "top-down transducer\n" in
         let q0 = List.hd (Mtt.get_initial_state mtt) in
         let seen = Hashtbl.create 101 
         in
            List.iter (fun p ->
               let domain = Mtt.get_domain mtt q0 in
               let rules' = 
                  match domain with
                     | Mtt.OneOf symbols -> 
                        eval_on_demand mtt symbols q0 p fsta 
                           states states' seen rs debug
                     | _ -> raise Domain_error in
               let _ = rule_counter := !rule_counter + (List.length rules')
               in
                  rules := rules' @ (!rules)
            ) all (*fin*)
      ) else (
            (* mtt has parameters *)
         let _ = if debug then Printf.printf "macro transducer\n" in
         let q0 = List.hd (Mtt.get_initial_state mtt) in
         let seen = Hashtbl.create 101 
         in
            List.iter (fun p ->
               let domain = Mtt.get_domain mtt q0 in
               let rules' =
                  match domain with
                     | Mtt.OneOf symbols -> 
                        eval_on_demand' mtt symbols q0 p fsta 
                           states states' seen rs debug
                     | _ -> raise Domain_error in
               let _ = rule_counter := !rule_counter + (List.length rules')
               in
                  rules := rules' @ (!rules)
            ) all
      ) 
   in
      (!rules,(Simple_table.size rs),!rule_counter,!error_rule_counter)




let pre mtt fsta delta fsta_states debug =
   let _ = if debug then (Printf.printf "mtt:\n"; Mtt_to_string.print mtt) in
   let _ = if debug then 
      Printf.printf "\n\nfsta:\n%s\n\n" (Automat.to_short_string fsta) in
   let sigma = Dictionary.create () in
   let sigma_list = mk_sigma mtt sigma in
   let rank = max_rank mtt in
      (* tuples : (int list) list *)
   let tuples = tuple rank (Dictionary.to_list fsta_states) in
   let (state_list,d) = mk_states mtt fsta_states tuples true in
   let _ = if debug then
      Printf.printf "\nnumber of computed states:\t%i\n\n"
      (List.length state_list) in
   let _ = if debug then
      Printf.printf "states of the new automaton:\n" in
   let _ = if debug then Simple_table.iter (fun i e ->
       Printf.printf "%i ->\n" i;
       Hashtbl.iter (fun (q,l) p ->
          Printf.printf "\t(%s%s) -> %s\n"
          ((Mtt.state_to_string mtt) (Mtt.state_of_ord mtt q))
          (if List.length l > 0 then List.fold_left (fun r a ->
             r ^ "," ^ (string_of_int a)
          ) (", "^(string_of_int (List.hd l))) (List.tl l) else "")
          (string_of_int p)
       ) e;
       Printf.printf "\n"
   ) d in
   let d' = compute_d' d in
   let (rules,rule_num,rules',errors) =
      mk_rules_on_demand mtt fsta sigma delta tuples d d' debug in
   let _ = Printf.printf "\n*** number of rules: %i\n" rules' in
   let _ = Printf.printf "*** number of (real) rules: %i\n" rule_num in
   let _ = Printf.printf "*** number of error rules: %i\n\n" errors in
   let finals = mk_final mtt fsta d in
   let a = Automat.from_list rules finals in
   (a, sigma, d)
(*    (Automat.from_list [] [], sigma, Simple_table.create ()) *)

let pre_2 mtt fsta delta fsta_states debug =
   let _ = if debug then (Printf.printf "mtt:\n"; Mtt_to_string.print mtt) in
   let _ = if debug then 
      Printf.printf "\n\nfsta:\n%s\n\n" (Automat.to_short_string fsta) in
   let sigma = Dictionary.create () in
   let sigma_list = mk_sigma mtt sigma in
   let rank = max_rank mtt in
      (* tuples : (int list) list *)
   let tuples = tuple rank (Dictionary.to_list fsta_states) in
   let (state_list,d) = mk_states mtt fsta_states tuples debug in
   let d' = compute_d d in
   let _ = if debug then
      Printf.printf "\nnumber of computed states:\t%i\n\n"
      (List.length state_list) in
   let _ = if false then
      Printf.printf "states of the new automaton:\n" in
   let _ = if true then Simple_table.iter (fun i e ->
       Printf.printf "%i ->\n" i;
       Hashtbl.iter (fun (q,l) p ->
          Printf.printf "\t(%s%s) -> %s\n"
          ((Mtt.state_to_string mtt) (Mtt.state_of_ord mtt q))
          (if List.length l > 0 then List.fold_left (fun r a ->
             r ^ "," ^ (string_of_int a)
          ) (", "^(string_of_int (List.hd l))) (List.tl l) else "")
          (string_of_int p)
       ) e;
       Printf.printf "\n"
   ) d in
   let d'' = compute_d' d in
   let (rules,rules',errors) =
      mk_rules mtt fsta sigma delta fsta_states tuples d d' debug in
   let _ = Printf.printf "\nnumber of rules: %i\n" rules' in
   let _ = Printf.printf "number of error rules: %i\n\n" errors in
   let finals = mk_final mtt fsta d in
   let a = Automat.from_list rules finals in
   (a, sigma, d)
