module type Domain =
   sig
      type t
      val lub : t -> t -> t
      val eq : t -> t -> bool
      val bot : t
   end

module type ItemEq =
   sig
      type t
      val eq : t -> t -> bool
   end

module type Solver =
   sig
      module I : ItemEq
      module D : Domain

      type item 
      type domain 

      type assign = item -> domain
      type system = item -> (assign -> domain)
      type equation = item * (assign -> domain)

      val solve : system -> item list -> (item,domain) Hashtbl.t
   end

module Make (Item : ItemEq) (Domain : Domain) =
   struct
      module I = Item
      module D = Domain
      module W = Worklist

      type item = I.t
      type domain = D.t

      type assign = item -> domain
      type system = item -> (assign -> domain)
      type equation = item * (assign -> domain)

      let init = 101

      let solve system xs =
         let sigma = Hashtbl.create init in
         let influence = Hashtbl.create init in
         let wl = W.empty () in
      
         let sigm x = 
            try Hashtbl.find sigma x
            with Not_found -> (Hashtbl.add sigma x D.bot; D.bot) in
         let set_sigm x y = Hashtbl.replace sigma x y in
      
         let empty = [] in
      
         let infl x = 
            try Hashtbl.find influence x
            with Not_found -> (Hashtbl.add influence x empty; empty) in
         let set_infl x y = Hashtbl.replace influence x y in
         let clear_infl x = Hashtbl.replace influence x empty in
         let add_infl x y = Hashtbl.replace influence x (y :: (infl x)) in
      
         let rec eval x y =
            _solve y;
            add_infl y x;
            sigm y

         and _solve x =
            let f = system x in
            let old = sigm x in
            let n = D.lub old (f (eval x)) in
               if not (D.eq n old) then
                  (set_sigm x n;
                   W.add wl (infl x);
                   clear_infl x) in

         let rec loop () = 
            if not (W.is_empty wl) then
               let x = W.extract wl in
                  (_solve x; loop ())
         in 
            (W.add wl xs; loop (); sigma)
   end
