(** Solver for Systems of Equations.

    This module implements a solver for general systems of equations.
    The solver uses a simple worklist algorithm to solve the equation
    system. 
    
    The algorith proceeds as follows. The set of variables yet to be
    evaluated is kept in a worklist. It is initialized with the set 
    {i X} of variables in which we are interested. For every variable
    {i x} considered so far, we (globally) maintain the current value
    together with a set {i infl}({i x}) of certain variables {i y}
    such that the evaluation of the right-hand side for {i y} may
    access value {i s}({i x}).
    
    As long as the worklist is nonempty, the algorithm iteratively
    extracts some variable {i x} from teh worklist and evaluates the
    right-hand side {i f} of {i x} on the current partial variable
    assignment {i s}. If the least upper bound of the old value
    {i s}({i s}) and {i f}({i s}) is different from the old value,
    the value of {i s} for all {i y} of {i infl}({i x}), may no longer
    be valid; therefore, they are added to the worklist. Afterwards,
    {i infl}({i x}) is reset to the empty set. *)


module type Domain =
   sig
      type t
         (** Type of domain. *)
      val lub : t -> t -> t
         (** The least upper bound of two elements of the domain. *)
      val eq : t -> t -> bool
         (** [eq x y] is [true] if [x] = [y], otherwise it is [false]. *)
      val bot : t
         (** The bottom element of the domain. *)
   end
(** Input signature of functor {!Solve.Make} representing the domain
    of the equations. *)

module type ItemEq =
   sig
      type t
         (** Type of the items. *)
      val eq : t -> t -> bool
         (** [eq x y] is [true] if [x] = [y], otherwise it is [false]. *)
   end
(** Input signature of functor {!Solve.Make} representing the variables. *)

module type Solver =
   sig
      module I : ItemEq
         (** Structure representing the items. *)
      module D : Domain
         (** Structure representing the domain. *)

      type item 
         (** Type of the items. *)
      type domain 
         (** Type of the domain. *)

      type assign = item -> domain
         (** Type of an assignment. *)
      type system = item -> (assign -> domain)
         (** Type of a system of equations. *)
      type equation = item * (assign -> domain)
         (** Type of an equation. *)

      val solve : system -> item list -> (item,domain) Hashtbl.t
         (** [solve s l] solves the given system [s] of equations at
             least for the variables given by [l]. *)
   end
(** Output signature of functor {!Solve.Make}. *)

module Make 
   (Item : ItemEq) 
   (Domain : Domain) : Solver with type item = Item.t 
                              and type domain = Domain.t
(** Functor building an implementation of the Solve structure given a
    variable set and a domain. *)
